
xVersion" 2011.11.5.1"

module NumConvert

// ////////////////////////////////////////////////////////////////////////////
// Numeric conversions
// ////////////////////////////////////////////////////////////////////////////


private

uses
  utils strings
endUses


  variable hld                      #// Pointer in hold buffer
  variable hld-len                  #// Length of hold buffer
  variable hld-init                 #// Initial hold buffer pointer (at and of buffer)
  variable #base                    #// Target number base, [[base]] is a non-smart meta word already
  variable double                   #// True when double number (containing dot) was parsed
  variable dp                       #// Location of dot in last parsed number

  varInterface hld
  varInterface hld-len
  varInterface hld-init
  varInterface #base
  varInterface double
  varinterface dp


  // //////////////////////////////////////////////////////////////////////////

  : #binary  ( -- ) %10 #base! ;
  : #decimal ( -- ) &10 #base! ;
  : #hex     ( -- ) $10 #base! ;


  // //////////////////////////////////////////////////////////////////////////

  : setup# ( $ram -- )

    #// Setup a user supplied buffer for numeric conversion
    #// Also see [[#Base@]] and [[#Base!]]
    #// For compatibility with older software use [[Init#]] (deprecated).

    hld-len!                        // Store buffer length
    hld-len@ +                      // Point to end of buffer
    hld-init!                       // Store that as the initial buffer pointer
  ;


  // //////////////////////////////////////////////////////////////////////////

  : init# ( $ram nBase -- )

    #// Setup a user supplied buffer for numeric conversion and supply a
    #// number base to use.
    #// Preferred word to use is [[Setup#]], this one stays in for
    #// backward compatibility with some older controller ware.
    #// Also see [[#Base@]] and [[#Base!]]

    #base!                          // Store number base to use
    setup#
  ;
    deprecated


  // //////////////////////////////////////////////////////////////////////////

  : hold ( n -- )

    #// Store char n into hld buffer

    hld--
    hld@ c!
  ;


  // //////////////////////////////////////////////////////////////////////////

  : $hld ( -- $ram )

    hld@                            // Get current pointer
    hld-init@                       // Get initial pointer
    over                            // Get current over it
    -                               // calculate length
  ;

  // //////////////////////////////////////////////////////////////////////////

  : <# ( -- )

    #// Init numerical conversion string

    hld-init@
    hld!
  ;


  // //////////////////////////////////////////////////////////////////////////

  alias $hld #>

  // : #> ( -- $ram )
  //
  //   #// End numeric conversion string
  //   #//
  //   #// This returns a forth string residing in the buffer supplied
  //   #// with calling [[Setup#]]
  //
  //   $hld                            // Return the built string
  // ;


  // //////////////////////////////////////////////////////////////////////////

  : sign ( n -- )

    #// When n < 0 append '-' to hld buffer

    d0<                             // If negative
    if
      '- hold                       // Append '-'
    endIf
  ;


  // //////////////////////////////////////////////////////////////////////////

  : #space ( -- )

    #// Adds a space to the holding buffer

    bl hold
  ;


  // //////////////////////////////////////////////////////////////////////////

  : #'0 ( -- )

    #// Adds a zero to the holding buffer

    '0 hold
  ;


  // //////////////////////////////////////////////////////////////////////////

  : (spaces) ( n -- )

    #// Adds b spaces to the holding buffer

    begin
      dup 0>
    while
      #space 1-
    repeat
    drop
  ;


  // //////////////////////////////////////////////////////////////////////////

  : (zeros) ( n -- )

    #// Adds b zeros to the holding buffer

    begin
      dup 0>
    while
      #'0 1-
    repeat
    drop
  ;


  // //////////////////////////////////////////////////////////////////////////

  : # ( ud1 -- ud2 )

    #// Convert one digit

    #base@ u>d                      // Number base as 32 bits unsigned
    ud/mod                          // ud1 / udNumBase -> udRemainder udQuotient
    dswap drop                      // udQuotient uRemainder
    9 over <                        // If 9 < remainder
    if
      7+                            // Add 7 ( 10 -> 'A')
    endIf
    '0+                             // Make ASCII
    Hold                            // Store into buffer
  ;



  // //////////////////////////////////////////////////////////////////////////

  : #s ( d -- )

    #// Convert all digits

    begin
      #                             // Convert a digit
      ddup d0=                      // If result = 0 stop converting digits
    until
    ddrop
  ;


  // //////////////////////////////////////////////////////////////////////////

  : (ud.) ( d -- $ram )

    #// 32 bits unsigned conversion

    <# #s #>
  ;


  // //////////////////////////////////////////////////////////////////////////

  : (u.) ( n -- $ram )

    #// 16 bits unsigned conversion

    u>d (ud.)
  ;


  // //////////////////////////////////////////////////////////////////////////

  : (d.) ( d -- $ram )

    #// 32 bits signed conversion

    ddup
    dabs
    <# #s sign #>
  ;


  // //////////////////////////////////////////////////////////////////////////

  : (.) ( n -- $ram )

    #// 16 bits signed conversion

    >d (d.)
  ;


  // //////////////////////////////////////////////////////////////////////////
  // Formatted number to asc conversion, leading spaces

  : (d.r) ( d nWidth -- $ram )

    #// 32 bit signed conversion with field width

    >r (d.) r>                      // ( nramAddr nLen nWidth )
    swap - (spaces)                 // ( nRamAddr )
    drop $hld
  ;


  // //////////////////////////////////////////////////////////////////////////

  : (.r) ( n nWidth -- $ram )

    #// 16 bit signed conversion with field width

    >r >d r> (d.r)
  ;


  // //////////////////////////////////////////////////////////////////////////

  : (ud.r) ( ud nWidth -- $ram )

    #// 32 bit unsigned conversion with field width

    >r (ud.) r>
    swap - (spaces)
    drop $hld
  ;


  // //////////////////////////////////////////////////////////////////////////

  : (u.r) ( u bWidth -- $ram )

    #// 16 bit unsigned conversion with field width

    >r u>d r> (ud.r)
  ;


  // //////////////////////////////////////////////////////////////////////////
  // Formatted number to asc conversion, leading zeros

  : (d.rz) ( d nWidth -- $ram )

    #// 32 bit signed conversion with field width

    >r
      ddup
      dabs
      (ud.)
    r>
    swap - (zeros) sign
    drop $hld
  ;


  // //////////////////////////////////////////////////////////////////////////

  : (.rz) ( n nWidth -- $ram )

    #// 16 bit signed conversion with field width

    >r >d r> (d.rz)
  ;


  // //////////////////////////////////////////////////////////////////////////

  : (ud.rz) ( ud nWidth -- $ram )

    #// 32 bit unsigned conversion with field width

    >r (ud.) r>
    swap - (zeros)
    drop $hld
  ;


  // //////////////////////////////////////////////////////////////////////////

  : (u.rz) ( u nWidth -- $ram )

    #// 16 bit unsigned conversion with field width

    >r u>d r> (ud.rz)
  ;



  // --------------------------------------------------------------------------

  code >number ( ud $ram -- ud $ram ) // Convert digits to number

    ; // $ram = ( nAddr nLen )
    ; // ud   = ( nLow nHigh )

    push      w8                    ; // w8 is used and must be conserved.

    PopReg    w2                    ; // w2 <- Length   - from data stack
    Popreg    w3                    ; // w3 <- ram addr - from data stack
    PopReg    w1                    ; // w1 <- high ud  - from data stack
                                    ; // w0 == low  ud  - top of stack
    cp0       w2
    bra       eq, 3f                ; // B/ length zero, done

    mov       <@" #base">, w4       ; // w4 <- base

    ; // Fall through to 1: , main loop
1:
    mov.b     [w3], w5              ; // w5 <- next character

    mov       #'0', w8              ; // if < '0 done
    cp.b      w5, w8
    bra       lt, 3f                ; // B/ character < '0, done

    mov       #'9', w8              ; // if <= 9 convert it
    cp.b      w5, w8
    bra       le, 2f                ; // B/ character <= '9, no tricks needed
                                    ; //    start conversion

    and       #0xdf, w5             ; // Make character upper case

    mov       #'A', w8              ; // if < 'A' done
    cp.b      w5, w8
    bra       lt, 3f                ; // B/ character < 'A, invalid digit

    sub.b     w5, #7                ; // 'A -> $3A
2:
    ; // Convert digit, '0 relative still

    sub.b     w5, #'0'              ; // '0 -> $00
    cp.b      w5, w4                ; // Check against base
    bra       ge, 3f                ; // B/ digit >= base, invalid digit

    mul.uu    w0, w4, w6            ; // w7:w6 <- base * ud lo
    mul.uu    w1, w4, w0            ; // w1:w0 <- base * ud hi
    mov       w0, w1                ; // High part into w1 again
    add       w6, w5, w0            ; // Add digit to new ud
    addc      w1, w7, w1            ; // Add overflow from lo mult plus carry
                                    ; // from add into hi word
    inc       w3, w3                ; // ram addr ++
    dec       w2, w2                ; // Length--
    bra       nz, 1b                ; // B/ Length <> 0, convert more
3:
                                    ; //      new ud lo is in w0 - top of stack
    PushReg   w1                    ; // Push new ud hi          - to data stack
    PushReg   w3                    ; // Push new ram addr       - to data stack
    PushReg   w2                    ; // Push new Length         - to data stack

    pop       w8                    ; // Restore w8

    return                          ; // B/ ----> done
  endCode


  // --------------------------------------------------------------------------

  : char= ( $ram nChar -- $ram f )

    2 pick c@ =
  ;


  // --------------------------------------------------------------------------

  : and_more? ( $ram f -- $ram f )

    over 0> and
  ;

  // --------------------------------------------------------------------------

  : number? ( $ram -- d f )

    #// Scan number from $ram using current number base.
    #// Returns ( d true ) on valid number ( d0 false ) otherwise.

    double.clear                    // Initially not a double number
    -1 dp!                          // Set decimal point position to -1
    '+ ramSkip                      // Ignore leading +
    '- char=                        // Check if number starts with a minus
    and_more? dup>r                 // Sign on loop stack (true when negative)
    if
      1 /ramstring                  // Proceed one byte in $ram
    endIf
    dup 0=                          // When zero remaining characters
    if                              // not a valid number
      r>drop                        // Drop sign from loop stack
      ddrop d0 false                // Always return zero on failure
      exit
    endIf
    d0 dswap                        // Start number with zero value
    >number                         // Convert digits
    '. char=                        // When next char is a '.'
    and_more?                       //   and more chars are available
    if
      dup 1- dp!                    // Store dot position into dp
      1 /ramstring                  // Proceed one byte in $ram
      >number                       // Convert more digits
      dup 0=                        // When zero characters remaining
      if
        double.set                  // flag last conversion as a valid double number
      endIf
    endIf
    nip                             // Drop ram address
    0=                              // When remaining length is zero it is a valid number
    r>                              // Get sign from loop stack (true when negative)
    if
      >r dnegate r>                 // Negate result when there was a leading minus
    endIf
  ;


  // //////////////////////////////////////////////////////////////////////////

public

  export  Init#   ( $ram nBase  --      ) // Setup a user supplied buffer for numeric conversion, deprecated
  export  Setup#  ( $ram        --      ) // Setup a user supplied buffer for numeric conversions, preferred over Init#
  export  #base@  (             -- n    ) // Get the number base used for conversions
  export  #base!  ( n           --      ) // Set the number base used for conversions
  export  Hold    ( n           --      ) // Store b into hld buffer
  export  Sign    ( d           --      ) // When d < 0 append '-' to hld buffer
  export  #       ( d1          -- d2   ) // Convert one digit
  export  #s      ( d           --      ) // Convert all digits -- d0 == 0
  export  <#      (             --      ) // Init numerical conversion
  export  #>      (             -- $ram ) // forth string :: End numeric conversion
  export  (u.)    ( n           -- $ram ) // forth string :: 16 bit unsigned conversion
  export  (.)     ( n           -- $ram ) // forth string :: 16 bit signed   conversion
  export  (.r)    ( n  nWidth   -- $ram ) // forth string :: 16 bit signed   conversion leading spaces
  export  (u.r)   ( u  nWidth   -- $ram ) // forth string :: 16 bit unsigned conversion leading spaces
  export  (.rz)   ( n  nWidth   -- $ram ) // forth string :: 16 bit signed   conversion leading zeros
  export  (u.rz)  ( u  nWidth   -- $ram ) // forth string :: 16 bit unsigned conversion leading zeros
  export  (ud.)   ( d           -- $ram ) // forth string :: 32 bit unsigned conversion
  export  (d.)    ( d           -- $ram ) // forth string :: 32 bit signed   conversion
  export  (d.r)   ( d  nWidth   -- $ram ) // forth string :: 32 bit signed   conversion leading spaces
  export  (ud.r)  ( ud nWidth   -- $ram ) // forth string :: 32 bit unsigned conversion leading spaces
  export  (d.rz)  ( d  nWidth   -- $ram ) // forth string :: 32 bit signed   conversion leading zeros
  export  (ud.rz) ( ud nWidth   -- $ram ) // forth string :: 32 bit unsigned conversion leading zeros
  export  number? ( $ram        -- d f  ) // Scan number from string using current number base, true on valid
  export  double? (             -- f    ) // True when last converted number was a double (had a dot)
  export  dp@     (             -- n    ) // Position of dot in last converted number

endModule

