
xVersion" 2012.10.10.1"

module work

  #document work

    // The main module

  #endDoc

private

  uses
    fatal queues timers hw_interface utils sets strings numconvert
    memory compiler_tests
  endUses


  variable      startupCause
  varInterface  startupCause


public


  uses
    fatal
    hw_config
    hw_setup
    hw_interface
    tmr1
    spi1
    uart1
    strings
    flash
    memory
    admin
    compiler_tests
    forthInterpreter
  endUses


  // //////////////////////////////////////////////////////////////////////////

  512 3 *    constant SectorSize            // Size of internal flash sector
  SectorSize array    SectorBuffer          // Sector buffer

  2variable           CurrentFlashOffset    // Current offset on external flash
  2variable           CurrentMemoryOffset   // Current offset in inernal flash
  2variable           ImageSize             // Remaining size of flash image

  arrayInterface      SectorBuffer
  dvarInterface       CurrentFlashOffset
  dvarInterface       CurrentMemoryOffset
  dvarInterface       ImageSize

  // //////////////////////////////////////////////////////////////////////////


  : SetupFlashStart ( -- )

    InitFlash1
    BootLoaderStart ddup
    CurrentFlashOffset!
    CurrentMemoryOffset!

    BootLoaderCount@ 2. d-                  // Get count minus CRC
    ImageSize!
  ;

  // //////////////////////////////////////////////////////////////////////////

  : ReadFlashData ( -- )

    #// Read from external flash

    CurrentFlashOffset@
    ImageSize@ SectorSize 0 d>
    if
      SectorSize
    else
      ImageSize@ drop
    endif
    @SectorBuffer
    Flash1ReadBlockRam                      // ( dFlashAddr wCount wRamAddr  --    )
  ;


  // //////////////////////////////////////////////////////////////////////////

  : WriteMemoryData ( -- )

    #// Write to internal flash

    CurrentMemoryOffset@
    MemEraseBlock                           // ( dRomOffs                  --    )

    @SectorBuffer
    CurrentMemoryOffset@
    SectorSize
    MemWriteBlock                           // ( nRamAddr dRomOffs nCount  --    )

    @SectorBuffer
    CurrentMemoryOffset@
    SectorSize
    MemWriteBlock                           // ( nRamAddr dRomOffs nCount  --    )
    MemVerifyBlock                          // ( nRamAddr dRomOffs nCount  -- f  )
    if
      ImageSize@ SectorSize 0 d>
      if
        ImageSize@
        SectorSize 0 d-
        ImageSize!
      else
        ImageSize.clear
      endif
    else
      todo' handle verification error'
    endif
  ;


  // //////////////////////////////////////////////////////////////////////////

  : WorkFlashOperationDone ( -- )

  ;
    implements Flash1OperationDone


  // //////////////////////////////////////////////////////////////////////////

  : FlashImage ( -- )

    BootLoaderCheckImage
    if                                      // Valid image
      BootLoaderDecrypt                     // Decrypt image
      SetupFlashStart
      begin
        ImageSize@ d0 d>
      while
        KickDog
        ReadFlashData
        WriteMemoryData
      repeat
    else                                    // Invalid image
      BootLoaderTrashImage                  // Trash the flashed image
      $ffff BootLoaderState!                // Clear state
      Reset                                 // And reboot
    endif
  ;

  // //////////////////////////////////////////////////////////////////////////

  : HandleBootImage ( -- )

    #// Check for a boot image and flash it when present
    #//
    #//   $0001 : receiving an image, must be dealt with by the user app
    #//   $0002 : received a complete and valid image, flash it
    #//   $0003 : Flashed the image, trash it
    #//   $ffff : Loader is idle
    #//
    #// All other values are invalid and will be reset to $ffff

    BootLoaderFixState                      // Fix damaged loader state if needced
    BootLoaderState@                        // Check state
    case

      $0002 of
          FlashImage                        // Flash the image
          $0003 BootLoaderState!            // Advance state
          Reset                             // And reboot
        endof

      $0003 of
          BootLoaderTrashImage              // Trash the flashed image
          $ffff BootLoaderState!            // Clear state
          Reset                             // And reboot
        endof

    default
      drop                                  // Other values handled by application
    endcase
  ;

  // //////////////////////////////////////////////////////////////////////////

  : handleUart1 ( -- )

    begin
      uart1.hasData
    while
      uart1.read drop
    repeat
  ;

  // //////////////////////////////////////////////////////////////////////////

  " : forth.testcode 3 4 + . ;" string forth.testcode

  // //////////////////////////////////////////////////////////////////////////

  : main ( -- )

    #// A standalone 4th program has to implement "main" as a global
    #// public word

    forthSetup                              // Initialize the 4th virtual machine, the
                                            // very first thing to be called in the entry
                                            // word.
    disable                                 // Disable all user interrupts
    initRegs                                // Setup hardware registers
    enable                                  // Enable user interrupts

    tmr1.setup
    spi1.setup
    uart1.setup

    try

      testBuiltins

      forth.init
      forth.testcode forth.rom.accept
   // forth.testcode uart1.writeRom

      // HandleBootImage

      begin
        ?stackEmpty                         // Assert forth data stack to be empty
        kickDog                             // Keep the watchdog happy
        handleUart1                         // Keep uart rx buffer empty
      again                                 // Loop forever

    except
      .// Outer exception handler
      breakPoint LoaderFailed               // Stop the debugger here - loader failed
      reset                                 // Reset the processor for now, must make an
                                            // outer exception handler later

    endTry

  noreturn; global                          // Main must be global for the linker to be
                                            // able to find it and it must be public for
                                            // the 4th system to be able to find it.

endModule

