
xVersion" 2012.10.10.1"

module hw_config

  #document

    // hw_config.4th, processor configuration file
    //
    // Used to set up peripheral devices for use with a particular
    // application.
    //
    // This module uses the name space hw_bit_defs as defined in
    // the processor dependent include file. The inclusion is automatic
    // through the 24xxxx.4th library provided that a valid processor was
    // defined in make.4th The selected processor is available through
    // the (string) variable ProcessorName in the Forth vocabulary. The
    // include file used is available through the ProcessorIclude (string)
    // variable also in the Forth vocabulary.
    //
    // Suppose the make file used  :
    //
    //   #define 24fj256da206
    //
    // Then ProcessorName    will have the string value " 24fj256da206"
    // and  ProcessorInclude will have the string value " hw_p24fj256da206.4th"
    //
    // The processor specific include file defines all of the names for
    // registers and bits as obtained through a translation of the
    // microchip processor dependent asm include file (using the x4th
    // built in translation utilty).
    //
    // Note :
    //
    //   This file was set up originally for the 24fj256da206 processor
    //   meaning that not all registers will be defined when another
    //   processor is being used. Other registers, mainly for I/O pin
    //   routing setup for the low pin count devices are not mentioned
    //   here.
    //

  #endDoc

public

  uses
    hw_bit_defs                   // From hw_p$ProcessorName$.4th
  endUses


  // ==========================================================================
  //
  //    Configuration Words
  //
  //    Configuration words exist in Program Space and their locations are
  //    defined in the device linker script. They can be set in source code
  //    or in the MPLAB IDE. Each configuration word should be specified
  //    only once (multiple settings may be combined.)
  //
  // ==========================================================================
  //
  //   Setting configuration words using macros:
  //
  //   The following macro named 'config' can be used to set configuration words:
  //
  //   For example, to set CONFIG_REG using the macro above, the following line
  //   can be pasted only at the beginning of the assembly source code,
  //   immediately below the '.include' directive.
  //
  //         config __CONFIG_REG, SETTING_A & SETTING_B
  //
  //   Note that the 'config' macro takes two arguments: the first is the name
  //   of a configuration word (in this case, __CONFIG_REG), the second argument
  //   is a boolean expression that may include multiple settings.
  //   The example above would enable SETTING_A and also SETTING_B.
  //
  //   A description of all possible settings for each configuration word
  //   appears below.
  //
  // ==========================================================================
  //
  //   Setting configuration words without using macros:
  //
  //   To set configuration words without using the 'config' macro,
  //   use the following format:
  //
  //            .section __CONFIG_REG.sec, code
  //            .global __CONFIG_REG
  // __CONFIG_REG: .pword SETTING_A & SETTING_B
  //
  //   This source code is equivalent to the expanded macro from the previous
  //   example.
  //
  // ==========================================================================
  //
  // ----- CONFIG4 (0x2abf8) --------------------------------------------------
  //
  //   The following settings are available for CONFIG4:
  //
  // ----- CONFIG3 (0x2abfa) --------------------------------------------------
  //
  //   The following settings are available for CONFIG3:
  //
  //    Write Protection Flash Page Segment Boundary:
  //      WPFP_WPFP0           Page 0 (0x00)
  //      WPFP_WPFP1           Page 1 (0x400)
  //      WPFP_WPFP2           Page 2 (0x800)
  //
  //      ...
  //
  //      WPFP_WPFP169         Page 169 (0x2A400)
  //      WPFP_WPFP170         Page 170 (0x2A800)
  //      WPFP_WPFP255         Highest Page (same as page 170)
  //
  //    Secondary Oscillator Power Mode Select:
  //      SOSCSEL_EC           External clock (SCLKI) or Digital I/O mode(
  //      SOSCSEL_LPSOSC       Secondary oscillator is in Low-Power (low drive strength) Oscillator mode
  //      SOSCSEL_SOSC         Secondary oscillator is in Default (high drive strength) Oscillator mode
  //
  //    Voltage Regulator Wake-up Time Select:
  //      WUTSEL_FST           Fast regulator start-up time is used
  //      WUTSEL_LEG           Default regulator start-up time is used
  //
  //    Alternate PMP Pin Mapping:
  //      ALTPMP_ALTPMPEN      EPMP pins are in alternate location mode
  //      ALTPMP_ALPMPDIS      EPMP pins are in default location mode
  //
  //    Segment Write Protection Disable:
  //      WPDIS_WPEN           Segmented code protection is enabled protected segment is defined by the WPEND, WPCFG and WPFPx Configuration bits
  //      WPDIS_WPDIS          Segmented code protection is disabled
  //
  //    Write Protect Configuration Page Select:
  //      WPCFG_WPCFGEN        Last page and Flash Configuration Words are write-protected, provided WPDIS = 0
  //      WPCFG_WPCFGDIS       Last page (at the top of program memory) and Flash Configuration Words are not write-protected
  //
  //    Segment Write Protection End Page Select:
  //      WPEND_WPSTARTMEM     Protected code segment lower boundary is at the bottom of the program memory (000000h) upper boundary is the code page specified by WPFP
  //      WPEND_WPENDMEM       Protected code segment upper boundary is at the last page of program memory the lower boundary is the code page specified by WPFP
  //

  config  CONFIG4 <WPFP_WPFP255 & SOSCSEL_EC & WUTSEL_LEG & ALTPMP_ALPMPDIS & WPDIS_WPDIS & WPCFG_WPCFGDIS & WPEND_WPENDMEM>

  // ----- CONFIG2 (0x2abfc) --------------------------------------------------
  //
  //   The following settings are available for CONFIG2:
  //
  //    Primary Oscillator Select:
  //      POSCMOD_EC           EC Oscillator mode is selected
  //      POSCMOD_XT           XT Oscillator mode is selected
  //      POSCMOD_HS           HS Oscillator mode is selected
  //      POSCMOD_NONE         Primary oscillator is disabled
  //
  //    IOLOCK One-Way Set Enable:
  //      IOL1WAY_OFF          The IOLOCK bit can be set and cleared as needed, provided the unlock sequence has been completed
  //      IOL1WAY_ON           The IOLOCK bit (OSCCON<6>) can be set once, provided the unlock sequence has been completed. Once set, the Peripheral Pin Select registers cannot be written to a second time.
  //
  //    OSCO Pin Configuration:
  //      OSCIOFNC_ON          OSCO/CLKO/RC15 functions as port I/O (RC15)
  //      OSCIOFNC_OFF         OSCO/CLKO/RC15 functions as CLKO (FOSC/2)
  //
  //    Clock Switching and Fail-Safe Clock Monitor:
  //      FCKSM_CSECME         Clock switching is enabled, Fail-Safe Clock Monitor is enabled
  //      FCKSM_CSECMD         Clock switching is enabled, Fail-Safe Clock Monitor is disabled
  //      FCKSM_CSDCMD         Clock switching and Fail-Safe Clock Monitor are disabled
  //
  //    Initial Oscillator Select:
  //      FNOSC_FRC            Fast RC Oscillator (FRC)
  //      FNOSC_FRCPLL         Fast RC Oscillator with Postscaler and PLL module (FRCPLL)
  //      FNOSC_PRI            Primary Oscillator (XT, HS, EC)
  //      FNOSC_PRIPLL         Primary Oscillator with PLL module (XTPLL, HSPLL, ECPLL)
  //      FNOSC_SOSC           Secondary Oscillator (SOSC)
  //      FNOSC_LPRC           Low-Power RC Oscillator (LPRC)
  //      FNOSC_FRCDIV         Fast RC Oscillator with Postscaler (FRCDIV)
  //
  //    96MHz PLL Startup Select:
  //      PLL96MHZ_OFF         96 MHz PLL is software controlled (can be enabled by setting the PLLEN bit in CLKDIV<5>)
  //      PLL96MHZ_ON          96 MHz PLL is enabled automatically on start-up
  //
  //    96 MHz PLL Prescaler Select:
  //      PLLDIV_NODIV         Oscillator input is used directly (4 MHz input)
  //      PLLDIV_DIV2          Oscillator input is divided by 2 (8 MHz input)
  //      PLLDIV_DIV3          Oscillator input is divided by 3 (12 MHz input)
  //      PLLDIV_DIV4          Oscillator input is divided by 4 (16 MHz input)
  //      PLLDIV_DIV5          Oscillator input is divided by 5 (20 MHz input)
  //      PLLDIV_DIV6          Oscillator input is divided by 6 (24 MHz input)
  //      PLLDIV_DIV8          Oscillator input is divided by 8 (32 MHz input)
  //      PLLDIV_DIV12         Oscillator input is divided by 12 (48 MHz input)
  //
  //    Internal External Switchover:
  //      IESO_OFF             IESO mode (Two-Speed Start-up) is disabled
  //      IESO_ON              IESO mode (Two-Speed Start-up) is enabled
  //

  config  CONFIG2 <POSCMOD_HS & IOL1WAY_ON & OSCIOFNC_OFF & FCKSM_CSDCMD & FNOSC_PRI & PLL96MHZ_ON & PLLDIV_DIV2 & IESO_OFF>

  //
  // ----- CONFIG1 (0x2abfe) --------------------------------------------------
  //
  //   The following settings are available for CONFIG1:
  //
  //    Watchdog Timer Postscaler:
  //      WDTPS_PS1            1:1
  //      WDTPS_PS2            1:2
  //      WDTPS_PS4            1:4
  //      WDTPS_PS8            1:8
  //      WDTPS_PS16           1:16
  //      WDTPS_PS32           1:32
  //      WDTPS_PS64           1:64
  //      WDTPS_PS128          1:128
  //      WDTPS_PS256          1:256
  //      WDTPS_PS512          1:512
  //      WDTPS_PS1024         1:1,024
  //      WDTPS_PS2048         1:2,048
  //      WDTPS_PS4096         1:4,096
  //      WDTPS_PS8192         1:8,192
  //      WDTPS_PS16384        1:16,384
  //      WDTPS_PS32768        1:32,768
  //
  //    WDT Prescaler:
  //      FWPSA_PR32           Prescaler ratio of 1:32
  //      FWPSA_PR128          Prescaler ratio of 1:128
  //
  //    Alternate VREF location Enable:
  //      ALTVREF_ALTVREFEN    VREF is on an alternate pin (VREF+ on RB0 and VREF- on RB1)
  //      ALTVREF_ALTVREDIS    VREF is on a default pin (VREF+ on RA9 and VREF- on RA10)
  //
  //    Windowed WDT:
  //      WINDIS_ON            Windowed Watchdog Timer is enabled FWDTEN must be 1
  //      WINDIS_OFF           Standard Watchdog Timer enabled,(Windowed-mode is disabled)
  //
  //    Watchdog Timer:
  //      FWDTEN_OFF           Watchdog Timer is disabled
  //      FWDTEN_ON            Watchdog Timer is enabled
  //
  //    Emulator Pin Placement Select bits:
  //      ICS_PGx3             Emulator functions are shared with PGEC3/PGED3
  //      ICS_PGx2             Emulator functions are shared with PGEC2/PGED2
  //      ICS_PGx1             Emulator functions are shared with PGEC1/PGED1
  //
  //    General Segment Write Protect:
  //      GWRP_ON              Writes to program memory are not allowed
  //      GWRP_OFF             Writes to program memory are allowed
  //
  //    General Segment Code Protect:
  //      GCP_ON               Code protection is enabled for the entire program memory space
  //      GCP_OFF              Code protection is disabled
  //
  //    JTAG Port Enable:
  //      JTAGEN_OFF           JTAG port is disabled
  //      JTAGEN_ON            JTAG port is enabled
  //

    config  CONFIG1 <WDTPS_PS32768 & FWPSA_PR32 & ALTVREF_ALTVREDIS & WINDIS_OFF & ICS_PGx1 & GWRP_OFF & GCP_OFF & JTAGEN_OFF>


  {
    #f 96e6 constant ProcessorSpeed
  }


  // =======================================================================
  //
  //    Peripheral Register setup
  //
  // =======================================================================


  // =======================================================================
  // PORT pin setup

  portDef B                              // use n output, n i/o, input or (nc)
    0 output  b0                        #//  0 out, not used
    0 output  b1                        #//  1 out, not used
    0 output  b2                        #//  2 out, not used
    0 output  b3                        #//  3 out, not used
    0 output  b4                        #//  4 out, not used
    0 output  b5                        #//  5 out, not used
    0 output  b6                        #//  6 out, not used
    0 output  b7                        #//  7 out, not used
    0 output  b8                        #//  8 out, not used
    0 output  b9                        #//  9 out, not used
    0 output  b10                       #// 10 out, not used
    0 output  b11                       #// 11 out, not used
    0 output  b12                       #// 12 out, not used
    0 output  b13                       #// 13 out, not used
    0 output  b14                       #// 14 out, not used
    0 output  b15                       #// 15 out, not used
  endPortDef

  RegDef ODCB                           // open drain control
    RegBit  ODB0            .off        //
    RegBit  ODB1            .off        //
    RegBit  ODB2            .off        //
    RegBit  ODB3            .off        //
    RegBit  ODB4            .off        //
    RegBit  ODB5            .off        //
    RegBit  ODB6            .off        //
    RegBit  ODB7            .off        //
    RegBit  ODB8            .off        //
    RegBit  ODB9            .off        //
    RegBit  ODB10           .off        //
    RegBit  ODB11           .off        //
    RegBit  ODB12           .off        //
    RegBit  ODB13           .off        //
    RegBit  ODB14           .off        //
    RegBit  ODB15           .off        //
  endRegDef

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  portDef C                              // use n output, n i/o, input or (nc)
      (nc)    c0                        #//  0 NC
      (nc)    c1                        #//  1 NC
      (nc)    c2                        #//  2 NC
      (nc)    c3                        #//  3 NC
      (nc)    c4                        #//  4 NC
      (nc)    c5                        #//  5 NC
      (nc)    c6                        #//  6 NC
      (nc)    c7                        #//  7 NC
      (nc)    c8                        #//  8 NC
      (nc)    c9                        #//  9 NC
      (nc)    c10                       #// 10 NC
      (nc)    c11                       #// 11 NC
    0 output  c12                       #// 12 out, not used
    0 output  c13                       #// 13 out, not used
    0 output  c14                       #// 14 out, not used
    0 output  c15                       #// 15 out, not used
  endPortDef

  RegDef ODCC                           // open drain control
    RegBit  ODC12           .off        //
    RegBit  ODC13           .off        //
    RegBit  ODC14           .off        //
    RegBit  ODC15           .off        //
  endRegDef

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  portDef D                              // use n output, n i/o, input or (nc)
    0 output  d0                        #//  0 out, not used
    0 output  d1                        #//  1 out, not used
    0 output  d2                        #//  2 out, not used
    0 output  d3                        #//  3 out, not used
    0 output  d4                        #//  4 out, not used
    0 output  d5                        #//  5 out, not used
    0 output  d6                        #//  6 out, not used
    0 output  d7                        #//  7 out, not used
    0 output  d8                        #//  8 out, not used
    0 output  d9                        #//  9 out, not used
    0 output  d10                       #// 10 out, not used
    0 output  d11                       #// 11 out, not used
  endPortDef

  RegDef ODCD                           // open drain control
    RegBit  ODD0            .off        //
    RegBit  ODD1            .off        //
    RegBit  ODD2            .off        //
    RegBit  ODD3            .off        //
    RegBit  ODD4            .off        //
    RegBit  ODD5            .off        //
    RegBit  ODD6            .off        //
    RegBit  ODD7            .off        //
    RegBit  ODD8            .off        //
    RegBit  ODD9            .off        //
    RegBit  ODD10           .off        //
    RegBit  ODD11           .off        //
  endRegDef

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  portDef E                              // use n output, n i/o, input or (nc)
    0 output  e0                        #//  0 out, not used
    0 output  e1                        #//  1 out, not used
    0 output  e2                        #//  2 out, not used
    0 output  e3                        #//  3 out, not used
    0 output  e4                        #//  4 out, not used
    0 output  e5                        #//  5 out, not used
    0 output  e6                        #//  6 out, not used
    0 output  e7                        #//  7 out, not used
  endPortDef

  RegDef ODCE                           // open drain control
    RegBit  ODE0            .off        //
    RegBit  ODE1            .off        //
    RegBit  ODE2            .off        //
    RegBit  ODE3            .off        //
    RegBit  ODE4            .off        //
    RegBit  ODE5            .off        //
    RegBit  ODE6            .off        //
    RegBit  ODE7            .off        //
  endRegDef

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  portDef F                              // use n output, n i/o, input or (nc)
    0 output  f0                        #//  0 out, not used
    0 output  f1                        #//  1 out, not used
        (nc)  f2                        #//  2 NC
    0 output  f3                        #//  3 out, not used
    0 output  f4                        #//  4 out, not used
    0 output  f5                        #//  5 out, not used
        (nc)  f6                        #//  6 NC
    0 output  f7                        #//  7 out, not used
  endPortDef

  RegDef ODCF                           // open drain control
    RegBit  ODF0            .off        //
    RegBit  ODF1            .off        //

    RegBit  ODF3            .off        //
    RegBit  ODF4            .off        //
    RegBit  ODF5            .off        //

    RegBit  ODF7            .off        //
  endRegDef

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  portDef G                              // use n output, n i/o, input or (nc)
      (nc)    g0                        #//  0 NC
      (nc)    g1                        #//  1 NC
    0 output  g2                        #//  2 out, not used
    0 output  g3                        #//  3 out, not used
      (nc)    g4                        #//  4 NC
      (nc)    g5                        #//  5 NC
    0 output  g6                        #//  6 out, not used
    0 output  g7                        #//  7 out, not used
    0 output  g8                        #//  8 out, not used
    0 output  g9                        #//  9 out, not used
  endPortDef

  RegDef ODCG                           // open drain control
    RegBit  ODG2            .off        //
    RegBit  ODG3            .off        //

    RegBit  ODG6            .off        //
    RegBit  ODG7            .off        //
    RegBit  ODG8            .off        //
    RegBit  ODG9            .off        //
  endRegDef



  // ///////////////////////////////////////////////////////////////////////
  //
  // Pin routing
  //
  // Output mapping, edit the second table, do not touch this one
  //
  //           function                    description

    0 constant RPOR_NULL                // NULL - disable output function.
    1 constant RPOR_C1OUT               // Comparator 1 Output.
    2 constant RPOR_C2OUT               // Comparator 2 Output.
    3 constant RPOR_U1TX                // UART1 Transmit.
    4 constant RPOR_U1RTS               // UART1 Ready To Send.
    5 constant RPOR_U2TX                // UART2 Transmit.
    6 constant RPOR_U2RTS               // UART2 Ready To Send.
    7 constant RPOR_SDO1                // SPI1 Data Output.
    8 constant RPOR_SCK1OUT             // SPI1 Clock Output.
    9 constant RPOR_SS1OUT              // SPI1 Slave Select Output.
   10 constant RPOR_SDO2                // SPI2 Data Output.
   11 constant RPOR_SCK2OUT             // SPI2 Clock Output.
   12 constant RPOR_SS2OUT              // SPI2 Slave Select Output.
   18 constant RPOR_OC1                 // Output Compare 1.
   19 constant RPOR_OC2                 // Output Compare 2.
   20 constant RPOR_OC3                 // Output Compare 3.
   21 constant RPOR_OC4                 // Output Compare 4.
   22 constant RPOR_OC5                 // Output Compare 5.


  //  Edit the table below, don't touch the one above, use RPOR_NULL to not set a mapping
  //  or use other values from the table above to map a function to an RP out (RPO)
  //
  // functions                                        register                  // RP number

    RPOR_NULL       RPOR_NULL       8 << or constant  RPOR0_IniVal              // Pins   0   1
    RPOR_NULL       RPOR_NULL       8 << or constant  RPOR1_IniVal              // Pins   2   3
    RPOR_NULL                               constant  RPOR2_IniVal              // Pins   4
    RPOR_NULL       RPOR_NULL       8 << or constant  RPOR3_IniVal              // Pins   6   7
    RPOR_NULL       RPOR_NULL       8 << or constant  RPOR4_IniVal              // Pins   8   9
    RPOR_NULL       RPOR_NULL       8 << or constant  RPOR5_IniVal              // Pins  10  11
    RPOR_NULL       RPOR_NULL       8 << or constant  RPOR6_IniVal              // Pins  12  13
    RPOR_NULL                               constant  RPOR7_IniVal              // Pins  14
    RPOR_NULL       RPOR_NULL       8 << or constant  RPOR8_IniVal              // Pins  16  17
    RPOR_NULL       RPOR_NULL       8 << or constant  RPOR9_IniVal              // Pins  18  19
    RPOR_NULL       RPOR_NULL       8 << or constant  RPOR10_IniVal             // Pins  20  21
    RPOR_NULL       RPOR_NULL       8 << or constant  RPOR11_IniVal             // Pins  22  23
    RPOR_NULL       RPOR_NULL       8 << or constant  RPOR12_IniVal             // Pins  24  25
    RPOR_NULL       RPOR_NULL       8 << or constant  RPOR13_IniVal             // Pins  26  27
    RPOR_NULL       RPOR_NULL       8 << or constant  RPOR14_IniVal             // Pins  28  29


  // ///////////////////////////////////////////////////////////////////////
  //
  // Input mapping, edit the first table (select 31 to not use a mapping)
  //
  // #hysical_pin functional_name       // description

    31 constant RPIN_INT1               // External Interrupt 1
    31 constant RPIN_INT2               // External Interrupt 2
    31 constant RPIN_INT3               // External Interrupt 3
    31 constant RPIN_INT4               // External Interrupt 4
    31 constant RPIN_T2CK               // Timer2 External Clock
    31 constant RPIN_T3CK               // Timer3 External Clock
    31 constant RPIN_T4CK               // Timer4 External Clock
    31 constant RPIN_T5CK               // Timer5 External Clock
    31 constant RPIN_IC1                // Input Capture 1
    31 constant RPIN_IC2                // Input Capture 2
    31 constant RPIN_IC3                // Input Capture 3
    31 constant RPIN_IC4                // Input Capture 4
    31 constant RPIN_IC5                // Input Capture 5
    31 constant RPIN_IC6                // Input Capture 6
    31 constant RPIN_IC7                // Input Capture 7
    31 constant RPIN_IC8                // Input Capture 8
    31 constant RPIN_OCFA               // Output Compare Fault A
    31 constant RPIN_OCFB               // Output Compare Fault B
    31 constant RPIN_IC9                // Input Capture 9
    31 constant RPIN_U3RX               // UART3 Receive
    31 constant RPIN_U1RX               // UART1 Receive
    31 constant RPIN_U1CTS              // UART1 Clear To Send
    31 constant RPIN_U2RX               // UART2 Receive
    31 constant RPIN_U2CTS              // UART2 Clear To Send
    31 constant RPIN_SDI1               // SPI1 Data Input
    31 constant RPIN_SCK1               // SPI1 Clock Input
    31 constant RPIN_SS1                // SPI1 Slave Select Input
    31 constant RPIN_U3CTS              // UART3 Clear To Send
    31 constant RPIN_SDI2               // SPI2 Data Input
    31 constant RPIN_SCK2               // SPI2 Clock Input
    31 constant RPIN_SS2                // SPI2 Slave Select Input
    31 constant RPIN_U4RX               // UART4 Receive
    31 constant RPIN_U4CTS              // UART4 Clear To Send
    31 constant RPIN_SDI3               // SPI3 Data Input
    31 constant RPIN_SCK3               // SPI3 Clock Input
    31 constant RPIN_SS3                // SPI3 Slave Select Input


  //  Edit the table above, don't touch the table below
  //
  // functional   name                              register

                  RPIN_INT1     8 <<      constant  RPINR0_IniVal
    RPIN_INT2     RPIN_INT3     8 <<  or  constant  RPINR1_IniVal
    RPIN_INT4                             constant  RPINR2_IniVal
    RPIN_T2CK     RPIN_T3CK     8 <<  or  constant  RPINR3_IniVal
    RPIN_T4CK     RPIN_T5CK     8 <<  or  constant  RPINR4_IniVal

    RPIN_IC1      RPIN_IC2      8 <<  or  constant  RPINR7_IniVal
    RPIN_IC3      RPIN_IC4      8 <<  or  constant  RPINR8_IniVal
    RPIN_IC5      RPIN_IC6      8 <<  or  constant  RPINR9_IniVal
    RPIN_IC7      RPIN_IC8      8 <<  or  constant  RPINR10_IniVal
    RPIN_OCFA     RPIN_OCFB     8 <<  or  constant  RPINR11_IniVal

                  RPIN_IC9      8 <<      constant  RPINR15_IniVal

                  RPIN_U3RX     8 <<      constant  RPINR17_IniVal
    RPIN_U1RX     RPIN_U1CTS    8 <<  or  constant  RPINR18_IniVal
    RPIN_U2RX     RPIN_U2CTS    8 <<  or  constant  RPINR19_IniVal
    RPIN_SDI1     RPIN_SCK1     8 <<  or  constant  RPINR20_IniVal
    RPIN_SS1      RPIN_U3CTS    8 <<  or  constant  RPINR21_IniVal
    RPIN_SDI2     RPIN_SCK2     8 <<  or  constant  RPINR22_IniVal
    RPIN_SS2                              constant  RPINR23_IniVal

    RPIN_U4RX     RPIN_U4CTS    8 <<  or  constant  RPINR27_IniVal
    RPIN_SDI3     RPIN_SCK3     8 <<  or  constant  RPINR28_IniVal
    RPIN_SS3                              constant  RPINR29_IniVal





  // =======================================================================
  // =======================================================================
  // SR - status register
  //          Also see : "29-interrupts.pdf"
  //
  // bit 15-8 Not used by Interrupt Controller.
  // (See the .dsPIC30F/33F Programmer.s Reference Manual. (DS70157) for descriptions of SR bits)
  //
  // bit 7-5 IPL<2:0>: CPU Interrupt Priority Level Status bits(1,2)
  //   111 = CPU Interrupt Priority Level is 7 (15), user interrupts disabled
  //   110 = CPU Interrupt Priority Level is 6 (14)
  //   101 = CPU Interrupt Priority Level is 5 (13)
  //   100 = CPU Interrupt Priority Level is 4 (12)
  //   011 = CPU Interrupt Priority Level is 3 (11)
  //   010 = CPU Interrupt Priority Level is 2 (10)
  //   001 = CPU Interrupt Priority Level is 1 (9)
  //   000 = CPU Interrupt Priority Level is 0 (8)
  //
  // bit 4-0 Not used by Interrupt Controller.
  // (See the .dsPIC30F/33F Programmer.s Reference Manual. (DS70157) for descriptions of SR bits)
  //
  // Note 1: The IPL<2:0> bits are concatenated with the IPL<3> bit (CORCON<3>) to form the CPU Interrupt Priority
  //         Level. The value in parentheses indicates the IPL if IPL<3> = 1. User interrupts are disabled when
  //         IPL<3> = 1.
  //
  // Note 2: The IPL<2:0> Status bits are read-only when NSTDIS (INTCON1<15>) = 1.

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  RegDef SR                                     // use .on or .off or (for vectors only) .use
    RegBit  (C)             .off                // Initial carry    flag
    RegBit  Z               .off                // Initial zero     flag
    RegBit  OV              .off                // Initial overflow flag
    RegBit  N               .off                // Initial negative flag
    RegBit  RA              .off                // Initial repeat inactive

    RegBit  (DC)            .off                // Initial half carry off

    RegBit  IPL0            .off                // Startup processor priority
    RegBit  IPL1            .off                // Startup processor priority
    RegBit  IPL2            .off                // Startup processor priority
  endRegDef


  // =======================================================================
  // CORCON -
  //          Also see : "29-interrupts.pdf"
  //
  // bit 15-4 Not used by the Interrupt Controller
  // (See .dsPIC30F/33F Programmer.s Reference Manual (DS70157). for descriptions of CORCON bits)
  //
  // bit 3 IPL3: CPU Interrupt Priority Level Status bit 3
  //   1 = CPU interrupt priority level is greater than 7
  //   0 = CPU interrupt priority level is 7 or less
  //
  //   Note: The IPL3 bit is concatenated with the IPL<2:0> bits (SR<7:5>) to form the CPU interrupt
  //   priority level.

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef CORCON
    RegBit  IPL3            .off                //
  endRegDef


  // ----- CNPD1 Bits -----------------------------------------------------
  regDef CNPD1
    RegBit  CN0PDE          .off                //
    RegBit  CN1PDE          .off                //
    RegBit  CN2PDE          .off                //
    RegBit  CN3PDE          .off                //
    RegBit  CN4PDE          .off                //
    RegBit  CN5PDE          .off                //
    RegBit  CN6PDE          .off                //
    RegBit  CN7PDE          .off                //
    RegBit  CN8PDE          .off                //
    RegBit  CN9PDE          .off                //
    RegBit  CN10PDE         .off                //
    RegBit  CN11PDE         .off                //
    RegBit  CN12PDE         .off                //
    RegBit  CN13PDE         .off                //
    RegBit  CN14PDE         .off                //
    RegBit  CN15PDE         .off                //
  endRegDef

  // ----- CNPD2 Bits -----------------------------------------------------
  regDef CNPD2
    RegBit  CN16PDE         .off                //
    RegBit  CN17PDE         .off                //
    RegBit  CN18PDE         .off                //
    RegBit  CN22PDE         .off                //
    RegBit  CN23PDE         .off                //
    RegBit  CN24PDE         .off                //
    RegBit  CN25PDE         .off                //
    RegBit  CN26PDE         .off                //
    RegBit  CN27PDE         .off                //
    RegBit  CN28PDE         .off                //
    RegBit  CN29PDE         .off                //
    RegBit  CN30PDE         .off                //
    RegBit  CN31PDE         .off                //
  endRegDef

  // ----- CNPD3 Bits -----------------------------------------------------
  regDef CNPD3
    RegBit  CN32PDE         .off                //
  endRegDef

  // ----- CNPD4 Bits -----------------------------------------------------
  regDef CNPD4
    RegBit  CN49PDE         .off                //
    RegBit  CN50PDE         .off                //
    RegBit  CN51PDE         .off                //
    RegBit  CN52PDE         .off                //
    RegBit  CN53PDE         .off                //
    RegBit  CN54PDE         .off                //
    RegBit  CN55PDE         .off                //
    RegBit  CN56PDE         .off                //
    RegBit  CN58PDE         .off                //
    RegBit  CN59PDE         .off                //
    RegBit  CN60PDE         .off                //
    RegBit  CN61PDE         .off                //
    RegBit  CN62PDE         .off                //
    RegBit  CN63PDE         .off                //
  endRegDef

  // ----- CNPD5 Bits -----------------------------------------------------
  regDef CNPD5
    RegBit  CN64PDE         .off                //
    RegBit  CN65PDE         .off                //
    RegBit  CN68PDE         .off                //
    RegBit  CN69PDE         .off                //
    RegBit  CN71PDE         .off                //
  endRegDef

  // ----- CNPD6 Bits -----------------------------------------------------
  regDef CNPD6
    RegBit  CN83PDE         .off                //
    RegBit  CN84PDE         .off                //
  endRegDef


  // =======================================================================
  // CNEN1 - Change Notification Enable register 1

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  RegDef CNEN1
    RegBit  CN0IE           .off                // Disable all pin change notifications
    RegBit  CN1IE           .off                //
    RegBit  CN2IE           .off                //
    RegBit  CN3IE           .off                //
    RegBit  CN4IE           .off                //
    RegBit  CN5IE           .off                //
    RegBit  CN6IE           .off                //
    RegBit  CN7IE           .off                //
    RegBit  CN8IE           .off                //
    RegBit  CN9IE           .off                //
    RegBit  CN10IE          .off                //
    RegBit  CN11IE          .off                //
    RegBit  CN12IE          .off                //
    RegBit  CN13IE          .off                //
    RegBit  CN14IE          .off                //
    RegBit  CN15IE          .off                //
  endRegDef


  // =======================================================================
  // CNEN2 - Change Notification Enable register 2

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  RegDef CNEN2
    RegBit  CN16IE          .off                // Disable all pin change notifications
    RegBit  CN17IE          .off                //
    RegBit  CN18IE          .off                //

    RegBit  CN22IE          .off                //
    RegBit  CN23IE          .off                //
    RegBit  CN24IE          .off                //
    RegBit  CN25IE          .off                //
    RegBit  CN26IE          .off                //
    RegBit  CN27IE          .off                //
    RegBit  CN28IE          .off                //
    RegBit  CN29IE          .off                //
    RegBit  CN30IE          .off                //
    RegBit  CN31IE          .off                //
  endRegDef


  // =======================================================================
  // CNEN3 - Change Notification Enable register 3

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  RegDef CNEN3
    RegBit  CN32IE          .off                // Disable all pin change notifications
  endRegDef


  // =======================================================================
  // CNEN4 - Change Notification Enable register 4

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  RegDef CNEN4
    RegBit  CN49IE          .off                // Disable all pin change notifications
    RegBit  CN50IE          .off                //
    RegBit  CN51IE          .off                //
    RegBit  CN52IE          .off                //
    RegBit  CN53IE          .off                //
    RegBit  CN54IE          .off                //
    RegBit  CN55IE          .off                //
    RegBit  CN56IE          .off                //

    RegBit  CN58IE          .off                //
    RegBit  CN59IE          .off                //
    RegBit  CN60IE          .off                //
    RegBit  CN61IE          .off                //
    RegBit  CN62IE          .off                //
    RegBit  CN63IE          .off                //
  endRegDef


  // =======================================================================
  // CNEN5 - Change Notification Enable register 5

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  RegDef CNEN5
    RegBit  CN64IE          .off                // Disable all pin change notifications
    RegBit  CN65IE          .off                //

    RegBit  CN68IE          .off                //
    RegBit  CN69IE          .off                //

    RegBit  CN71IE          .off                //
  endRegDef


  // =======================================================================
  // CNEN6 - Change Notification Enable register 6

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  RegDef CNEN6
    RegBit  CN83IE          .off                // Disable all pin change notifications
    RegBit  CN84IE          .off                //
  endRegDef


  // =======================================================================
  // CNPU1 - Change Notification Pull=up Enable register 1

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  RegDef CNPU1
    RegBit  CN0PUE          .off                // Disable all pull-ups
    RegBit  CN1PUE          .off                //
    RegBit  CN2PUE          .off                //
    RegBit  CN3PUE          .off                //
    RegBit  CN4PUE          .off                //
    RegBit  CN5PUE          .off                //
    RegBit  CN6PUE          .off                //
    RegBit  CN7PUE          .off                //
    RegBit  CN8PUE          .off                //
    RegBit  CN9PUE          .off                //
    RegBit  CN10PUE         .off                //
    RegBit  CN11PUE         .off                //
    RegBit  CN12PUE         .off                //
    RegBit  CN13PUE         .off                //
    RegBit  CN14PUE         .off                //
    RegBit  CN15PUE         .off                //
  endRegDef


  // =======================================================================
  // CNPU2 - Change Notification Pull=up Enable register 2

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  RegDef CNPU2
    RegBit  CN16PUE         .off                // Disable all pull-ups
    RegBit  CN17PUE         .off                //
    RegBit  CN18PUE         .off                //

    RegBit  CN22PUE         .off                //
    RegBit  CN23PUE         .off                //
    RegBit  CN24PUE         .off                //
    RegBit  CN25PUE         .off                //
    RegBit  CN26PUE         .off                //
    RegBit  CN27PUE         .off                //
    RegBit  CN28PUE         .off                //
    RegBit  CN29PUE         .off                //
    RegBit  CN30PUE         .off                //
    RegBit  CN31PUE         .off                //
  endRegDef


  // =======================================================================
  // CNPU3 - Change Notification Pull=up Enable register 3

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  RegDef CNPU3
    RegBit  CN32PUE         .off                // Disable all pull-ups
  endRegDef


  // =======================================================================
  // CNPU4 - Change Notification Pull=up Enable register 4

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  RegDef CNPU4
    RegBit  CN49PUE         .off                // Disable all pull-ups
    RegBit  CN50PUE         .off                //
    RegBit  CN51PUE         .off                //
    RegBit  CN52PUE         .off                //
    RegBit  CN53PUE         .off                //
    RegBit  CN54PUE         .off                //
    RegBit  CN55PUE         .off                //
    RegBit  CN56PUE         .off                //

    RegBit  CN58PUE         .off                //
    RegBit  CN59PUE         .off                //
    RegBit  CN60PUE         .off                //
    RegBit  CN61PUE         .off                //
    RegBit  CN62PUE         .off                //
    RegBit  CN63PUE         .off                //
  endRegDef


  // =======================================================================
  // CNPU5 - Change Notification Pull=up Enable register 5

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  RegDef CNPU5
    RegBit  CN64PUE         .off                // Disable all pull-ups
    RegBit  CN65PUE         .off                //

    RegBit  CN68PUE         .off                //
    RegBit  CN69PUE         .off                //

    RegBit  CN71PUE         .off                //
  endRegDef


  // =======================================================================
  // CNPU6 - Change Notification Pull=up Enable register 6

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  RegDef CNPU6
    RegBit  CN83PUE         .off                // Disable all pull-ups
    RegBit  CN84PUE         .off                //
  endRegDef


  // =======================================================================
  // INTCON1 -
  //          Also see : "29-interrupts.pdf"
  //
  // bit 15 NSTDIS: Interrupt Nesting Disable bit
  //   1 = Interrupt nesting is disabled
  //   0 = Interrupt nesting is enabled
  //
  // bit 14-7 Unimplemented: Read as .0.
  //
  // bit 5 Unimplemented: Read as .0.
  //
  // bit 4 MATHERR: Arithmetic Error Status bit
  //   1 = Math error trap has occurred
  //   0 = Math error trap has not occurred
  //
  // bit 3 ADDRERR: Address Error Trap Status bit
  //   1 = Address error trap has occurred
  //   0 = Address error trap has not occurred
  //
  // bit 2 STKERR: Stack Error Trap Status bit
  //   1 = Stack error trap has occurred
  //   0 = Stack error trap has not occurred
  //
  // bit 1 OSCFAIL: Oscillator Failure Trap Status bit
  //   1 = Oscillator failure trap has occurred
  //   0 = Oscillator failure trap has not occurred
  //
  // bit 0 Unimplemented: Read as .0.


  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef INTCON1
    RegBit  OSCFAIL         .off                //
    RegBit  STKERR          .off                //
    RegBit  ADDRERR         .off                //
    RegBit  MATHERR         .off                //
    RegBit  NSTDIS          .off                //
  endRegDef


  // =======================================================================
  // INTCON2 -
  //          Also see : "29-interrupts.pdf"
  //
  // bit 15 ALTIVT: Enable Alternate Interrupt Vector Table bit
  //   1 = Use alternate vector table
  //   0 = Use standard (default) vector table
  //
  // bit 14 DISI: DISI (disable interrupts) Instruction Status bit
  //   1 = DISI instruction is active
  //   0 = DISI instruction is not active
  //
  // bit 13-3 Unimplemented: Read as .0.
  //
  // bit 2 INT2EP: External Interrupt 2 Edge Detect Polarity Select bit
  //   1 = Interrupt on negative edge
  //   0 = Interrupt on positive edge
  //
  // bit 1 INT1EP: External Interrupt 1 Edge Detect Polarity Select bit
  //   1 = Interrupt on negative edge
  //   0 = Interrupt on positive edge
  //
  // bit 0 INT0EP: External Interrupt 0 Edge Detect Polarity Select bit
  //   1 = Interrupt on negative edge
  //   0 = Interrupt on positive edge


  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef INTCON2
    RegBit  INT0EP          .off                //
    RegBit  INT1EP          .off                //
    RegBit  INT2EP          .off                //
    RegBit  INT3EP          .off                //
    RegBit  INT4EP          .off                //
    RegBit  DISI            .off                //
    RegBit  ALTIVT          .off                //
  endRegDef


  // =======================================================================
  // IFS0 - interrupt flag status 0
  //          Also see : "29-interrupts.pdf"
  //
  // bit 15-14 Unimplemented: Read as .0.
  //
  // bit 13 AD1IF: ADC1 Conversion Complete Interrupt Flag Status bit
  //   1 = Interrupt request has occurred
  //   0 = Interrupt request has not occurred
  //
  // bit 12 U1TXIF: UART1 Transmitter Interrupt Flag Status bit
  //   1 = Interrupt request has occurred
  //   0 = Interrupt request has not occurred
  //
  // bit 11 U1RXIF: UART1 Receiver Interrupt Flag Status bit
  //   1 = Interrupt request has occurred
  //   0 = Interrupt request has not occurred
  //
  // bit 10 SPI1IF: SPI1 Event Interrupt Flag Status bit
  //   1 = Interrupt request has occurred
  //   0 = Interrupt request has not occurred
  //
  // bit 9 SPI1EIF: SPI1 Fault Interrupt Flag Status bit
  //   1 = Interrupt request has occurred
  //   0 = Interrupt request has not occurred
  //
  // bit 8 T3IF: Timer3 Interrupt Flag Status bit
  //   1 = Interrupt request has occurred
  //   0 = Interrupt request has not occurred
  //
  // bit 7 T2IF: Timer2 Interrupt Flag Status bit
  //   1 = Interrupt request has occurred
  //   0 = Interrupt request has not occurred
  //
  // bit 6 OC2IF: Output Compare Channel 2 Interrupt Flag Status bit
  //   1 = Interrupt request has occurred
  //   0 = Interrupt request has not occurred
  //
  // bit 5 IC2IF: Input Capture Channel 2 Interrupt Flag Status bit
  //   1 = Interrupt request has occurred
  //   0 = Interrupt request has not occurred
  //
  // bit 4 Unimplemented: Read as .0.
  //
  // bit 3 T1IF: Timer1 Interrupt Flag Status bit
  //   1 = Interrupt request has occurred
  //   0 = Interrupt request has not occurred
  //
  // bit 2 OC1IF: Output Compare Channel 1 Interrupt Flag Status bit
  //   1 = Interrupt request has occurred
  //   0 = Interrupt request has not occurred
  //
  // bit 1 IC1IF: Input Capture Channel 1 Interrupt Flag Status bit
  //   1 = Interrupt request has occurred
  //   0 = Interrupt request has not occurred
  //
  // bit 0 INT0IF: External Interrupt 0 Flag Status bit
  //   1 = Interrupt request has occurred
  //   0 = Interrupt request has not occurred

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IFS0
    RegBit  INT0IF          .off                // int 0 interrupt flag
    RegBit  IC1IF           .off                //
    RegBit  OC1IF           .off                //
    RegBit  T1IF            .off                // timer 1 intgerrupt flag
    RegBit  IC2IF           .off                //
    RegBit  OC2IF           .off                //
    RegBit  T2IF            .off                // timer 2 intgerrupt flag
    RegBit  T3IF            .off                // timer 3 intgerrupt flag
    RegBit  SPF1IF          .off                //
    RegBit  SPI1IF          .off                //
    RegBit  U1RXIF          .off                // uart 1 rx interrupt flag
    RegBit  U1TXIF          .off                // uart 2 tx interrupt flag
    RegBit  AD1IF           .off                //
  endRegDef


  // =======================================================================
  // IFS1 - interrupt flag status 1
  //          Also see : "29-interrupts.pdf"
  //
  // bit 15-14 Unimplemented: Read as .0.
  //
  // bit 13 INT2IF: External Interrupt 2 Flag Status bit
  //   1 = Interrupt request has occurred
  //   0 = Interrupt request has not occurred
  //
  // bit 12-8 Unimplemented: Read as .0.
  //
  // bit 7 IC8IF: Input Capture Channel 8 Interrupt Flag Status bit
  //   1 = Interrupt request has occurred
  //   0 = Interrupt request has not occurred
  //
  // bit 6 IC7IF: Input Capture Channel 7 Interrupt Flag Status bit
  //   1 = Interrupt request has occurred
  //   0 = Interrupt request has not occurred
  //
  // bit 5 Unimplemented: Read as .0.
  //
  // bit 4 INT1IF: External Interrupt 1 Flag Status bit
  //   1 = Interrupt request has occurred
  //   0 = Interrupt request has not occurred
  //
  // bit 3 CNIF: Input Change Notification Interrupt Flag Status bit
  //   1 = Interrupt request has occurred
  //   0 = Interrupt request has not occurred
  //
  // bit 2 CMIF
  //   1 = Interrupt request has occurred
  //   0 = Interrupt request has not occurred
  //
  // bit 1 MI2C1IF: I2C1 Master Events Interrupt Flag Status bit
  //   1 = Interrupt request has occurred
  //   0 = Interrupt request has not occurred
  //
  // bit 0 SI2C1IF: I2C1 Slave Events Interrupt Flag Status bit
  //   1 = Interrupt request has occurred
  //   0 = Interrupt request has not occurred

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IFS1
    RegBit  SI2C1IF         .off                //
    RegBit  MI2C1IF         .off                //
    RegBit  CMIF            .off                //
    RegBit  CNIF            .off                //
    RegBit  INT1IF          .off                //
    RegBit  IC7IF           .off                //
    RegBit  IC8IF           .off                //
    RegBit  OC3IF           .off                //
    RegBit  OC4IF           .off                //
    RegBit  T4IF            .off                //
    RegBit  T5IF            .off                //
    RegBit  INT2IF          .off                //
    RegBit  U2RXIF          .off                //
    RegBit  U2TXIF          .off                //
  endRegDef


  // =======================================================================
  // IFS2 - interrupt flag status 2
  //          Also see : "29-interrupts.pdf"
  //

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IFS2
    RegBit  SPI2IF          .off                //
    RegBit  SPI2IF          .off                //
    RegBit  IC3IF           .off                //
    RegBit  IC4IF           .off                //
    RegBit  IC5IF           .off                //
    RegBit  IC6IF           .off                //
    RegBit  OC5IF           .off                //
    RegBit  OC6IF           .off                //
    RegBit  OC7IF           .off                //
    RegBit  OC8IF           .off                //
  endRegDef


  // =======================================================================
  // IFS3 - interrupt flag status 3
  //          Also see : "29-interrupts.pdf"
  //

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IFS3
    RegBit  SI2C2IF         .off                //
    RegBit  MI2C2IF         .off                //
    RegBit  INT3IF          .off                //
    RegBit  INT4IF          .off                //
    RegBit  RTCIF           .off                //
  endRegDef


  // =======================================================================
  // IFS4 - interrupt flag status 4
  //          Also see : "29-interrupts.pdf"
  //

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IFS4
    RegBit  U1ERIF          .off                // uart 1 error interrupt flag
    RegBit  U2ERIF          .off                //
    RegBit  CRCIF           .off                //
    RegBit  LVDIF           .off                //
    RegBit  CTMUIF          .off                //
  endRegDef


  // =======================================================================
  // IFS5 - interrupt flag status 5
  //          Also see : "29-interrupts.pdf"
  //

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IFS5
    RegBit  U3ERIF          .off                //
    RegBit  U3RXIF          .off                //
    RegBit  U3TXIF          .off                //
    RegBit  SI2C3IF         .off                //
    RegBit  MI2C3IF         .off                //
    RegBit  USB1IF          .off                //
    RegBit  U4ERIF          .off                //
    RegBit  U4RXIF          .off                //
    RegBit  U4TXIF          .off                //
    RegBit  SPF3IF          .off                //
    RegBit  SPI3IF          .off                //
    RegBit  OC9IF           .off                //
    RegBit  IC9IF           .off                //
  endRegDef


  // =======================================================================
  // IFS6 - interrupt flag status 6
  //          Also see : "29-interrupts.pdf"
  //

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IFS6
    RegBit  GFX1IF          .off                //
  endRegDef


  // =======================================================================
  // IEC0 - interrupt enable control 0
  //          Also see : "29-interrupts.pdf"
  //
  // bit 15-14 Unimplemented: Read as .0.
  //
  // bit 13 AD1IE: ADC1 Conversion Complete Interrupt Enable bit
  //   1 = Interrupt request enabled
  //   0 = Interrupt request not enabled
  //
  // bit 12 U1TXIE: UART1 Transmitter Interrupt Enable bit
  //   1 = Interrupt request enabled
  //   0 = Interrupt request not enabled
  //
  // bit 11 U1RXIE: UART1 Receiver Interrupt Enable bit
  //   1 = Interrupt request enabled
  //   0 = Interrupt request not enabled
  //
  // bit 10 SPI1IE: SPI1 Event Interrupt Enable bit
  //   1 = Interrupt request enabled
  //   0 = Interrupt request not enabled
  //
  // bit 9 SPI1EIE: SPI1 Event Interrupt Enable bit
  //   1 = Interrupt request enabled
  //   0 = Interrupt request not enabled
  //
  // bit 8 T3IE: Timer3 Interrupt Enable bit
  //   1 = Interrupt request enabled
  //   0 = Interrupt request not enabled
  //
  // bit 7 T2IE: Timer2 Interrupt Enable bit
  //   1 = Interrupt request enabled
  //   0 = Interrupt request not enabled
  //
  // bit 6 OC2IE: Output Compare Channel 2 Interrupt Enable bit
  //   1 = Interrupt request enabled
  //   0 = Interrupt request not enabled
  //
  // bit 5 IC2IE: Input Capture Channel 2 Interrupt Enable bit
  //   1 = Interrupt request enabled
  //   0 = Interrupt request not enabled
  //
  // bit 4 Unimplemented: Read as .0.
  //
  // bit 3 T1IE: Timer1 Interrupt Enable bit
  //   1 = Interrupt request enabled
  //   0 = Interrupt request not enabled
  //
  // bit 2 OC1IE: Output Compare Channel 1 Interrupt Enable bit
  //   1 = Interrupt request enabled
  //   0 = Interrupt request not enabled
  //
  // bit 1 IC1IE: Input Capture Channel 1 Interrupt Enable bit
  //   1 = Interrupt request enabled
  //   0 = Interrupt request not enabled
  //
  // bit 0 INT0IE: External Interrupt 0 Enable bit
  //   1 = Interrupt request enabled
  //   0 = Interrupt request not enabled

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IEC0
    RegBit  INT0IE          .off                //
    RegBit  IC1IE           .off                //
    RegBit  OC1IE           .off                //
    RegBit  T1IE                  .on           // timer 1 interrupt enable
    RegBit  IC2IE           .off                //
    RegBit  OC2IE           .off                //
    RegBit  T2IE            .off                // timer 2 interrupt enable
    RegBit  T3IE            .off                // timer 3 interrupt enable
    RegBit  SPF1IE          .off                //
    RegBit  SPI1IE                .on           // SPI1       interrupt enable
    RegBit  U1RXIE          .off                //
    RegBit  U1TXIE          .off                //
    RegBit  AD1IE           .off                //
  endRegDef


  // =======================================================================
  // IEC1 - inrterrupt enaqble control 1
  //          Also see : "29-interrupts.pdf"
  //
  // bit 15-14 Unimplemented: Read as .0.
  //
  // bit 13 INT2IE: External Interrupt 2 Enable bit
  //   1 = Interrupt request enabled
  //   0 = Interrupt request not enabled
  //
  // bit 12-8 Unimplemented: Read as .0.
  //
  // bit 7 IC8IE: Input Capture Channel 8 Interrupt Enable bit
  //   1 = Interrupt request enabled
  //   0 = Interrupt request not enabled
  //
  // bit 6 IC7IE: Input Capture Channel 7 Interrupt Enable bit
  //   1 = Interrupt request enabled
  //   0 = Interrupt request not enabled
  //
  // bit 5 Unimplemented: Read as .0.
  //
  // bit 4 INT1IE: External Interrupt 1 Enab2le bit
  //   1 = Interrupt request enabled
  //   0 = Interrupt request not enabled
  //
  // bit 3 CNIE: Input Change Notification Interrupt Enable bit
  //   1 = Interrupt request enabled
  //   0 = Interrupt request not enabled
  //
  // bit 2 Unimplemented: Read as .0.
  //
  // bit 1 MI2C1IE: I2C1 Master Events Interrupt Enable bit
  //   1 = Interrupt request enabled
  //   0 = Interrupt request not enabled
  //
  // bit 0 SI2C1IE: I2C1 Slave Events Interrupt Enable bit
  //   1 = Interrupt request enabled
  //   0 = Interrupt request not enabled

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IEC1
    RegBit  SI2C1IE         .off                //
    RegBit  MI2C1IE         .off                //
    RegBit  CMIE            .off                //
    RegBit  CNIE            .off                //
    RegBit  INT1IE          .off                //
    RegBit  IC7IE           .off                //
    RegBit  IC8IE           .off                //
    RegBit  OC3IE           .off                //
    RegBit  OC4IE           .off                //
    RegBit  T4IE            .off                //
    RegBit  T5IE            .off                //
    RegBit  INT2IE          .off                //
    RegBit  U2RXIE          .off                //
    RegBit  U2TXIE          .off                //
  endRegDef


  // =======================================================================
  // IEC2 - interrupt enable control 2
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IEC2
    RegBit  SPF2IE          .off
    RegBit  SPI2IE          .off                //
    RegBit  IC3IE           .off                //
    RegBit  IC4IE           .off                //
    RegBit  IC5IE           .off                //
    RegBit  IC6IE           .off                //
    RegBit  OC5IE           .off                //
    RegBit  OC6IE           .off                //
    RegBit  OC7IE           .off                //
    RegBit  OC8IE           .off                //
  endRegDef


  // =======================================================================
  // IEC3 - interrupt enable control 3
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IEC3
    RegBit  SI2C2IE         .off                //
    RegBit  MI2C2IE         .off                //
    RegBit  INT3IE          .off                //
    RegBit  INT4IE          .off                //
    RegBit  RTCIE           .off                //
  endRegDef


  // =======================================================================
  // IEC4 - interrupt enable control 4
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IEC4
    RegBit  U1ERIE          .off                //
    RegBit  U2ERIE          .off                //
    RegBit  CRCIE           .off                //
    RegBit  LVDIE           .off                //
    RegBit  CTMUIE          .off                //
  endRegDef


  // =======================================================================
  // IEC5 - interrupt enable control 5
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IEC5
    RegBit  U3ERIE          .off                //
    RegBit  U3RXIE          .off                //
    RegBit  U3TXIE          .off                //
    RegBit  SI2C3IE         .off                //
    RegBit  MI2C3IE         .off                //
    RegBit  USB1IE          .off                //
    RegBit  U4ERIE          .off                //
    RegBit  U4RXIE          .off                //
    RegBit  U4TXIE          .off                //
    RegBit  SPF3IE          .off                //
    RegBit  SPI3IE          .off                //
    RegBit  OC9IE           .off                //
    RegBit  IC9IE           .off                //
  endRegDef


  // =======================================================================
  // IEC6 - interrupt enable control 6
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IEC6
    RegBit  GFX1IE          .off                //
  endRegDef


  // =======================================================================
  // IPC0 - interrupt priority control 0
  //          Also see : "29-interrupts.pdf"
  //
  // bit 15 Unimplemented: Read as .0.
  //
  // bit 14-12 T1IP<2:0>: Timer1 Interrupt Priority bits
  // 111 = Interrupt is priority 7 (highest priority interrupt)
  // ...
  // 001 = Interrupt is priority 1
  // 000 = Interrupt source is disabled

    %000 12 bits T1IP_DISABLE
    %001 12 bits T1IP_1
    %010 12 bits T1IP_2
    %011 12 bits T1IP_3
    %100 12 bits T1IP_4
    %101 12 bits T1IP_5
    %110 12 bits T1IP_6
    %111 12 bits T1IP_7
    %111 12 bits T1IP_MASK

  // bit 11 Unimplemented: Read as .0.
  //
  // bit 10-8 OC1IP<2:0>: Output Compare Channel 1 Interrupt Priority bits
  // 111 = Interrupt is priority 7 (highest priority interrupt)
  // ...
  // 001 = Interrupt is priority 1
  // 000 = Interrupt source is disabled

    %000 8 bits OC1IP_DISABLE
    %001 8 bits OC1IP_1
    %010 8 bits OC1IP_2
    %011 8 bits OC1IP_3
    %100 8 bits OC1IP_4
    %101 8 bits OC1IP_5
    %110 8 bits OC1IP_6
    %111 8 bits OC1IP_7
    %111 8 bits OC1IP_MASK

  // bit 7 Unimplemented: Read as .0.
  //
  // bit 6-4 IC1IP<2:0>: Input Capture Channel 1 Interrupt Priority bits
  // 111 = Interrupt is priority 7 (highest priority interrupt)
  // ...
  // 001 = Interrupt is priority 1
  // 000 = Interrupt source is disabled

    %000 4 bits IC1IP_DISABLE
    %001 4 bits IC1IP_1
    %010 4 bits IC1IP_2
    %011 4 bits IC1IP_3
    %100 4 bits IC1IP_4
    %101 4 bits IC1IP_5
    %110 4 bits IC1IP_6
    %111 4 bits IC1IP_7
    %111 4 bits IC1IP_MASK

  // bit 3 Unimplemented: Read as .0.
  //
  // bit 2-0 INT0IP<2:0>: External Interrupt 0 Priority bits
  // 111 = Interrupt is priority 7 (highest priority interrupt)
  // ...
  // 001 = Interrupt is priority 1
  // 000 = Interrupt source is disabled

    %000 0 bits INT0IP_DISABLE
    %001 0 bits INT0IP_1
    %010 0 bits INT0IP_2
    %011 0 bits INT0IP_3
    %100 0 bits INT0IP_4
    %101 0 bits INT0IP_5
    %110 0 bits INT0IP_6
    %111 0 bits INT0IP_7
    %111 0 bits INT0IP_MASK

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC0
    INT0IP_4              .use          //
    IC1IP_4               .use          //
    OC1IP_4               .use          //
    T1IP_4                .use          //
  endRegDef


  todo' next action ... interrupts ... "29-interrupts.pdf"'

  // =======================================================================
  // IPC1 - interrupt priority control 1
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC1
    RegBit  IC2IP0          .off                //
    RegBit  IC2IP1          .off                //
    RegBit  IC2IP2                .on           //

    RegBit  OC2IP0          .off                //
    RegBit  OC2IP1          .off                //
    RegBit  OC2IP2                .on           //

    RegBit  T2IP0           .off                //
    RegBit  T2IP1           .off                //
    RegBit  T2IP2                 .on           //
  endRegDef


  // =======================================================================
  // IPC2 - interrupt priority control 2
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC2
    RegBit  T3IP0           .off                //
    RegBit  T3IP1           .off                //
    RegBit  T3IP2                 .on           //

    RegBit  SPI1IP0         .off                //
    RegBit  SPI1IP1         .off                //
    RegBit  SPI1IP2               .on           //

    RegBit  U1RXIP0         .off                //
    RegBit  U1RXIP1         .off                //
    RegBit  U1RXIP2               .on           //
  endRegDef


  // =======================================================================
  // IPC3 - interrupt priority control 3
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC3
    RegBit  U1TXIP0         .off                //
    RegBit  U1TXIP1         .off                //
    RegBit  U1TXIP2               .on           //

    RegBit  AD1IP0          .off                //
    RegBit  AD1IP1          .off                //
    RegBit  AD1IP2                .on           //
  endRegDef


  // =======================================================================
  // IPC4 - interrupt priority control 4
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC4
    RegBit  SI2C1IP0        .off                //
    RegBit  SI2C1IP1        .off                //
    RegBit  SI2C1IP2              .on           //

    RegBit  MI2C1IP0        .off                //
    RegBit  MI2C1IP1        .off                //
    RegBit  MI2C1IP2              .on           //

    RegBit  CNIP0           .off                //
    RegBit  CNIP1           .off                //
    RegBit  CNIP2                 .on           //
  endRegDef


  // =======================================================================
  // IPC5 - interrupt priority control 5
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC5
    RegBit  INT1IP0         .off                //
    RegBit  INT1IP1         .off                //
    RegBit  INT1IP2               .on           //

    RegBit  IC7IP0          .off                //
    RegBit  IC7IP1          .off                //
    RegBit  IC7IP2                .on           //

    RegBit  IC8IP0          .off                //
    RegBit  IC8IP1          .off                //
    RegBit  IC8IP2                .on           //
  endRegDef


  // =======================================================================
  // IPC6 - interrupt priority control 6
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC6
    RegBit  OC3IP0          .off                //
    RegBit  OC3IP1          .off                //
    RegBit  OC3IP2                .on           //

    RegBit  OC4IP0          .off                //
    RegBit  OC4IP1          .off                //
    RegBit  OC4IP2                .on           //

    RegBit  T4IP0           .off                //
    RegBit  T4IP1           .off                //
    RegBit  T4IP2                 .on           //
  endRegDef


  // =======================================================================
  // IPC7 - interrupt priority control 7
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC7
    RegBit  T5IP0           .off                //
    RegBit  T5IP1           .off                //
    RegBit  T5IP2                 .on           //

    RegBit  INT2IP0         .off                //
    RegBit  INT2IP1         .off                //
    RegBit  INT2IP2               .on           //

    RegBit  U2RXIP0         .off                //
    RegBit  U2RXIP1         .off                //
    RegBit  U2RXIP2               .on           //

    RegBit  U2TXIP0         .off                //
    RegBit  U2TXIP1         .off                //
    RegBit  U2TXIP2               .on           //
  endRegDef


  // =======================================================================
  // IPC8 - interrupt priority control 8
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC8
    RegBit  SPI2IP0         .off                //
    RegBit  SPI2IP1         .off                //
    RegBit  SPI2IP2               .on           //
  endRegDef


  // =======================================================================
  // IPC9 - interrupt priority control 9
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC9
    RegBit  IC3IP0          .off                //
    RegBit  IC3IP1          .off                //
    RegBit  IC3IP2                .on           //

    RegBit  IC4IP0          .off                //
    RegBit  IC4IP1          .off                //
    RegBit  IC4IP2                .on           //

    RegBit  IC5IP0          .off                //
    RegBit  IC5IP1          .off                //
    RegBit  IC5IP2                .on           //
  endRegDef


  // =======================================================================
  // IPC10 - interrupt priority control 10
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC10
    RegBit  IC6IP0          .off                //
    RegBit  IC6IP1          .off                //
    RegBit  IC6IP2                .on           //

    RegBit  OC5IP0          .off                //
    RegBit  OC5IP1          .off                //
    RegBit  OC5IP2                .on           //

    RegBit  OC6IP0          .off                //
    RegBit  OC6IP1          .off                //
    RegBit  OC6IP2                .on           //

    RegBit  OC7IP0          .off                //
    RegBit  OC7IP1          .off                //
    RegBit  OC7IP2                .on           //
  endRegDef


  // =======================================================================
  // IPC11 - interrupt priority control 11
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC11
    RegBit  OC8IP0          .off                //
    RegBit  OC8IP1          .off                //
    RegBit  OC8IP2                .on           //
  endRegDef


  // =======================================================================
  // IPC12 - interrupt priority control 12
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC12
    RegBit  SI2C2IP0        .off                //
    RegBit  SI2C2IP1        .off                //
    RegBit  SI2C2IP2              .on           //

    RegBit  MI2C2IP0        .off                //
    RegBit  MI2C2IP1        .off                //
    RegBit  MI2C2IP2              .on           //
  endRegDef


  // =======================================================================
  // IPC13 - interrupt priority control 13
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC13
    RegBit  INT3IP0         .off                //
    RegBit  INT3IP1         .off                //
    RegBit  INT3IP2               .on           //

    RegBit  INT4IP0         .off                //
    RegBit  INT4IP1         .off                //
    RegBit  INT4IP2               .on           //
  endRegDef


  // =======================================================================
  // IPC15 - interrupt priority control 15
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC15
    RegBit  RTCIP0          .off                //
    RegBit  RTCIP1          .off                //
    RegBit  RTCIP2                .on           //
  endRegDef


  // =======================================================================
  // IPC16 - interrupt priority control 16
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC16
    RegBit  U1ERIP0         .off                //
    RegBit  U1ERIP1         .off                //
    RegBit  U1ERIP2               .on           //

    RegBit  U2ERIP0         .off                //
    RegBit  U2ERIP1         .off                //
    RegBit  U2ERIP2               .on           //

    RegBit  CRCIP0          .off                //
    RegBit  CRCIP1          .off                //
    RegBit  CRCIP2                .on           //
  endRegDef


  // =======================================================================
  // IPC18 - interrupt priority control 18
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC18
    RegBit  LVDIP0          .off                //
    RegBit  LVDIP1          .off                //
    RegBit  LVDIP2                .on           //
  endRegDef


  // =======================================================================
  // IPC19 - interrupt priority control 19
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC19
    RegBit  CTMUIP0         .off                //
    RegBit  CTMUIP1         .off                //
    RegBit  CTMUIP2               .on           //
  endRegDef


  // =======================================================================
  // IPC21 - interrupt priority control 21
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC20
    RegBit  U3ERIP0         .off                //
    RegBit  U3ERIP1         .off                //
    RegBit  U3ERIP2               .on           //

    RegBit  U3RXIP0         .off                //
    RegBit  U3RXIP1         .off                //
    RegBit  U3RXIP2               .on           //

    RegBit  U3TXIP0         .off                //
    RegBit  U3TXIP1         .off                //
    RegBit  U3TXIP2               .on           //
  endRegDef


  // =======================================================================
  // IPC21 - interrupt priority control 21
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC21
    RegBit  SI2C3IP0        .off                //
    RegBit  SI2C3IP1        .off                //
    RegBit  SI2C3IP2              .on           //

    RegBit  MI2C3IP0        .off                //
    RegBit  MI2C3IP1        .off                //
    RegBit  MI2C3IP2              .on           //

    RegBit  USB1IP0         .off                //
    RegBit  USB1IP1         .off                //
    RegBit  USB1IP2               .on           //

    RegBit  U4ERIP0         .off                //
    RegBit  U4ERIP1         .off                //
    RegBit  U4ERIP2               .on           //
  endRegDef


  // =======================================================================
  // IPC22 - interrupt priority control 22
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC22
    RegBit  U4RXIP0         .off                //
    RegBit  U4RXIP1         .off                //
    RegBit  U4RXIP2               .on           //

    RegBit  U4TXIP0         .off                //
    RegBit  U4TXIP1         .off                //
    RegBit  U4TXIP2               .on           //

    RegBit  SPF3IP0         .off                //
    RegBit  SPF3IP1         .off                //
    RegBit  SPF3IP2               .on           //

    RegBit  SPI3IP0         .off                //
    RegBit  SPI3IP1         .off                //
    RegBit  SPI3IP2               .on           //
  endRegDef


  // =======================================================================
  // IPC23 - interrupt priority control 23
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC23
    RegBit  OC9IP0          .off                //
    RegBit  OC9IP1          .off                //
    RegBit  OC9IP2                .on           //

    RegBit  IC9IP0          .off                //
    RegBit  IC9IP1          .off                //
    RegBit  IC9IP2                .on           //
  endRegDef


  // =======================================================================
  // IPC25 - interrupt priority control 25
  //          Also see : "29-interrupts.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IPC25
    RegBit  GFX1IP0         .off                //
    RegBit  GFX1IP1         .off                //
    RegBit  GFX1IP2               .on           //
  endRegDef


  // =======================================================================
  // INTTREG - INTREG -
  //          Also see : "23-security.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef INTTREG
    RegBit  VHOLD           .off                //
    RegBit  CPUIRQ          .off                //

    RegBit  VECNUM0         .off                //
    RegBit  VECNUM1         .off                //
    RegBit  VECNUM2         .off                //
    RegBit  VECNUM3         .off                //
    RegBit  VECNUM4         .off                //
    RegBit  VECNUM5         .off                //
    RegBit  VECNUM6         .off                //

    RegBit  ILR0            .off                //
    RegBit  ILR1            .off                //
    RegBit  ILR2            .off                //
    RegBit  ILR3            .off                //
  endRegDef


  // =======================================================================
  // TMR1 - Timer count register 1

  0 constant TMR1_IniVal


  // =======================================================================
  // PR1 - period register 1

  $fff constant PR1_IniVal                      // Default startup $ffff


  // =======================================================================
  // T1CON - timer 1 configuration
  //         see "11-timers.pdf" page 11-6
  //         the following bits are not implemented on all timers,
  //         see the pdf.

  // TCS : Timer Clock Source Select
  //       no effect for timer 3, 5, 7, 9 when used in 32 bit mode
  // 0 - Internal Clock (Fosc / 2)
  // 1 - External clock from TxCK pin

  // TSYNC : External clock input synchronization, not used when TCS = 0
  //         timer 1 only
  // 0 - Do not sync external clock
  // 1 - Sync external clock

  // T32 : Time in 32 bit mode
  //       timer 2, 4, 6, 8 only
  // 0 - not in 32 bit mode
  // 1 - cascaded with next timer to make a 32 bit timer

  // TGATE : Timer Gated Time Accumulation bit, not used when TCS = 1
  // 0 - Gated time accumulation disabled
  // 1 - Gated time accumulation enabled

  // TSIDL : Stop In Idle Mode
  //       no effect for timer 3, 5, 7, 9 when used in 32 bit mode
  // 0 - keep timer running in idle mode
  // 1 - stop timer when device is in idle mode

  // TON : set to turn timer on
  //       no effect for timer 3, 5, 7, 9 when used in 32 bit mode

  // TCKPS : Timer input clock pre scale
  // 10
  // 00 - div   1 prescale
  // 01 - div   8 prescale
  // 10 - div  64 prescale
  // 11 - div 256 prescale

    %00 4 bits TCKPS_1
    %01 4 bits TCKPS_8
    %10 4 bits TCKPS_64
    %11 4 bits TCKPS_256

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef T1CON
    RegBit  TCS             .off                // Internal clock
    RegBit  TSYNC           .off                // Don't care
    RegBit  TGATE           .off                // No gated time accumulation
    RegBit  TSIDL           .off                // Timer runs in idle mode
    RegBit  TON                   .on           // Timer is on
            TCKPS_1               .use          // Prescaler (1, 8, 64 or 256, 1 is default)
  endRegDef

  // =======================================================================
  // TMR2 - Timer count register 2

  0 constant TMR2_IniVal


  // =======================================================================
  // TMR3 - Timer count register 3

  0 constant TMR3_IniVal


  // =======================================================================
  // PR2 - period register 2

  $ddd constant PR2_IniVal                      // Default startup $ffff


  // =======================================================================
  // PR3 - period register 3

  $ffff constant PR3_IniVal                     // Default startup $ffff


  // =======================================================================
  // T2CON - timer 2 configuration
  //         see "11-timers.pdf" page 11-6
  //         see T1CON for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef T2CON
    RegBit  TCS             .off                //
    RegBit  T32             .off                // do not chain with tmr3
    RegBit  TGATE           .off                //
    RegBit  TSIDL           .off                //
    RegBit  TON                   .on           //
            TCKPS_1               .use          // Prescaler (1, 8, 64 or 256, 1 is default)
  endRegDef


  // =======================================================================
  // T3CON - timer 3 configuration
  //         see "11-timers.pdf" page 11-6
  //         see T1CON for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef T3CON
    RegBit  TCS             .off                //
    RegBit  TGATE           .off                //
    RegBit  TSIDL           .off                //
    RegBit  TON             .off                //
            TCKPS_1               .use          // Prescaler (1, 8, 64 or 256, 1 is default)
  endRegDef


  // =======================================================================
  // TMR4 - Timer count register 4

  0 constant TMR4_IniVal


  // =======================================================================
  // TMR5 - Timer count register 5

  0 constant TMR5_IniVal


  // =======================================================================
  // PR4 - period register 4

  $ffff constant PR4_IniVal                     // Default startup $ffff


  // =======================================================================
  // PR5 - period register 5

  $ffff constant PR5_IniVal                     // Default startup $ffff


  // =======================================================================
  // T4CON - timer 4 configuration
  //         see "11-timers.pdf" page 11-6
  //         see T1CON for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef T4CON
    RegBit  TCS             .off                //
    RegBit  T32             .off                // do not chain with tmr5
    RegBit  TGATE           .off                //
    RegBit  TSIDL           .off                //
    RegBit  TON             .off                //
            TCKPS_1               .use          // Prescaler (1, 8, 64 or 256, 1 is default)
  endRegDef


  // =======================================================================
  // T5CON - timer 5 configuration
  //         see "11-timers.pdf" page 11-6
  //         see T1CON for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef T5CON
    RegBit  TCS             .off                //
    RegBit  TGATE           .off                //
    RegBit  TSIDL           .off                //
    RegBit  TON             .off                //
            TCKPS_1               .use          // Prescaler (1, 8, 64 or 256; 1 is default)
  endRegDef


  // =======================================================================
  // IC1CON1 - Input Capture Control register 1
  //         see "12-input-capture.pdf" page 12-3

  // ICBNE : Input Capture Buffer Empty Status
  //   0 - capture buffer is empty
  //   1 - capture buffer has 1 to 4 items

  // ICOV : Input Capture Overflow
  //   0 - No capture overflow detected
  //   1 - Capture bufferr overflow detected

  // ICTMTR : Input Capture Timer select
  //   0 - Timer 2 value is captured on capture event
  //   1 - Timer 3 value is captured on capture event

  // ICSIDL : Input Capture Stop In Idle control
  //   0 - Input capture continues in CPU idle mode
  //   1 - Input capture halts in CPU idle mode

  // ICM : Input Capture Mode select

    %000 0 bits ICM_OFF
    %001 0 bits ICM_ALL_EDGES
    %010 0 bits ICM_FALLING_EDGE
    %011 0 bits ICM_RISING_EDGE
    %100 0 bits ICM_4TH_RISING_EDGE
    %101 0 bits ICM_16TH_RISING_EDGE
    %111 0 bits ICM_INT_MODE

  // ICI : Nr of captures per interrupt

    %00 5 bits ICI_1
    %01 5 bits ICI_2
    %10 5 bits ICI_3
    %11 5 bits ICI_4

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IC1CON1
    RegBit  ICBNE           .off                // RO - Buffer empty
    RegBit  ICOV            .off                // RO - no overflow
    RegBit  ICSIDL          .off                // Continue capture in IDLE mode
            ICM_OFF               .use          // Capture module is off
            ICI_1                 .use          // Interrupt after every 1 capture
  endRegDef


  // ----- IC1CON2 Bits -----------------------------------------------------

  regDef IC1CON2
    RegBit  TRIGSTAT        .off                //
    RegBit  ICTRIG          .off                //
    RegBit  IC32            .off                //

    RegBit  SYNCSEL0        .off                //
    RegBit  SYNCSEL1        .off                //
    RegBit  SYNCSEL2        .off                //
    RegBit  SYNCSEL3        .off                //
    RegBit  SYNCSEL4        .off                //
  endRegDef


  // =======================================================================
  // IC2CON1 - Input Capture Control register 2
  //         see "12-input-capture.pdf" page 12-3
  //         see IC1CON for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IC2CON1
    RegBit  ICBNE           .off                // RO - Buffer empty
    RegBit  ICOV            .off                // RO - no overflow
    RegBit  ICSIDL          .off                // Continue capture in IDLE mode
            ICM_OFF               .use          // Capture module is off
            ICI_1                 .use          // Interrupt after every 1 capture
  endRegDef


  // ----- IC2CON2 Bits -----------------------------------------------------

  regDef IC2CON2
    RegBit  TRIGSTAT        .off                //
    RegBit  ICTRIG          .off                //
    RegBit  IC32            .off                //

    RegBit  SYNCSEL0        .off                //
    RegBit  SYNCSEL1        .off                //
    RegBit  SYNCSEL2        .off                //
    RegBit  SYNCSEL3        .off                //
    RegBit  SYNCSEL4        .off                //
  endRegDef


  // =======================================================================
  // IC3CON1 - Input Capture Control register 3
  //         see "12-input-capture.pdf" page 12-3
  //         see IC1CON for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IC3CON1
    RegBit  ICBNE           .off                // RO - Buffer empty
    RegBit  ICOV            .off                // RO - no overflow
    RegBit  ICSIDL          .off                // Continue capture in IDLE mode
            ICM_OFF               .use          // Capture module is off
            ICI_1                 .use          // Interrupt after every 1 capture
  endRegDef


  // ----- IC3CON2 Bits -----------------------------------------------------

  regDef IC3CON2
    RegBit  TRIGSTAT        .off                //
    RegBit  ICTRIG          .off                //
    RegBit  IC32            .off                //

    RegBit  SYNCSEL0        .off                //
    RegBit  SYNCSEL1        .off                //
    RegBit  SYNCSEL2        .off                //
    RegBit  SYNCSEL3        .off                //
    RegBit  SYNCSEL4        .off                //
  endRegDef


  // =======================================================================
  // IC4CON1 - Input Capture Control register 1
  //         see "12-input-capture.pdf" page 12-3
  //         see IC1CON for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IC4CON1
    RegBit  ICBNE           .off                // RO - Buffer empty
    RegBit  ICOV            .off                // RO - no overflow
    RegBit  ICSIDL          .off                // Continue capture in IDLE mode
            ICM_OFF               .use          // Capture module is off
            ICI_1                 .use          // Interrupt after every 1 capture
  endRegDef


  // ----- IC4CON2 Bits -----------------------------------------------------

  regDef IC4CON2
    RegBit  TRIGSTAT        .off                //
    RegBit  ICTRIG          .off                //
    RegBit  IC32            .off                //

    RegBit  SYNCSEL0        .off                //
    RegBit  SYNCSEL1        .off                //
    RegBit  SYNCSEL2        .off                //
    RegBit  SYNCSEL3        .off                //
    RegBit  SYNCSEL4        .off                //
  endRegDef


  // =======================================================================
  // IC5CON1 - Input Capture Control register 5
  //         see "12-input-capture.pdf" page 12-3
  //         see IC1CON for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IC5CON1
    RegBit  ICBNE           .off                // RO - Buffer empty
    RegBit  ICOV            .off                // RO - no overflow
    RegBit  ICSIDL          .off                // Continue capture in IDLE mode
            ICM_OFF               .use          // Capture module is off
            ICI_1                 .use          // Interrupt after every 1 capture
  endRegDef


  // ----- IC5CON2 Bits -----------------------------------------------------

  regDef IC5CON2
    RegBit  TRIGSTAT        .off                //
    RegBit  ICTRIG          .off                //
    RegBit  IC32            .off                //

    RegBit  SYNCSEL0        .off                //
    RegBit  SYNCSEL1        .off                //
    RegBit  SYNCSEL2        .off                //
    RegBit  SYNCSEL3        .off                //
    RegBit  SYNCSEL4        .off                //
  endRegDef


  // =======================================================================
  // IC6CON1 - Input Capture Control register 6
  //         see "12-input-capture.pdf" page 12-3
  //         see IC1CON for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IC6CON1
    RegBit  ICBNE           .off                // RO - Buffer empty
    RegBit  ICOV            .off                // RO - no overflow
    RegBit  ICSIDL          .off                // Continue capture in IDLE mode
            ICM_OFF               .use          // Capture module is off
            ICI_1                 .use          // Interrupt after every 1 capture
  endRegDef


  // ----- IC6CON2 Bits -----------------------------------------------------

  regDef IC6CON2
    RegBit  TRIGSTAT        .off                //
    RegBit  ICTRIG          .off                //
    RegBit  IC32            .off                //

    RegBit  SYNCSEL0        .off                //
    RegBit  SYNCSEL1        .off                //
    RegBit  SYNCSEL2        .off                //
    RegBit  SYNCSEL3        .off                //
    RegBit  SYNCSEL4        .off                //
  endRegDef


  // =======================================================================
  // IC7CON1 - Input Capture Control register 7
  //         see "12-input-capture.pdf" page 12-3
  //         see IC1CON for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IC7CON1
    RegBit  ICBNE           .off                // RO - Buffer empty
    RegBit  ICOV            .off                // RO - no overflow
    RegBit  ICSIDL          .off                // Continue capture in IDLE mode
            ICM_OFF               .use          // Capture module is off
            ICI_1                 .use          // Interrupt after every 1 capture
  endRegDef


  // ----- IC7CON2 Bits -----------------------------------------------------

  regDef IC7CON2
    RegBit  TRIGSTAT        .off                //
    RegBit  ICTRIG          .off                //
    RegBit  IC32            .off                //

    RegBit  SYNCSEL0        .off                //
    RegBit  SYNCSEL1        .off                //
    RegBit  SYNCSEL2        .off                //
    RegBit  SYNCSEL3        .off                //
    RegBit  SYNCSEL4        .off                //
  endRegDef


  // =======================================================================
  // IC8CON1 - Input Capture Control register 8
  //         see "12-input-capture.pdf" page 12-3
  //         see IC1CON for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IC8CON1
    RegBit  ICBNE           .off                // RO - Buffer empty
    RegBit  ICOV            .off                // RO - no overflow
    RegBit  ICSIDL          .off                // Continue capture in IDLE mode
            ICM_OFF               .use          // Capture module is off
            ICI_1                 .use          // Interrupt after every 1 capture
  endRegDef


  // ----- IC8CON2 Bits -----------------------------------------------------

  regDef IC8CON2
    RegBit  TRIGSTAT        .off                //
    RegBit  ICTRIG          .off                //
    RegBit  IC32            .off                //

    RegBit  SYNCSEL0        .off                //
    RegBit  SYNCSEL1        .off                //
    RegBit  SYNCSEL2        .off                //
    RegBit  SYNCSEL3        .off                //
    RegBit  SYNCSEL4        .off                //
  endRegDef


  // =======================================================================
  // IC9CON1 - Input Capture Control register 9
  //         see "12-input-capture.pdf" page 12-3
  //         see IC1CON for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef IC9CON1
    RegBit  ICBNE           .off                // RO - Buffer empty
    RegBit  ICOV            .off                // RO - no overflow
    RegBit  ICSIDL          .off                // Continue capture in IDLE mode
            ICM_OFF               .use          // Capture module is off
            ICI_1                 .use          // Interrupt after every 1 capture
  endRegDef


  // ----- IC9CON2 Bits -----------------------------------------------------

  regDef IC9CON2
    RegBit  TRIGSTAT        .off                //
    RegBit  ICTRIG          .off                //
    RegBit  IC32            .off                //

    RegBit  SYNCSEL0        .off                //
    RegBit  SYNCSEL1        .off                //
    RegBit  SYNCSEL2        .off                //
    RegBit  SYNCSEL3        .off                //
    RegBit  SYNCSEL4        .off                //
  endRegDef


  // =======================================================================
  // OC1RS - Secondary output compare register
  //          see "13-outout-compare.pdf" page 13-4

  0 constant OC1RS_IniVal


  // =======================================================================
  // OC1R - Output compare register
  //          see "13-outout-compare.pdf" page 13-4

  0 constant OC1R_IniVal


  // =======================================================================
  // OC1CON1 - Output Compare Control register
  //          see "13-outout-compare.pdf" page 13-3

  // OCTSEL : Output Compare Timer Select
  // 0 - Timer 2
  // 1 - Timer 3

  // OCFLT : PWM Fault Condition status bit
  //         only used for OCM = 111 (PWM fault protection mode)
  // 0 - no PWM fault condition was detected
  // 1 - a PEW fault condition was detected

  // OCSIDL : Stop OutputCompare in Idle mode when set
  // 0 - Keep running in idle mode
  // 1 - Stop running in idle mode

  // OCM : Output Compare Mode select

    %000 0 bits OCM_OFF               // Module disabled
    %001 0 bits OCM_ALOSM             // Active Low  One Shot Mode
    %010 0 bits OCM_AHOSM             // Active high One Shot Mode
    %011 0 bits OCM_TM                // Toggle mode
    %100 0 bits OCM_DOSM              // Delayed One Shot Mode
    %101 0 bits OCM_CPM               // Continues Pulse Mode
    %110 0 bits OCM_PWM               // PWM mode
    %111 0 bits OCM_PWMF              // PWM mode with Fault detection

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef OC1CON1
    RegBit  OCSIDL          .off                // Run in Idle mode
            OCM_OFF               .use          // Module disabled
  endRegDef


  // ----- OC1CON2 Bits -----------------------------------------------------

  regDef OC1CON2
    RegBit  OCTRIS          .off                //
    RegBit  TRIGSTAT        .off                //
    RegBit  OCTRIG          .off                //
    RegBit  OC32            .off                //
    RegBit  OCINV           .off                //
    RegBit  FLTTRIEN        .off                //
    RegBit  FLTOUT          .off                //
    RegBit  FLTMD           .off                //
    RegBit  SYNCSEL0        .off                //
    RegBit  SYNCSEL1        .off                //
    RegBit  SYNCSEL2        .off                //
    RegBit  SYNCSEL3        .off                //
    RegBit  SYNCSEL4        .off                //
    RegBit  (DCB0)          .off                //
    RegBit  (DCB1)          .off                //
    RegBit  FLTMODE         .off                //
  endRegDef


  // =======================================================================
  // OC2RS - Secondary output compare register
  //          see "13-outout-compare.pdf" page 13-4

  0 constant OC2RS_IniVal


  // =======================================================================
  // OC2R - Output compare register
  //          see "13-outout-compare.pdf" page 13-4

  0 constant OC2R_IniVal


  // =======================================================================
  // OC2CON1 - Output Compare Control register
  //          see "13-outout-compare.pdf" page 13-3
  //          see OC1CON for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef OC2CON1
    RegBit  OCSIDL          .off                // Run in Idle mode
            OCM_OFF               .use          // Module disabled
  endRegDef


  // ----- OC2CON2 Bits -----------------------------------------------------

  regDef OC2CON2
    RegBit  OCTRIS          .off                //
    RegBit  TRIGSTAT        .off                //
    RegBit  OCTRIG          .off                //
    RegBit  OC32            .off                //
    RegBit  OCINV           .off                //
    RegBit  FLTTRIEN        .off                //
    RegBit  FLTOUT          .off                //
    RegBit  FLTMD           .off                //
    RegBit  SYNCSEL0        .off                //
    RegBit  SYNCSEL1        .off                //
    RegBit  SYNCSEL2        .off                //
    RegBit  SYNCSEL3        .off                //
    RegBit  SYNCSEL4        .off                //
    RegBit  (DCB0)          .off                //
    RegBit  (DCB1)          .off                //
    RegBit  FLTMODE         .off                //
  endRegDef


  // =======================================================================
  // OC3RS - Secondary output compare register
  //          see "13-outout-compare.pdf" page 13-4

  0 constant OC3RS_IniVal


  // =======================================================================
  // OC3R - Output compare register
  //          see "13-outout-compare.pdf" page 13-4

  0 constant OC3R_IniVal


  // =======================================================================
  // OC3CON1 - Output Compare Control register
  //          see "13-outout-compare.pdf" page 13-3
  //          see OC1CON for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef OC3CON1
    RegBit  OCSIDL          .off                // Run in Idle mode
            OCM_OFF               .use          // Module disabled
  endRegDef


  // ----- OC3CON2 Bits -----------------------------------------------------

  regDef OC3CON2
    RegBit  OCTRIS          .off                //
    RegBit  TRIGSTAT        .off                //
    RegBit  OCTRIG          .off                //
    RegBit  OC32            .off                //
    RegBit  OCINV           .off                //
    RegBit  FLTTRIEN        .off                //
    RegBit  FLTOUT          .off                //
    RegBit  FLTMD           .off                //
    RegBit  SYNCSEL0        .off                //
    RegBit  SYNCSEL1        .off                //
    RegBit  SYNCSEL2        .off                //
    RegBit  SYNCSEL3        .off                //
    RegBit  SYNCSEL4        .off                //
    RegBit  (DCB0)          .off                //
    RegBit  (DCB1)          .off                //
    RegBit  FLTMODE         .off                //
  endRegDef


  // =======================================================================
  // OC4RS - Secondary output compare register
  //         see "12-input-capture.pdf" page 12-4

  0 constant OC4RS_IniVal


  // =======================================================================
  // OC4R - Output compare register
  //          see "13-outout-compare.pdf" page 13-4

  0 constant OC4R_IniVal


  // =======================================================================
  // OC4CON1 - Output Compare Control register
  //          see "13-outout-compare.pdf" page 13-3
  //          see OC1CON for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef OC4CON1
    RegBit  OCSIDL          .off                // Run in Idle mode
            OCM_OFF               .use          // Module disabled
  endRegDef


  // ----- OC4CON2 Bits -----------------------------------------------------

  regDef OC4CON2
    RegBit  OCTRIS          .off                //
    RegBit  TRIGSTAT        .off                //
    RegBit  OCTRIG          .off                //
    RegBit  OC32            .off                //
    RegBit  OCINV           .off                //
    RegBit  FLTTRIEN        .off                //
    RegBit  FLTOUT          .off                //
    RegBit  FLTMD           .off                //
    RegBit  SYNCSEL0        .off                //
    RegBit  SYNCSEL1        .off                //
    RegBit  SYNCSEL2        .off                //
    RegBit  SYNCSEL3        .off                //
    RegBit  SYNCSEL4        .off                //
    RegBit  (DCB0)          .off                //
    RegBit  (DCB1)          .off                //
    RegBit  FLTMODE         .off                //
  endRegDef


  // =======================================================================
  // OC5RS - Secondary output compare register
  //          see "13-outout-compare.pdf" page 13-4

  0 constant OC5RS_IniVal


  // =======================================================================
  // OC5R - Output compare register
  //          see "13-outout-compare.pdf" page 13-4

  0 constant OC5R_IniVal


  // =======================================================================
  // OC5CON1 - Output Compare Control register
  //          see "13-outout-compare.pdf" page 13-3
  //          see OC1CON for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef OC5CON1
    RegBit  OCSIDL          .off                // Run in Idle mode
            OCM_OFF               .use          // Module disabled
  endRegDef


  // ----- OC5CON2 Bits -----------------------------------------------------

  regDef OC5CON2
    RegBit  OCTRIS          .off                //
    RegBit  TRIGSTAT        .off                //
    RegBit  OCTRIG          .off                //
    RegBit  OC32            .off                //
    RegBit  OCINV           .off                //
    RegBit  FLTTRIEN        .off                //
    RegBit  FLTOUT          .off                //
    RegBit  FLTMD           .off                //
    RegBit  SYNCSEL0        .off                //
    RegBit  SYNCSEL1        .off                //
    RegBit  SYNCSEL2        .off                //
    RegBit  SYNCSEL3        .off                //
    RegBit  SYNCSEL4        .off                //
    RegBit  (DCB0)          .off                //
    RegBit  (DCB1)          .off                //
    RegBit  FLTMODE         .off                //
  endRegDef


  // =======================================================================
  // OC6RS - Secondary output compare register
  //          see "13-outout-compare.pdf" page 13-4

  0 constant OC6RS_IniVal


  // =======================================================================
  // OC6R - Output compare register
  //          see "13-outout-compare.pdf" page 13-4

  0 constant OC6R_IniVal


  // =======================================================================
  // OC6CON1 - Output Compare Control register
  //          see "13-outout-compare.pdf" page 13-3
  //          see OC1CON for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef OC6CON1
    RegBit  OCSIDL          .off                // Run in Idle mode
            OCM_OFF               .use          // Module disabled
  endRegDef


  // ----- OC6CON2 Bits -----------------------------------------------------

  regDef OC6CON2
    RegBit  OCTRIS          .off                //
    RegBit  TRIGSTAT        .off                //
    RegBit  OCTRIG          .off                //
    RegBit  OC32            .off                //
    RegBit  OCINV           .off                //
    RegBit  FLTTRIEN        .off                //
    RegBit  FLTOUT          .off                //
    RegBit  FLTMD           .off                //
    RegBit  SYNCSEL0        .off                //
    RegBit  SYNCSEL1        .off                //
    RegBit  SYNCSEL2        .off                //
    RegBit  SYNCSEL3        .off                //
    RegBit  SYNCSEL4        .off                //
    RegBit  (DCB0)          .off                //
    RegBit  (DCB1)          .off                //
    RegBit  FLTMODE         .off                //
  endRegDef


  // =======================================================================
  // OC7RS - Secondary output compare register
  //          see "13-outout-compare.pdf" page 13-4

  0 constant OC7RS_IniVal


  // =======================================================================
  // OC7R - Output compare register
  //          see "13-outout-compare.pdf" page 13-4

  0 constant OC7R_IniVal


  // =======================================================================
  // OC7CON1 - Output Compare Control register
  //          see "13-outout-compare.pdf" page 13-3
  //          see OC1CON for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef OC7CON1
    RegBit  OCSIDL          .off                // Run in Idle mode
            OCM_OFF               .use          // Module disabled
  endRegDef


  // ----- OC7CON2 Bits -----------------------------------------------------

  regDef OC7CON2
    RegBit  OCTRIS          .off                //
    RegBit  TRIGSTAT        .off                //
    RegBit  OCTRIG          .off                //
    RegBit  OC32            .off                //
    RegBit  OCINV           .off                //
    RegBit  FLTTRIEN        .off                //
    RegBit  FLTOUT          .off                //
    RegBit  FLTMD           .off                //
    RegBit  SYNCSEL0        .off                //
    RegBit  SYNCSEL1        .off                //
    RegBit  SYNCSEL2        .off                //
    RegBit  SYNCSEL3        .off                //
    RegBit  SYNCSEL4        .off                //
    RegBit  (DCB0)          .off                //
    RegBit  (DCB1)          .off                //
    RegBit  FLTMODE         .off                //
  endRegDef


  // =======================================================================
  // OC8RS - Secondary output compare register
  //          see "13-outout-compare.pdf" page 13-4

  0 constant OC8RS_IniVal


  // =======================================================================
  // OC8R - Output compare register
  //          see "13-outout-compare.pdf" page 13-4

  0 constant OC8R_IniVal


  // =======================================================================
  // OC8CON1 - Output Compare Control register
  //          see "13-outout-compare.pdf" page 13-3
  //          see OC1CON for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef OC8CON1
    RegBit  OCSIDL          .off                // Run in Idle mode
            OCM_OFF               .use          // Module disabled
  endRegDef


  // ----- OC8CON2 Bits -----------------------------------------------------

  regDef OC8CON2
    RegBit  OCTRIS          .off                //
    RegBit  TRIGSTAT        .off                //
    RegBit  OCTRIG          .off                //
    RegBit  OC32            .off                //
    RegBit  OCINV           .off                //
    RegBit  FLTTRIEN        .off                //
    RegBit  FLTOUT          .off                //
    RegBit  FLTMD           .off                //
    RegBit  SYNCSEL0        .off                //
    RegBit  SYNCSEL1        .off                //
    RegBit  SYNCSEL2        .off                //
    RegBit  SYNCSEL3        .off                //
    RegBit  SYNCSEL4        .off                //
    RegBit  (DCB0)          .off                //
    RegBit  (DCB1)          .off                //
    RegBit  FLTMODE         .off                //
  endRegDef



  // =======================================================================
  // OC9RS - Secondary output compare register
  //          see "13-outout-compare.pdf" page 13-4

  0 constant OC9RS_IniVal


  // =======================================================================
  // OC9R - Output compare register
  //          see "13-outout-compare.pdf" page 13-4

  0 constant OC9R_IniVal


  // =======================================================================
  // OC9CON1 - Output Compare Control register
  //          see "13-outout-compare.pdf" page 13-3
  //          see OC1CON for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef OC9CON1
    RegBit  OCSIDL          .off                // Run in Idle mode
            OCM_OFF               .use          // Module disabled
  endRegDef


  // ----- OC9CON2 Bits -----------------------------------------------------

  regDef OC9CON2
    RegBit  OCTRIS          .off                //
    RegBit  TRIGSTAT        .off                //
    RegBit  OCTRIG          .off                //
    RegBit  OC32            .off                //
    RegBit  OCINV           .off                //
    RegBit  FLTTRIEN        .off                //
    RegBit  FLTOUT          .off                //
    RegBit  FLTMD           .off                //
    RegBit  SYNCSEL0        .off                //
    RegBit  SYNCSEL1        .off                //
    RegBit  SYNCSEL2        .off                //
    RegBit  SYNCSEL3        .off                //
    RegBit  SYNCSEL4        .off                //
    RegBit  (DCB0)          .off                //
    RegBit  (DCB1)          .off                //
    RegBit  FLTMODE         .off                //
  endRegDef



  // =======================================================================
  // I2C1BRG -
  //            see "19-i2c.pdf"

  0 constant I2C1BRG_IniVal


  // =======================================================================
  // I2C1CON -
  //            see "19-i2c.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef I2C1CON
    RegBit  SEN             .off                //
    RegBit  RSEN            .off                //
    RegBit  PEN             .off                //
    RegBit  RCEN            .off                //
    RegBit  ACKEN           .off                //
    RegBit  ACKDT           .off                //
    RegBit  STREN           .off                //
    RegBit  GCEN            .off                //
    RegBit  SMEN            .off                //
    RegBit  DISSLW          .off                //
    RegBit  A10M            .off                //
    RegBit  IPMIEN          .off                //
    RegBit  SCLREL          .off                //
    RegBit  I2CSIDL         .off                //
    RegBit  I2CEN           .off                //
  endRegDef


  // =======================================================================
  // I2C1STAT
  //            see "19-i2c.pdf"

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef I2C1STAT
    RegBit  TBF             .off                //
    RegBit  RBF             .off                //
    RegBit  R_W             .off                //
    RegBit  S               .off                //
    RegBit  P               .off                //
    RegBit  D_A             .off                //
    RegBit  I2COV           .off                //
    RegBit  IWCOL           .off                //
    RegBit  (ADD10)         .off                //
    RegBit  GCSTAT          .off                //
    RegBit  BCL             .off                //
    RegBit  TRSTAT          .off                //
    RegBit  ACKSTAT         .off                //
  endRegDef


  // =======================================================================
  // I2C2BRG
  //            see "19-i2c.pdf"
  //            see i2c1 for bit definitions

  0 constant I2C2BRG_IniVal


  // =======================================================================
  // I2C2CON -
  //            see "19-i2c.pdf"
  //            see i2c1 for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef I2C2CON
    RegBit  SEN             .off                //
    RegBit  RSEN            .off                //
    RegBit  PEN             .off                //
    RegBit  RCEN            .off                //
    RegBit  ACKEN           .off                //
    RegBit  ACKDT           .off                //
    RegBit  STREN           .off                //
    RegBit  GCEN            .off                //
    RegBit  SMEN            .off                //
    RegBit  DISSLW          .off                //
    RegBit  A10M            .off                //
    RegBit  IPMIEN          .off                //
    RegBit  SCLREL          .off                //
    RegBit  I2CSIDL         .off                //
    RegBit  I2CEN           .off                //
  endRegDef


  // =======================================================================
  // I2C2STAT -
  //            see "19-i2c.pdf"
  //            see i2c1 for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef I2C2STAT
    RegBit  TBF             .off                //
    RegBit  RBF             .off                //
    RegBit  R_W             .off                //
    RegBit  S               .off                //
    RegBit  P               .off                //
    RegBit  D_A             .off                //
    RegBit  I2COV           .off                //
    RegBit  IWCOL           .off                //
    RegBit  (ADD10)         .off                //
    RegBit  GCSTAT          .off                //
    RegBit  BCL             .off                //
    RegBit  TRSTAT          .off                //
    RegBit  ACKSTAT         .off                //
  endRegDef


  // =======================================================================
  // I2C3BRG
  //            see "19-i2c.pdf"
  //            see i2c1 for bit definitions

  0 constant I2C3BRG_IniVal


  // =======================================================================
  // I2C3CON -
  //            see "19-i2c.pdf"
  //            see i2c1 for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef I2C3CON
    RegBit  SEN             .off                //
    RegBit  RSEN            .off                //
    RegBit  PEN             .off                //
    RegBit  RCEN            .off                //
    RegBit  ACKEN           .off                //
    RegBit  ACKDT           .off                //
    RegBit  STREN           .off                //
    RegBit  GCEN            .off                //
    RegBit  SMEN            .off                //
    RegBit  DISSLW          .off                //
    RegBit  A10M            .off                //
    RegBit  IPMIEN          .off                //
    RegBit  SCLREL          .off                //
    RegBit  I2CSIDL         .off                //
    RegBit  I2CEN           .off                //
  endRegDef


  // =======================================================================
  // I2C3STAT -
  //            see "19-i2c.pdf"
  //            see i2c1 for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef I2C3STAT
    RegBit  TBF             .off                //
    RegBit  RBF             .off                //
    RegBit  R_W             .off                //
    RegBit  S               .off                //
    RegBit  P               .off                //
    RegBit  D_A             .off                //
    RegBit  I2COV           .off                //
    RegBit  IWCOL           .off                //
    RegBit  (ADD10)         .off                //
    RegBit  GCSTAT          .off                //
    RegBit  BCL             .off                //
    RegBit  TRSTAT          .off                //
    RegBit  ACKSTAT         .off                //
  endRegDef


  // =======================================================================
  // U1MODE : Uart 1 mode register
  //          see "17-uart.pdf"

  // STSEL : Stop selection bits
  // 0 : 1 stop bit
  // 1 : 2 stop bits

    0 0 bits ST_1
    1 0 bits ST_2

  // BRGH :
  // 0 : Low speed
  // 1 : High speed

    0 3 bits BRGH_LOW
    1 3 bits BRGH_HIGH

  // ABAUD : Auto Baud detection
  // 0 : Baud rate measurement disable or completed
  // 1 : Enable Baud Rate measurement on the next character.
  //     requires reception of a syunc fiel $55; cleared in
  //     hardware upon completion.

  // LPBACK : Loopback Mode select
  // 0 - Loopback mode is off
  // 1 - Loopback mode is on

  // WAKE : Enable wake-up on start bit detection durin Sleep mode.

  // RTSMD : RTS Mode Select
  // 0 - RTS_FLWOC   RTS in flow control mode
  // 1 - RTS_SIMPLEX RTS in simplex mode

    0 11 bits RTS_FLOWC
    1 11 bits RTS_SIMPLEX

  // IREN : IrDA encoder/decoder enable
  // 0 - IrDA mode off
  // 1 - IrDA mode on

  // USIDL : Uart Stop In Idle mode
  // 0 - Continue operation in idle mide
  // 1 - Discontinue operation when device enters idle mode

  // UARTEN :
  // 0 - UART disabled, associated pins ar port pins
  // 1 - UART enabled, UART pins controlled by UART

  // PDSEL : Parity and Data selection
  // 21
  // 00 - 8 bit no parity
  // 01 - 8 bit even parity
  // 10 - 8 bit odd parity
  // 11 - 9 bit

    %00 1 bits PDSEL_N8
    %01 1 bits PDSEL_E8
    %10 1 bits PDSEL_O8
    %11 1 bits PDSEL_N9

  // UEN : Uart Tx Enable
  // 98
  // 00 - Tx and Rx pins are enabled and used; CTS RTS and BCLK are not used
  // 01 - Tx, Rx and RTS are used; CTS is not used
  // 10 - Tx, Rx, CTS and RTS are used
  // 11 - Tx, Rx and BCLK are used; CTS is not used
  // BCLK : Baud Clock Output

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef U1MODE
            ST_1                  .use          // ST_1 or ST_2 for 1 or two stop bits
            BRGH_LOW              .use          // BRG_LOW or BRG_HIGH for low or high
                                                // baud rates
    RegBit  RXINV           .off                //
    RegBit  ABAUD           .off                // Autobaud detection
    RegBit  LPBACK          .off                // Loopback mode
    RegBit  WAKE            .off                // Wake-up on start bit detection
            RTS_FLOWC             .use          // RTS_FLOWC (default) or RTS_SIMPLEX
    RegBit  IREN            .off                // IrDA mode
    RegBit  USIDL           .off                // Disable in Idle mode (when .on)
    RegBit  UARTEN          .off                 // Uart enable
            PDSEL_N8              .use          // N8 (default), E8, O8 or N9
    RegBit  UEN0            .off                //
    RegBit  UEN1            .off                //
  endRegDef


  // =======================================================================
  // U1STA : Uart 1 Status register
  //          see "17-uart.pdf"

  // URXDA : Rx data available
  // 0 - no data available
  // 1 - FIFO holds 1 to 4 Rx characters

  // OERR : RO : when set : Receive buffer overrun error

  // FERR : RO : when set : Framing error has been detected for the current
  //                        character

  // PERR : RO : when set : Parity error has been detected for the currenr
  //                        character.

  // RIDLE : RO : when set : Receiver is idle, otherwise Rx in progress

  // ADDEN : Address Character Detect bit
  // 0 - Address detect mode disabled
  // 1 - Address detect mode on; only valid in 9 bit mode

  // TRMT : RO : when set : transmit shift register and transmit register
  //                        both empty; otherwise transmission in progress.

  // UTXBF : RO : Transmit Buffer Full status
  // 0 - Transmit buffer not full, can send at least one more char
  // 1 - Transmit buffer full

  // UTXEN : Transmit Enable
  // 0 - Tx disabled
  // 1 - Tx Enabled

  // UTXBRK : Transmit Break
  // 0 - Sync break transmissin complete or disabled
  // 1 - Start sync break transmiission (start 12 zeros stop)

  // UTXINV : Transmit polarity inversion
  // When IREN = 0 :
  //   0 - Tx idle state = '1'
  //   1 - Tx idle state = '0'
  // When IREN = 1 :
  //   0 - IrDA encoded Tx idle state = '1'
  //   1 - IrDA encoded Tx idle state = '0'

  // UTXISEL : Transmission Interrupt Mode Selection
  // fd
  // 00 - Interrupt generated when any character is transferred to the
  //      transmit shift register and the transmit buffer becones empty.
  // 01 - Interrupt generated when the last transmission is over (last
  //      char shifted out of transmit shift register) and all the tx
  //      operations are completed.
  // 10 - Interrupt generated when a character is transferred to the
  //      transmit shift register and the transmit buffer becomes empty.
  // 11 - Reserved

  // URXISEL :
  // 76
  // 0x - Rx interrupt on each rx char
  // 10 - Rx interrupt after three rx chars  (fifo 3/4)
  // 11 - Rx interrupt after four  rtx chars (fifo full)

    %00 6 bits RXI_ALL
    %10 6 bits RXI_F3/4
    %11 6 bits RXI_FFULL

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef U1STA
    RegBit  URXDA           .off                // RO, set off
    RegBit  OERR            .off                // RO, set off
    RegBit  FERR            .off                // RO, set off
    RegBit  PERR            .off                // RO, set off
    RegBit  RIDLE           .off                // RO, set off
    RegBit  ADDEN           .off                // Address character detect in 9 bit mode
    RegBit  TRMT            .off                // RO, set off
    RegBit  UTXBF           .off                // RO, set off
    RegBit  UTXEN           .off                // Tx enable
    RegBit  UTXBRK          .off                // Transmit a 14 bit break sync character
    RegBit  UTXINV          .off                // Invert Tx polarity
    RegBit  UTXISEL0        .off                // Interrupt generation mode, ee table above
    RegBit  UTXISEL1        .off                // Interrupt generation mode, ee table above
            RXI_ALL               .use          // ALL, F3/4 or FFULL
  endRegDef


  // =======================================================================
  // U1BRG : Uart 1 Buad Rate Generator register
  //          see "17-uart.pdf"
  //
  // BRGH = 0
  //   Baud Rate = Fcy / ( 16 * ( UxBRG + 1))
  //   UxBRG = Fcy / ( 16 * BaudRate) - 1
  //
  // BRGH = 1
  //   Baud Rate = Fcy / ( 4 * ( UxBRG + 1))
  //   UxBRG = Fcy / ( 4 * BaudRate) - 1

  ProcessorSpeed >f 4 / 9600 16 * 1 - /

    constant U1BRG_IniVal


  // =======================================================================
  // U2MODE : Uart 2 Mode register
  //          see "17-uart.pdf"
  //          see UART 1 for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef U2MODE
            ST_1                  .use          // ST_1 or ST_2 for 1 or two stop bits
            BRGH_LOW              .use          // BRG_LOW or BRG_HIGH for low or high
                                                // baud rates
    RegBit  RXINV           .off                //
    RegBit  ABAUD           .off                // Autobaud detection
    RegBit  LPBACK          .off                // Loopback mode
    RegBit  WAKE            .off                // Wake-up on start bit detection
            RTS_FLOWC             .use          // RTS_FLOWC (default) or RTS_SIMPLEX
    RegBit  IREN            .off                // IrDA mode
    RegBit  USIDL           .off                // Idle in sleep mode
    RegBit  UARTEN          .off                // Uart enable
            PDSEL_N8              .use          // N8 (default), E8, O8 or N9
    RegBit  UEN0            .off                //
    RegBit  UEN1            .off                //
  endRegDef


  // =======================================================================
  // U2STA : Uart 2 Status register
  //          see "17-uart.pdf"
  //          see UART 1 for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef U2STA
    RegBit  URXDA           .off                // RO, set off
    RegBit  OERR            .off                // RO, set off
    RegBit  FERR            .off                // RO, set off
    RegBit  PERR            .off                // RO, set off
    RegBit  RIDLE           .off                // RO, set off
    RegBit  ADDEN           .off                // Address character detect in 9 bit mode
    RegBit  TRMT            .off                // RO, set off
    RegBit  UTXBF           .off                // RO, set off
    RegBit  UTXEN           .off                // Tx enable
    RegBit  UTXBRK          .off                // Transmit a 14 bit break sync character
    RegBit  UTXINV          .off                // Invert Tx polarity
    RegBit  UTXISEL0        .off                // Interrupt generation mode, ee table above
    RegBit  UTXISEL1        .off                // Interrupt generation mode, ee table above
            RXI_ALL               .use          // ALL, F3/4 or FFULL
  endRegDef


  // =======================================================================
  // U2BRG : Uart 2 Baud Rate Generator register
  //          see "17-uart.pdf"
  //
  // BRGH = 0
  //   Baud Rate = Fcy / ( 16 * ( UxBRG + 1))
  //   UxBRG = Fcy / ( 16 * BaudRate) - 1
  //
  // BRGH = 1
  //   Baud Rate = Fcy / ( 4 * ( UxBRG + 1))
  //   UxBRG = Fcy / ( 4 * BaudRate) - 1
  //

  0 constant U2BRG_IniVal


  // =======================================================================
  // U3MODE : Uart 3 Mode register
  //          see "17-uart.pdf"
  //          see UART 1 for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef U3MODE
            ST_1                  .use          // ST_1 or ST_2 for 1 or two stop bits
            BRGH_LOW              .use          // BRG_LOW or BRG_HIGH for low or high
                                                // baud rates
    RegBit  RXINV           .off                //
    RegBit  ABAUD           .off                // Autobaud detection
    RegBit  LPBACK          .off                // Loopback mode
    RegBit  WAKE            .off                // Wake-up on start bit detection
            RTS_FLOWC             .use          // RTS_FLOWC (default) or RTS_SIMPLEX
    RegBit  IREN            .off                // IrDA mode
    RegBit  USIDL           .off                // Idle in sleep mode
    RegBit  UARTEN          .off                // Uart enable
            PDSEL_N8              .use          // N8 (default), E8, O8 or N9
    RegBit  UEN0            .off                //
    RegBit  UEN1            .off                //
  endRegDef


  // =======================================================================
  // U3STA : Uart 3 Status register
  //          see "17-uart.pdf"
  //          see UART 1 for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef U3STA
    RegBit  URXDA           .off                // RO, set off
    RegBit  OERR            .off                // RO, set off
    RegBit  FERR            .off                // RO, set off
    RegBit  PERR            .off                // RO, set off
    RegBit  RIDLE           .off                // RO, set off
    RegBit  ADDEN           .off                // Address character detect in 9 bit mode
    RegBit  TRMT            .off                // RO, set off
    RegBit  UTXBF           .off                // RO, set off
    RegBit  UTXEN           .off                // Tx enable
    RegBit  UTXBRK          .off                // Transmit a 14 bit break sync character
    RegBit  UTXINV          .off                // Invert Tx polarity
    RegBit  UTXISEL0        .off                // Interrupt generation mode, ee table above
    RegBit  UTXISEL1        .off                // Interrupt generation mode, ee table above
            RXI_ALL               .use          // ALL, F3/4 or FFULL
  endRegDef


  // =======================================================================
  // U3BRG : Uart 3 Baud Rate Generator register
  //          see "17-uart.pdf"
  //
  // BRGH = 0
  //   Baud Rate = Fcy / ( 16 * ( UxBRG + 1))
  //   UxBRG = Fcy / ( 16 * BaudRate) - 1
  //
  // BRGH = 1
  //   Baud Rate = Fcy / ( 4 * ( UxBRG + 1))
  //   UxBRG = Fcy / ( 4 * BaudRate) - 1
  //

  0 constant U3BRG_IniVal


  // =======================================================================
  // U4MODE : Uart 4 Mode register
  //          see "17-uart.pdf"
  //          see UART 1 for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef U4MODE
            ST_1                  .use          // ST_1 or ST_2 for 1 or two stop bits
            BRGH_LOW              .use          // BRG_LOW or BRG_HIGH for low or high
                                                // baud rates
    RegBit  RXINV           .off                //
    RegBit  ABAUD           .off                // Autobaud detection
    RegBit  LPBACK          .off                // Loopback mode
    RegBit  WAKE            .off                // Wake-up on start bit detection
            RTS_FLOWC             .use          // RTS_FLOWC (default) or RTS_SIMPLEX
    RegBit  IREN            .off                // IrDA mode
    RegBit  USIDL           .off                // Idle in sleep mode
    RegBit  UARTEN          .off                // Uart enable
            PDSEL_N8              .use          // N8 (default), E8, O8 or N9
    RegBit  UEN0            .off                //
    RegBit  UEN1            .off                //
  endRegDef


  // =======================================================================
  // U4STA : Uart 4 Status register
  //          see "17-uart.pdf"
  //          see UART 1 for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef U4STA
    RegBit  URXDA           .off                // RO, set off
    RegBit  OERR            .off                // RO, set off
    RegBit  FERR            .off                // RO, set off
    RegBit  PERR            .off                // RO, set off
    RegBit  RIDLE           .off                // RO, set off
    RegBit  ADDEN           .off                // Address character detect in 9 bit mode
    RegBit  TRMT            .off                // RO, set off
    RegBit  UTXBF           .off                // RO, set off
    RegBit  UTXEN           .off                // Tx enable
    RegBit  UTXBRK          .off                // Transmit a 14 bit break sync character
    RegBit  UTXINV          .off                // Invert Tx polarity
    RegBit  UTXISEL0        .off                // Interrupt generation mode, ee table above
    RegBit  UTXISEL1        .off                // Interrupt generation mode, ee table above
            RXI_ALL               .use          // ALL, F3/4 or FFULL
  endRegDef


  // =======================================================================
  // U4BRG : Uart 4 Baud Rate Generator register
  //          see "17-uart.pdf"
  //
  // BRGH = 0
  //   Baud Rate = Fcy / ( 16 * ( UxBRG + 1))
  //   UxBRG = Fcy / ( 16 * BaudRate) - 1
  //
  // BRGH = 1
  //   Baud Rate = Fcy / ( 4 * ( UxBRG + 1))
  //   UxBRG = Fcy / ( 4 * BaudRate) - 1
  //

  0 constant U4BRG_IniVal


  // =======================================================================
  // SPI1STAT : SPI 1 Statu register
  //            see "18-spi.pdf"

  // SPIRBF : SPI Rx full status
  // 0 - Rx not complete
  // 1 - Rx complete

  // SPITBF : SPI Tx buffer full status
  // 0 - Tx buffer free
  // 1 - Tx buffer in use

  // SPIROCV : Receive Overflow flag
  // 0 - No Rx overflow
  // 1 - Rx overflow - an Rx byte was discarded

  // SPISIDL : Stop In Idle Mode
  // 0 - Do not stop SPI in idle mode
  // 1 - Stop SPI in Idle mode

  // SPIEN : SPI Enable
  // 0 - SPI is disabled
  // 1 -  SPI is enabled

  // BUFELM : ?? undocumented

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef SPI1STAT
    RegBit  SPIRBF          .off                // RO : SPI Rx full status
    RegBit  SPITBF          .off                // RO : SPI Tx buffer full status
    RegBit  SRXMPT          .off                //
    RegBit  SPIROV          .off                // RO : Receive Overflow flag
    RegBit  SRMPT           .off                //
    RegBit  SPISIDL         .off                // Stop In Idle Mode
    RegBit  SPIEN           .off                // SPI Enable
    RegBit  SISEL0          .off                //
    RegBit  SISEL1          .off                //
    RegBit  SISEL2          .off                //
    RegBit  SPIBEC0         .off                //
    RegBit  SPIBEC1         .off                //
    RegBit  SPIBEC2         .off                //
  endRegDef


  // =======================================================================
  // SPI1CON1 : SPI 1 Control Register 1
  //            see "18-spi.pdf"

  // MSTEN : Master Mode Enable
  // 0 - Slave Mode
  // 1 - Master mode

  // CKP : Clock Polarity Select
  // 0 - Idle state for clock is high
  // 1 - Idle state for clock is low

  // SSEN : Slave Select Enable
  // 0 - SSx pin not used by SPI unit
  // 1 - SSx pin used for slave mode

  // CKE : Clock Edge Select
  // 0 - Data out change on clock idle to clock active
  // 1 - Data out xhange on clock active to clock idle

  // SMP : Data Input Sample Phase select
  // 0 - Input sampled at middle of data output time
  // 1 - Input sampled at end of data input time

  // MODE16 : Word/byte communication select mode
  // 0 - operate in byte mode (  8 bits)
  // 1 - operate in word mode ( 16 bits)

  // DISSDO : Disable SDO pin
  // 0 - SDO controlled by SPI unit
  // 1 - SDO free for other use

  // DISSCK : Disable SCK pin
  // 0 - Internal SPI clock is enabled
  // 1 - Internal SPI clock disabled, pin free for I/O

  // PPRE : Primary Prescal bits (Master mode)
  // 10
  // 00 - Primary prescale 64:1
  // 01 - Primary prescale 16:1
  // 10 - Primary prescale  4:1
  // 11 - Primary prescale  1:1

    %00 0 bits PPRE_64
    %01 0 bits PPRE_16
    %10 0 bits PPRE_4
    %11 0 bits PPRE_1

  // SPRE : Secondary Prescale bits (Master mode)

    %000 2 bits SPRE_8
    %001 2 bits SPRE_7
    %010 2 bits SPRE_6
    %011 2 bits SPRE_5
    %100 2 bits SPRE_4
    %101 2 bits SPRE_3
    %110 2 bits SPRE_2
    %111 2 bits SPRE_1

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef SPI1CON1
    RegBit  MSTEN           .off                // Master enable
    RegBit  CKP             .off                // Clock polarity
    RegBit  SSEN            .off                // Slave Select enable
    RegBit  CKE             .off                // Clock Edge select
    RegBit  SMP             .off                // Data input sample phase select
    RegBit  MODE16          .off                // .on for 16 bit mode, .off for 8 bit mode
    RegBit  DISSDO          .off                // Disable SDO pin
    RegBit  DISSCK          .off                // Disable SCK pin
            PPRE_64               .use          // Primary   prescale 64 (default), 16, 4 or 1
            SPRE_8                .use          // Secondary prescale  8 (default) .. 1
  endRegDef


  // =======================================================================
  // SPI1CON2 : SPI 1 Control register 2
  //            see "18-spi.pdf"

  // SPIFSD : Frame Sync Pulse Direction control
  // 0 - Frame sync puls output (master)
  // 1 - Frame sync pulse input (slave)

  // FRMEN : Framed SPI Support select
  // 0 - Framed SPI support disabled
  // 1 - Framed SPI support enabled (SSx oin used as frame sync I/O pin)

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef SPI1CON2
    RegBit  SPIBEN          .off                //
    RegBit  SPIFE           .off                //
    RegBit  SPIFPOL         .off                //
    RegBit  SPIFSD          .off                // Frame Sync Pulse Direction control
    RegBit  FRMEN           .off                // Framed SPI Support select
  endRegDef


  // =======================================================================
  // SPI2STAT : SPI 2 Status register
  //            see "18-spi.pdf"
  //            see SPI1 for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef SPI2STAT
    RegBit  SPIRBF          .off                // RO : SPI Rx full status
    RegBit  SPITBF          .off                // RO : SPI Tx buffer full status
    RegBit  SRXMPT          .off                //
    RegBit  SPIROV          .off                // RO : Receive Overflow flag
    RegBit  SRMPT           .off                //
    RegBit  SPISIDL         .off                // Stop In Idle Mode
    RegBit  SPIEN           .off                // SPI Enable
    RegBit  SISEL0          .off                //
    RegBit  SISEL1          .off                //
    RegBit  SISEL2          .off                //
    RegBit  SPIBEC0         .off                //
    RegBit  SPIBEC1         .off                //
    RegBit  SPIBEC2         .off                //
  endRegDef


  // =======================================================================
  // SPI2CON1 : SPI 2 Control register 1
  //            see "18-spi.pdf"
  //            see SPI1 for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef SPI2CON1
    RegBit  MSTEN           .off                // Master enable
    RegBit  CKP             .off                // Clock polarity
    RegBit  SSEN            .off                // Slave Select enable
    RegBit  CKE             .off                // Clock Edge select
    RegBit  SMP             .off                // Data input sample phase select
    RegBit  MODE16          .off                // .on for 16 bit mode, .off for 8 bit mode
    RegBit  DISSDO          .off                // Disable SDO pin
    RegBit  DISSCK          .off                // Disable SCK pin
    RegBit  PPRE_64               .use          // Primary   prescale 64 (default), 16, 4 or 1
    RegBit  SPRE_8                .use          // Secondary prescale 8 (default) .. 1
  endRegDef


  // =======================================================================
  // SPI2CON2 : SPI 2 Control register 2
  //            see "18-spi.pdf"
  //            see SPI1 for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef SPI2CON2
    RegBit  SPIBEN          .off                //
    RegBit  SPIFE           .off                //
    RegBit  SPIFPOL         .off                //
    RegBit  SPIFSD          .off                // Frame Sync Pulse Direction control
    RegBit  FRMEN           .off                // Framed SPI Support select
  endRegDef


  // =======================================================================
  // SPI3STAT : SPI 3 Status register
  //            see "18-spi.pdf"
  //            see SPI1 for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef SPI3STAT
    RegBit  SPIRBF          .off                // RO : SPI Rx full status
    RegBit  SPITBF          .off                // RO : SPI Tx buffer full status
    RegBit  SRXMPT          .off                //
    RegBit  SPIROV          .off                // RO : Receive Overflow flag
    RegBit  SRMPT           .off                //
    RegBit  SPISIDL         .off                // Stop In Idle Mode
    RegBit  SPIEN           .off                // SPI Enable
    RegBit  SISEL0          .off                //
    RegBit  SISEL1          .off                //
    RegBit  SISEL2          .off                //
    RegBit  SPIBEC0         .off                //
    RegBit  SPIBEC1         .off                //
    RegBit  SPIBEC2         .off                //
  endRegDef


  // =======================================================================
  // SPI3CON1 : SPI 3 Control register 1
  //            see "18-spi.pdf"
  //            see SPI1 for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef SPI3CON1
    RegBit  MSTEN           .off                // Master enable
    RegBit  CKP             .off                // Clock polarity
    RegBit  SSEN            .off                // Slave Select enable
    RegBit  CKE             .off                // Clock Edge select
    RegBit  SMP             .off                // Data input sample phase select
    RegBit  MODE16          .off                // .on for 16 bit mode, .off for 8 bit mode
    RegBit  DISSDO          .off                // Disable SDO pin
    RegBit  DISSCK          .off                // Disable SCK pin
    RegBit  PPRE_64               .use          // Primary   prescale 64 (default), 16, 4 or 1
    RegBit  SPRE_8                .use          // Secondary prescale 8 (default) .. 1
  endRegDef


  // =======================================================================
  // SPI3CON2 : SPI 3 Control register 2
  //            see "18-spi.pdf"
  //            see SPI1 for bit definitions

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef SPI3CON2
    RegBit  SPIBEN          .off                //
    RegBit  SPIFE           .off                //
    RegBit  SPIFPOL         .off                //
    RegBit  SPIFSD          .off                // Frame Sync Pulse Direction control
    RegBit  FRMEN           .off                // Framed SPI Support select
  endRegDef


  // ----- PADCFG1 Bits -----------------------------------------------------

  regDef PADCFG1
    RegBit  RTSECSEL        .off                //
  endRegDef


  // =======================================================================
  // AD1CON1 - ADC 1 Control register 1
  //          see "16-adc.pdf"

  // DONE : ADC conversion status bit, set when done
  //        can be cleared in SW, but that has no effect on module.

  // SAMP : ADC Sample Enable
  // 0 - S&H is holding
  // 1 - S&H is soampling
  // When ASM = 0 software can starrt sampling by setting a 1
  // When ASM = 1 it's aute set by hardware
  // When SSRC = 000 software can write 0 to end sampling and start
  // conversion
  // When SSRC <> 000 auto cleared by hardware when sampling ends
  // and conversion starts.

  // ASAM : Auto Sample
  // 0 - Sampling begins when SAMP bit is set
  // 1 - Sampling begins imediately after last conversion. SAMP is auto
  //     set.

  // AD12B : 10 or 12 bit operation mode bit
  // 0 - 10 bit 4 channel ADC operation
  // 1 - 12 bit 1 channel ADC operation

  // ADSIDL : Stop In Idle Mode
  // 0 - do not stop module in idle mode
  // 1 - stop module operation in idle mode

  // ADON : ADC operating mode bit
  // 0 - ADC os off
  // 1 - ADC is operating

  // SSRC : Sample Clock Source Select
  // 765
  // 000 - Clearing sample bit ends sampling and starts conversion
  // 001 - Active transition on INTx pin ends sampling and starts
  //       conversion
  // 010 - GP timer (timer 3 for ADC1, timer 5 for ADC2) compare
  //       ends sampling and starts conversion
  // 011 - Reserved
  // 100 - Reserved
  // 101 - Reserved
  // 110 - Reserved
  // 111 - Internal counter ends sampling and starts conversion


    %000 5 bits SSRC_AUTO
    %001 5 bits SSRC_EXT_INT
    %010 5 bits SSRC_GP_TIMER
    %111 5 bits SSRC_INT_COUNTER

  // FORM : Data Output Format
  // 98 -  10 bit operation - 12 bit operation
  // 00 -  Integer          - integer
  // 01 -  Signed int       - signed int
  // 10 -  reserved         - fractional
  // 11 -  reserved         - signed fractional

    %00 8 bits FORM_UINT
    %01 8 bits FORM_SINT

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef AD1CON1
    RegBit  DONE            .off                //
    RegBit  SAMP            .off                //
    RegBit  ASAM            .off                //
    RegBit  ADSIDL          .off                //
    RegBit  ADON            .off                //
            SSRC_AUTO             .use          //
            FORM_UINT             .use          //
  endRegDef


  // =======================================================================
  // AD1CON2 - ADC 1 Control register 2
  //          see "16-adc.pdf"

  // ALTS : Alternate Input Sample Mode
  // 0 - Always use channel input selects for sample A
  // 1 - Uses channel input selects for sample A on first sample and
  //     sample B on next sample.

  // BUFM : Buffer fill mode select
  // 0 - Always start filling the buffer from the start address
  // 1 - Starts buffer filling the first half of the buffer on the
  //     first interrupt and the 2nd half of the buffer on the next
  //     interrupt.

  // BUFS : Buffer Fill Status bit
  //        only valid when BUFM = 1
  // 0 - ADC is currently filling the first half of the buffer. The
  //     user application should access data in the 2nd half of the buffer.
  // 1 - ADC is currently filling the 2nd half of the buffer. The user
  //     application should access data in the first half of the buffer.

  // CSNA : Input Scan Select
  // 0 - Do not scan inputs
  // 1 - Scan inputs for CH0+ during sample A bit

  // VCFG : Voltage Reference Configuration
  // fed
  // 000 - VrefH = Avdd , VrefL = Avss
  // 001 - VrefH = Vref+, VrefL = Avss
  // 010 - VrefH = Avdd , VrefL = Vref-
  // 011 - VrefH = Vref+, VrefL = Vref-
  // 100 - reserved
  // 101 - reserved
  // 110 - reserved
  // 111 - reserved

    %000 13 bits VCFG_AVDD_AVSS
    %001 13 bits VCFG_AVREF+_AVSS
    %010 13 bits VCFG_AVDD_AVREF-
    %011 13 bits VCFG_AVREF+_AVREF-


  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef AD1CON2
    RegBit  ALTS            .off                //
    RegBit  BUFM            .off                //
    RegBit  BUFS            .off                //
    RegBit  CSCNA           .off                //

    RegBit  SMPI0           .off                //
    RegBit  SMPI1           .off                //
    RegBit  SMPI2           .off                //
    RegBit  SMPI3           .off                //
    RegBit  SMPI4           .off                //

            VCFG_AVDD_AVSS        .use
  endRegDef


  // =======================================================================
  // AD1CON3 - ADC 1 Control register 3
  //          see "16-adc.pdf"

  // ADRC : AD Conversion Clock Source
  // 0 - Clock derived from system clock
  // 1 - ADC internal RC clock

  // ADCS : ADC Conversion Clock Select
  // 76543210
  // 00000000 - Tad =   1 * Tcy
  // ....
  // 11111111 - Tad = 256 * Tcy

    {
      : ADCS ( n -- n )
        1 -
        $ff and
      ;
    }

  // SAM : Auto sample time
  // cba98
  // 00000 -  0 * Tad
  // ....
  // 11111 - 31 * Tad

    {
      : SAMS ( n -- n )
        $1f and
        8 <<
      ;
    }


  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef AD1CON3
    RegBit  ADRC            .off                //
            1 ADCS                .use          // Tad = n * Tcy ( 1 <= n <= 256)
            0 SAMS                .use          // Tsa = n * Tad ( 0 <= n <= 31)
  endRegDef


  // =======================================================================
  // AD1CHS0 - ADC 1 input channel 0 register
  //          see "16-adc.pdf"

  // CH0NA : CH0 -input select for sample A bit
  // 0 - CH0 -input = Vrefl
  // 1 - CH0 -input = AN1

  // CH0NB : CH0 -input select for sample B bit
  // 0 - CH0 -input = Vrefl
  // 1 - CH0 -input = AN1

  // CH0SA : CH0 +input select for sample A
  // 43210
  // 00000 - CH0 +input = AN0
  // ....
  // 11111 - CH0 +input = AN31

    {
      : CH0SA ( n -- n )
        $1f and
      ;
    }

  // CH0SB : CH0 +input select for sample B
  // cba98
  // 00000 - CH0 +input = AN0
  // ....
  // 11111 - CH0 +input = AN31

    {
      : CH0SB ( n -- n )
        $1f and
        8 <<
      ;
    }

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef AD1CHS0
    RegBit  CH0NA           .off                //
    RegBit  CH0NB           .off                //
            0 CH0SA               .use          // 0 .. 31
            0 CH0SB               .use          // 0 .. 31
  endRegDef


  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef AD1CSSH
    RegBit  CSSL24          .off                //
    RegBit  CSSL25          .off                //
    RegBit  CSSL26          .off                //
    RegBit  CSSL27          .off                //
  endRegDef


  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef AD1CSSL
    RegBit  CSSL0           .off                //
    RegBit  CSSL1           .off                //
    RegBit  CSSL2           .off                //
    RegBit  CSSL3           .off                //
    RegBit  CSSL4           .off                //
    RegBit  CSSL5           .off                //
    RegBit  CSSL6           .off                //
    RegBit  CSSL7           .off                //
    RegBit  CSSL8           .off                //
    RegBit  CSSL9           .off                //
    RegBit  CSSL10          .off                //
    RegBit  CSSL11          .off                //
    RegBit  CSSL12          .off                //
    RegBit  CSSL13          .off                //
    RegBit  CSSL14          .off                //
    RegBit  CSSL15          .off                //
  endRegDef


  // ----- CTMUCON Bits -----------------------------------------------------
  regDef CTMUCON
    RegBit  EDG1STAT        .off                //
    RegBit  EDG2STAT        .off                //
    RegBit  EDG1POL         .off                //
    RegBit  EDG2POL         .off                //
    RegBit  CTTRIG          .off                //
    RegBit  IDISSEN         .off                //
    RegBit  EDGSEQEN        .off                //
    RegBit  EDGEN           .off                //
    RegBit  TGEN            .off                //
    RegBit  CTMUSIDL        .off                //
    RegBit  CTMUEN          .off                //

    RegBit  EDG1SEL0        .off                //
    RegBit  EDG1SEL1        .off                //
    RegBit  EDG2SEL0        .off                //
    RegBit  EDG2SEL1        .off                //
  endRegDef
  //
  // ----- CTMUICON Bits -----------------------------------------------------
  regDef CTMUICON
    RegBit  IRNG0           .off                //
    RegBit  IRNG1           .off                //
    RegBit  ITRIM0          .off                //
    RegBit  ITRIM1          .off                //
    RegBit  ITRIM2          .off                //
    RegBit  ITRIM3          .off                //
    RegBit  ITRIM4          .off                //
    RegBit  ITRIM5          .off                //
  endRegDef
  //
  // ----- U1OTGIR Bits -----------------------------------------------------
  regDef U1OTGIR
    RegBit  VBUSVDIF        .off                //
    RegBit  SESENDIF        .off                //
    RegBit  SESVDIF         .off                //
    RegBit  ACTVIF          .off                //
    RegBit  LSTATEIF        .off                //
    RegBit  T1MSECIF        .off                //
    RegBit  IDIF            .off                //
  endRegDef
  //
  // ----- U1OTGIE Bits -----------------------------------------------------
  regDef U1OTGIE
    RegBit  VBUSVDIE        .off                //
    RegBit  SESENDIE        .off                //
    RegBit  SESVDIE         .off                //
    RegBit  ACTVIE          .off                //
    RegBit  LSTATEIE        .off                //
    RegBit  T1MSECIE        .off                //
    RegBit  IDIE            .off                //
  endRegDef
  //
  // ----- U1OTGSTAT Bits -----------------------------------------------------
  regDef U1OTGSTAT
    RegBit  VBUSVD          .off                //
    RegBit  SESEND          .off                //
    RegBit  SESVD           .off                //
    RegBit  LSTATE          .off                //
    RegBit  ID              .off                //
  endRegDef
  //
  // ----- U1OTGCON Bits -----------------------------------------------------
  regDef U1OTGCON
    RegBit  VBUSDIS         .off                //
    RegBit  VBUSCHG         .off                //
    RegBit  OTGEN           .off                //
    RegBit  VBUSON          .off                //
    RegBit  DMPULDWN        .off                //
    RegBit  DPPULDWN        .off                //
    RegBit  DMPULUP         .off                //
    RegBit  DPPULUP         .off                //
  endRegDef
  //
  // ----- U1PWRC Bits -----------------------------------------------------
  regDef U1PWRC
    RegBit  USBPWR          .off                //
    RegBit  USUSPND         .off                //
    RegBit  USLPGRD         .off                //
    RegBit  UACTPND         .off                //
  endRegDef
  //
  // ----- U1IR Bits -----------------------------------------------------
  regDef U1IR
    RegBit  URSTIF          .off                //
    RegBit  UERRIF          .off                //
    RegBit  SOFIF           .off                //
    RegBit  TRNIF           .off                //
    RegBit  IDLEIF          .off                //
    RegBit  RESUMEIF        .off                //
    RegBit  ATTACHIF        .off                //
    RegBit  STALLIF         .off                //
  endRegDef
  //
  // ----- U1IE Bits -----------------------------------------------------
  regDef U1IE
    RegBit  URSTIE          .off                //
    RegBit  UERRIE          .off                //
    RegBit  SOFIE           .off                //
    RegBit  TRNIE           .off                //
    RegBit  IDLEIE          .off                //
    RegBit  RESUMEIE        .off                //
    RegBit  ATTACHIE        .off                //
    RegBit  STALLIE         .off                //
  endRegDef
  //
  // ----- U1EIR Bits -----------------------------------------------------
  regDef U1EIR
    RegBit  PIDEF           .off                //
    RegBit  CRC5EF          .off                //
    RegBit  CRC16EF         .off                //
    RegBit  DFN8EF          .off                //
    RegBit  BTOEF           .off                //
    RegBit  DMAEF           .off                //
    RegBit  BTSEF           .off                //
  endRegDef
  //
  // ----- U1EIE Bits -----------------------------------------------------
  regDef U1EIE
    RegBit  PIDEE           .off                //
    RegBit  CRC5EE          .off                //
    RegBit  CRC16EE         .off                //
    RegBit  DFN8EE          .off                //
    RegBit  BTOEE           .off                //
    RegBit  DMAEE           .off                //
    RegBit  BTSEE           .off                //
  endRegDef
  //
  // ----- U1STAT Bits -----------------------------------------------------
  regDef U1STAT
    RegBit  PPBI            .off                //
    RegBit  DIR             .off                //

    RegBit  ENDPT0          .off                //
    RegBit  ENDPT1          .off                //
    RegBit  ENDPT2          .off                //
    RegBit  ENDPT3          .off                //
  endRegDef
  //
  // ----- U1CON Bits -----------------------------------------------------
  regDef U1CON
    RegBit  USBEN           .off                //
    RegBit  PPBRST          .off                //
    RegBit  RESUME          .off                //
    RegBit  HOSTEN          .off                //
    RegBit  USBRST          .off                //
    RegBit  PKTDIS          .off                //
    RegBit  SE0             .off                //
    RegBit  JSTATE          .off                //
  endRegDef
  //
  // ----- U1ADDR Bits -----------------------------------------------------
  regDef U1ADDR
    RegBit  LOWSPDEN        .off                //

    RegBit  DEVADDR0        .off                //
    RegBit  DEVADDR1        .off                //
    RegBit  DEVADDR2        .off                //
    RegBit  DEVADDR3        .off                //
    RegBit  DEVADDR4        .off                //
    RegBit  DEVADDR5        .off                //
    RegBit  DEVADDR6        .off                //
  endRegDef
  //
  // ----- U1BDTP1 Bits -----------------------------------------------------
  regDef U1BDTP1
    RegBit  BDTPTRL0        .off                //
    RegBit  BDTPTRL1        .off                //
    RegBit  BDTPTRL2        .off                //
    RegBit  BDTPTRL3        .off                //
    RegBit  BDTPTRL4        .off                //
    RegBit  BDTPTRL5        .off                //
    RegBit  BDTPTRL6        .off                //
  endRegDef
  //
  // ----- U1FRML Bits -----------------------------------------------------
  regDef U1FRML
    RegBit  FRM0            .off                //
    RegBit  FRM1            .off                //
    RegBit  FRM2            .off                //
    RegBit  FRM3            .off                //
    RegBit  FRM4            .off                //
    RegBit  FRM5            .off                //
    RegBit  FRM6            .off                //
    RegBit  FRM7            .off                //
  endRegDef
  //
  // ----- U1FRMH Bits -----------------------------------------------------
  regDef U1FRMH
    RegBit  FRM8            .off                //
    RegBit  FRM9            .off                //
    RegBit  FRM10           .off                //
  endRegDef
  //
  // ----- U1TOK Bits -----------------------------------------------------
  regDef U1TOK
    RegBit  EP0             .off                //
    RegBit  EP1             .off                //
    RegBit  EP2             .off                //
    RegBit  EP3             .off                //
    RegBit  PID0            .off                //
    RegBit  PID1            .off                //
    RegBit  PID2            .off                //
    RegBit  PID3            .off                //
  endRegDef
  //
  // ----- U1SOF Bits -----------------------------------------------------
  regDef U1SOF
    RegBit  CNT0            .off                //
    RegBit  CNT1            .off                //
    RegBit  CNT2            .off                //
    RegBit  CNT3            .off                //
    RegBit  CNT4            .off                //
    RegBit  CNT5            .off                //
    RegBit  CNT6            .off                //
    RegBit  CNT7            .off                //
  endRegDef
  //
  // ----- U1CNFG1 Bits -----------------------------------------------------
  regDef U1CNFG1
    RegBit  USBSIDL         .off                //
    RegBit  UOEMON          .off                //
    RegBit  UTEYE           .off                //

    RegBit  PPB0            .off                //
    RegBit  PPB1            .off                //
  endRegDef
  //
  // ----- U1CNFG2 Bits -----------------------------------------------------
  regDef U1CNFG2
    RegBit  UTRDIS          .off                //
    RegBit  UVCMPDIS        .off                //
    RegBit  UVBUSDIS        .off                //
    RegBit  EXTI2CEN        .off                //
    RegBit  PUVBUS          .off                //
    RegBit  UVCMPSEL        .off                //
  endRegDef
  //
  // ----- U1EP0 Bits -----------------------------------------------------
  regDef U1EP0
    RegBit  EPHSHK          .off                //
    RegBit  EPSTALL         .off                //
    RegBit  EPTXEN          .off                //
    RegBit  EPRXEN          .off                //
    RegBit  EPCONDIS        .off                //
    RegBit  RETRYDIS        .off                //
    RegBit  LSPD            .off                //

    RegBit  EPINEN          .off                //
    RegBit  EPOUTEN         .off                //
    RegBit  LOWSPD          .off                //
  endRegDef
  //
  // ----- U1EP1 Bits -----------------------------------------------------
  regDef U1EP1
    RegBit  EPHSHK          .off                //
    RegBit  EPSTALL         .off                //
    RegBit  EPTXEN          .off                //
    RegBit  EPRXEN          .off                //
    RegBit  EPCONDIS        .off                //

    RegBit  EPOUTEN         .off                //
    RegBit  LOWSPD          .off                //
  endRegDef
  //
  // ----- U1EP2 Bits -----------------------------------------------------
  regDef U1EP2
    RegBit  EPHSHK          .off                //
    RegBit  EPSTALL         .off                //
    RegBit  EPTXEN          .off                //
    RegBit  EPRXEN          .off                //
    RegBit  EPCONDIS        .off                //

    RegBit  EPOUTEN         .off                //
    RegBit  LOWSPD          .off                //
  endRegDef
  //
  // ----- U1EP3 Bits -----------------------------------------------------
  regDef U1EP3
    RegBit  EPHSHK          .off                //
    RegBit  EPSTALL         .off                //
    RegBit  EPTXEN          .off                //
    RegBit  EPRXEN          .off                //
    RegBit  EPCONDIS        .off                //

    RegBit  EPOUTEN         .off                //
    RegBit  LOWSPD          .off                //
  endRegDef
  //
  // ----- U1EP4 Bits -----------------------------------------------------
  regDef U1EP4
    RegBit  EPHSHK          .off                //
    RegBit  EPSTALL         .off                //
    RegBit  EPTXEN          .off                //
    RegBit  EPRXEN          .off                //
    RegBit  EPCONDIS        .off                //

    RegBit  EPOUTEN         .off                //
    RegBit  LOWSPD          .off                //
  endRegDef
  //
  // ----- U1EP5 Bits -----------------------------------------------------
  regDef U1EP5
    RegBit  EPHSHK          .off                //
    RegBit  EPSTALL         .off                //
    RegBit  EPTXEN          .off                //
    RegBit  EPRXEN          .off                //
    RegBit  EPCONDIS        .off                //

    RegBit  EPOUTEN         .off                //
    RegBit  LOWSPD          .off                //
  endRegDef
  //
  // ----- U1EP6 Bits -----------------------------------------------------
  regDef U1EP6
    RegBit  EPHSHK          .off                //
    RegBit  EPSTALL         .off                //
    RegBit  EPTXEN          .off                //
    RegBit  EPRXEN          .off                //
    RegBit  EPCONDIS        .off                //

    RegBit  EPOUTEN         .off                //
    RegBit  LOWSPD          .off                //
  endRegDef
  //
  // ----- U1EP7 Bits -----------------------------------------------------
  regDef U1EP7
    RegBit  EPHSHK          .off                //
    RegBit  EPSTALL         .off                //
    RegBit  EPTXEN          .off                //
    RegBit  EPRXEN          .off                //
    RegBit  EPCONDIS        .off                //

    RegBit  EPOUTEN         .off                //
    RegBit  LOWSPD          .off                //
  endRegDef
  //
  // ----- U1EP8 Bits -----------------------------------------------------
  regDef U1EP8
    RegBit  EPHSHK          .off                //
    RegBit  EPSTALL         .off                //
    RegBit  EPTXEN          .off                //
    RegBit  EPRXEN          .off                //
    RegBit  EPCONDIS        .off                //

    RegBit  EPOUTEN         .off                //
    RegBit  LOWSPD          .off                //
  endRegDef
  //
  // ----- U1EP9 Bits -----------------------------------------------------
  regDef U1EP9
    RegBit  EPHSHK          .off                //
    RegBit  EPSTALL         .off                //
    RegBit  EPTXEN          .off                //
    RegBit  EPRXEN          .off                //
    RegBit  EPCONDIS        .off                //

    RegBit  EPOUTEN         .off                //
    RegBit  LOWSPD          .off                //
  endRegDef
  //
  // ----- U1EP10 Bits -----------------------------------------------------
  regDef U1EP10
    RegBit  EPHSHK          .off                //
    RegBit  EPSTALL         .off                //
    RegBit  EPTXEN          .off                //
    RegBit  EPRXEN          .off                //
    RegBit  EPCONDIS        .off                //

    RegBit  EPOUTEN         .off                //
    RegBit  LOWSPD          .off                //
  endRegDef
  //
  // ----- U1EP11 Bits -----------------------------------------------------
  regDef U1EP11
    RegBit  EPHSHK          .off                //
    RegBit  EPSTALL         .off                //
    RegBit  EPTXEN          .off                //
    RegBit  EPRXEN          .off                //
    RegBit  EPCONDIS        .off                //

    RegBit  EPOUTEN         .off                //
    RegBit  LOWSPD          .off                //
  endRegDef
  //
  // ----- U1EP12 Bits -----------------------------------------------------
  regDef U1EP12
    RegBit  EPHSHK          .off                //
    RegBit  EPSTALL         .off                //
    RegBit  EPTXEN          .off                //
    RegBit  EPRXEN          .off                //
    RegBit  EPCONDIS        .off                //

    RegBit  EPOUTEN         .off                //
    RegBit  LOWSPD          .off                //
  endRegDef
  //
  // ----- U1EP13 Bits -----------------------------------------------------
  regDef U1EP13
    RegBit  EPHSHK          .off                //
    RegBit  EPSTALL         .off                //
    RegBit  EPTXEN          .off                //
    RegBit  EPRXEN          .off                //
    RegBit  EPCONDIS        .off                //

    RegBit  EPOUTEN         .off                //
    RegBit  LOWSPD          .off                //
  endRegDef
  //
  // ----- U1EP14 Bits -----------------------------------------------------
  regDef U1EP14
    RegBit  EPHSHK          .off                //
    RegBit  EPSTALL         .off                //
    RegBit  EPTXEN          .off                //
    RegBit  EPRXEN          .off                //
    RegBit  EPCONDIS        .off                //

    RegBit  EPOUTEN         .off                //
    RegBit  LOWSPD          .off                //
  endRegDef
  //
  // ----- U1EP15 Bits -----------------------------------------------------
  regDef U1EP15
    RegBit  EPHSHK          .off                //
    RegBit  EPSTALL         .off                //
    RegBit  EPTXEN          .off                //
    RegBit  EPRXEN          .off                //
    RegBit  EPCONDIS        .off                //

    RegBit  EPOUTEN         .off                //
    RegBit  LOWSPD          .off                //
  endRegDef
  //
  // ----- U1PWMRRS Bits -----------------------------------------------------
  regDef U1PWMRRS
    RegBit  PER0            .off                //
    RegBit  PER1            .off                //
    RegBit  PER2            .off                //
    RegBit  PER3            .off                //
    RegBit  PER4            .off                //
    RegBit  PER5            .off                //
    RegBit  PER6            .off                //
    RegBit  PER7            .off                //
    RegBit  (DC0)           .off                //
    RegBit  (DC1)           .off                //
    RegBit  (DC2)           .off                //
    RegBit  (DC3)           .off                //
    RegBit  (DC4)           .off                //
    RegBit  (DC5)           .off                //
    RegBit  (DC6)           .off                //
    RegBit  (DC7)           .off                //
  endRegDef
  //
  // ----- U1PWMCON Bits -----------------------------------------------------
  regDef U1PWMCON
    RegBit  CNTEN           .off                //
    RegBit  PWMPOL          .off                //
    RegBit  PWMEN           .off                //
  endRegDef
  //
  // ----- ANCFG Bits -----------------------------------------------------
  regDef ANCFG
    RegBit  VBGEN           .off                //
    RegBit  VBG2EN          .off                //
    RegBit  VBG6EN          .off                //
  endRegDef
  //
  // ----- ANSB Bits -----------------------------------------------------
  regDef ANSB
    RegBit  ANSB0           .off                //
    RegBit  ANSB1           .off                //
    RegBit  ANSB2           .off                //
    RegBit  ANSB3           .off                //
    RegBit  ANSB4           .off                //
    RegBit  ANSB5           .off                //
    RegBit  ANSB6           .off                //
    RegBit  ANSB7           .off                //
    RegBit  ANSB8           .off                //
    RegBit  ANSB9           .off                //
    RegBit  ANSB10          .off                //
    RegBit  ANSB11          .off                //
    RegBit  ANSB12          .off                //
    RegBit  ANSB13          .off                //
    RegBit  ANSB14          .off                //
    RegBit  ANSB15          .off                //
  endRegDef
  //
  // ----- ANSC Bits -----------------------------------------------------
  regDef ANSC
    RegBit  ANSC13          .off                //
    RegBit  ANSC14          .off                //
  endRegDef
  //
  // ----- ANSD Bits -----------------------------------------------------
  regDef ANSD
    RegBit  ANSD6           .off                //
    RegBit  ANSD7           .off                //
  endRegDef
  //
  // ----- ANSF Bits -----------------------------------------------------
  regDef ANSF
    RegBit  ANSF0           .off                //
  endRegDef
  //
  // ----- ANSG Bits -----------------------------------------------------
  regDef ANSG
    RegBit  ANSG6           .off                //
    RegBit  ANSG7           .off                //
    RegBit  ANSG8           .off                //
    RegBit  ANSG9           .off                //
  endRegDef
  //
  // ----- ALCFGRPT Bits -----------------------------------------------------
  regDef ALCFGRPT
    RegBit  CHIME           .off                //
    RegBit  ALRMEN          .off                //

    RegBit  ARPT0           .off                //
    RegBit  ARPT1           .off                //
    RegBit  ARPT2           .off                //
    RegBit  ARPT3           .off                //
    RegBit  ARPT4           .off                //
    RegBit  ARPT5           .off                //
    RegBit  ARPT6           .off                //
    RegBit  ARPT7           .off                //
    RegBit  ALRMPTR0        .off                //
    RegBit  ALRMPTR1        .off                //
    RegBit  AMASK0          .off                //
    RegBit  AMASK1          .off                //
    RegBit  AMASK2          .off                //
    RegBit  AMASK3          .off                //
  endRegDef
  //
  // ----- RCFGCAL Bits -----------------------------------------------------
  regDef RCFGCAL
    RegBit  RTCOE           .off                //
    RegBit  HALFSEC         .off                //
    RegBit  RTCSYNC         .off                //
    RegBit  RTCWREN         .off                //
    RegBit  RTCEN           .off                //

    RegBit  CAL0            .off                //
    RegBit  CAL1            .off                //
    RegBit  CAL2            .off                //
    RegBit  CAL3            .off                //
    RegBit  CAL4            .off                //
    RegBit  CAL5            .off                //
    RegBit  CAL6            .off                //
    RegBit  CAL7            .off                //
    RegBit  RTCPTR0         .off                //
    RegBit  RTCPTR1         .off                //
  endRegDef
  //
  // ----- CMSTAT Bits -----------------------------------------------------
  regDef CMSTAT
    RegBit  C1OUT           .off                //
    RegBit  C2OUT           .off                //
    RegBit  C3OUT           .off                //
    RegBit  C1EVT           .off                //
    RegBit  C2EVT           .off                //
    RegBit  C3EVT           .off                //
    RegBit  CMIDL           .off                //
  endRegDef
  //
  // ----- CVRCON Bits -----------------------------------------------------
  regDef CVRCON
    RegBit  CVRSS           .off                //
    RegBit  CVRR            .off                //
    RegBit  CVROE           .off                //
    RegBit  CVREN           .off                //
    RegBit  CVREFP          .off                //

    RegBit  CVR0            .off                //
    RegBit  CVR1            .off                //
    RegBit  CVR2            .off                //
    RegBit  CVR3            .off                //
    RegBit  CVREFM0         .off                //
    RegBit  CVREFM1         .off                //
  endRegDef
  //
  // ----- CM1CON Bits -----------------------------------------------------
  regDef CM1CON
    RegBit  CREF            .off                //
    RegBit  COUT            .off                //
    RegBit  CEVT            .off                //
    RegBit  CPOL            .off                //
    RegBit  COE             .off                //
    RegBit  CON             .off                //

    RegBit  CCH0            .off                //
    RegBit  CCH1            .off                //
    RegBit  EVPOL0          .off                //
    RegBit  EVPOL1          .off                //
  endRegDef
  //
  // ----- CM2CON Bits -----------------------------------------------------
  regDef CM2CON
    RegBit  CREF            .off                //
    RegBit  COUT            .off                //
    RegBit  CEVT            .off                //
    RegBit  CPOL            .off                //
    RegBit  COE             .off                //
    RegBit  CON             .off                //

    RegBit  CCH0            .off                //
    RegBit  CCH1            .off                //
    RegBit  EVPOL0          .off                //
    RegBit  EVPOL1          .off                //
  endRegDef
  //
  // ----- CM3CON Bits -----------------------------------------------------
  regDef CM3CON
    RegBit  CREF            .off                //
    RegBit  COUT            .off                //
    RegBit  CEVT            .off                //
    RegBit  CPOL            .off                //
    RegBit  COE             .off                //
    RegBit  CON             .off                //

    RegBit  CCH0            .off                //
    RegBit  CCH1            .off                //
    RegBit  EVPOL0          .off                //
    RegBit  EVPOL1          .off                //
  endRegDef
  //
  // ----- CRCCON1 Bits -----------------------------------------------------
  regDef CRCCON1
    RegBit  LENDIAN         .off                //
    RegBit  CRCGO           .off                //
    RegBit  CRCISEL         .off                //
    RegBit  CRCMPT          .off                //
    RegBit  CRCFUL          .off                //
    RegBit  CSIDL           .off                //
    RegBit  CRCEN           .off                //

    RegBit  VWORD0          .off                //
    RegBit  VWORD1          .off                //
    RegBit  VWORD2          .off                //
    RegBit  VWORD3          .off                //
    RegBit  VWORD4          .off                //
  endRegDef
  //
  // ----- CRCCON2 Bits -----------------------------------------------------
  regDef CRCCON2
    RegBit  PLEN0           .off                //
    RegBit  PLEN1           .off                //
    RegBit  PLEN2           .off                //
    RegBit  PLEN3           .off                //
    RegBit  PLEN4           .off                //
    RegBit  DWIDTH0         .off                //
    RegBit  DWIDTH1         .off                //
    RegBit  DWIDTH2         .off                //
    RegBit  DWIDTH3         .off                //
    RegBit  DWIDTH4         .off                //
  endRegDef
  //
  // ----- CRCXORL Bits -----------------------------------------------------
  regDef CRCXORL
    RegBit  X1              .off                //
    RegBit  X2              .off                //
    RegBit  X3              .off                //
    RegBit  X4              .off                //
    RegBit  X5              .off                //
    RegBit  X6              .off                //
    RegBit  X7              .off                //
    RegBit  X8              .off                //
    RegBit  X9              .off                //
    RegBit  X10             .off                //
    RegBit  X11             .off                //
    RegBit  X12             .off                //
    RegBit  X13             .off                //
    RegBit  X14             .off                //
    RegBit  X15             .off                //
  endRegDef
  //
  // ----- CRCXORH Bits -----------------------------------------------------
  regDef CRCXORH
    RegBit  X16             .off                //
    RegBit  X17             .off                //
    RegBit  X18             .off                //
    RegBit  X19             .off                //
    RegBit  X20             .off                //
    RegBit  X21             .off                //
    RegBit  X22             .off                //
    RegBit  X23             .off                //
    RegBit  X24             .off                //
    RegBit  X25             .off                //
    RegBit  X26             .off                //
    RegBit  X27             .off                //
    RegBit  X28             .off                //
    RegBit  X29             .off                //
    RegBit  X30             .off                //
    RegBit  X31             .off                //
  endRegDef
  //
  // ----- G1CMDL Bits -----------------------------------------------------
  regDef G1CMDL
    RegBit  GCMD0           .off                //
    RegBit  GCMD1           .off                //
    RegBit  GCMD2           .off                //
    RegBit  GCMD3           .off                //
    RegBit  GCMD4           .off                //
    RegBit  GCMD5           .off                //
    RegBit  GCMD6           .off                //
    RegBit  GCMD7           .off                //
    RegBit  GCMD8           .off                //
    RegBit  GCMD9           .off                //
    RegBit  GCMD10          .off                //
    RegBit  GCMD11          .off                //
    RegBit  GCMD12          .off                //
    RegBit  GCMD13          .off                //
    RegBit  GCMD14          .off                //
    RegBit  GCMD15          .off                //
  endRegDef
  //
  // ----- G1CMDH Bits -----------------------------------------------------
  regDef G1CMDH
    RegBit  GCMD16          .off                //
    RegBit  GCMD17          .off                //
    RegBit  GCMD18          .off                //
    RegBit  GCMD19          .off                //
    RegBit  GCMD20          .off                //
    RegBit  GCMD21          .off                //
    RegBit  GCMD22          .off                //
    RegBit  GCMD23          .off                //
    RegBit  GCMD24          .off                //
    RegBit  GCMD25          .off                //
    RegBit  GCMD26          .off                //
    RegBit  GCMD27          .off                //
    RegBit  GCMD28          .off                //
    RegBit  GCMD29          .off                //
    RegBit  GCMD30          .off                //
    RegBit  GCMD31          .off                //
  endRegDef
  //
  // ----- G1CON1 Bits -----------------------------------------------------
  regDef G1CON1
    RegBit  G1SIDL          .off                //
    RegBit  G1EN            .off                //

    RegBit  GCMDCNT0        .off                //
    RegBit  GCMDCNT1        .off                //
    RegBit  GCMDCNT2        .off                //
    RegBit  GCMDCNT3        .off                //
    RegBit  GCMDCNT4        .off                //
    RegBit  PUBPP0          .off                //
    RegBit  PUBPP1          .off                //
    RegBit  PUBPP2          .off                //
    RegBit  GCMDWMK0        .off                //
    RegBit  GCMDWMK1        .off                //
    RegBit  GCMDWMK2        .off                //
    RegBit  GCMDWMK3        .off                //
    RegBit  GCMDWMK4        .off                //
  endRegDef
  //
  // ----- G1STAT Bits -----------------------------------------------------
  regDef G1STAT
    RegBit  CMDMPT          .off                //
    RegBit  CMDFUL          .off                //
    RegBit  CMDLV           .off                //
    RegBit  HMRGN           .off                //
    RegBit  VMRGN           .off                //
    RegBit  CHRBUSY         .off                //
    RegBit  RCCBUSY         .off                //
    RegBit  IPUBUSY         .off                //
    RegBit  PUBUSY          .off                //
  endRegDef
  //
  // ----- G1IE Bits -----------------------------------------------------
  regDef G1IE
    RegBit  CMDMPTIE        .off                //
    RegBit  CMDFULIE        .off                //
    RegBit  CMDLVIE         .off                //
    RegBit  HMRGNIE         .off                //
    RegBit  VMRGNIE         .off                //
    RegBit  CHRIE           .off                //
    RegBit  RCCIE           .off                //
    RegBit  IPUIE           .off                //
    RegBit  PUIE            .off                //
  endRegDef
  //
  // ----- G1IR Bits -----------------------------------------------------
  regDef G1IR
    RegBit  CMDMPTIF        .off                //
    RegBit  CMDFULIF        .off                //
    RegBit  CMDLVIF         .off                //
    RegBit  HMRGNIF         .off                //
    RegBit  VMRGNIF         .off                //
    RegBit  CHRIF           .off                //
    RegBit  RCCIF           .off                //
    RegBit  IPUIF           .off                //
    RegBit  PUIF            .off                //
  endRegDef
  //
  // ----- G1W1ADRL Bits -----------------------------------------------------
  regDef G1W1ADRL
    RegBit  W1ADR0          .off                //
    RegBit  W1ADR1          .off                //
    RegBit  W1ADR2          .off                //
    RegBit  W1ADR3          .off                //
    RegBit  W1ADR4          .off                //
    RegBit  W1ADR5          .off                //
    RegBit  W1ADR6          .off                //
    RegBit  W1ADR7          .off                //
    RegBit  W1ADR8          .off                //
    RegBit  W1ADR9          .off                //
    RegBit  W1ADR10         .off                //
    RegBit  W1ADR11         .off                //
    RegBit  W1ADR12         .off                //
    RegBit  W1ADR13         .off                //
    RegBit  W1ADR14         .off                //
    RegBit  W1ADR15         .off                //
  endRegDef
  //
  // ----- G1W1ADRH Bits -----------------------------------------------------
  regDef G1W1ADRH
    RegBit  W1ADR16         .off                //
    RegBit  W1ADR17         .off                //
    RegBit  W1ADR18         .off                //
    RegBit  W1ADR19         .off                //
    RegBit  W1ADR20         .off                //
    RegBit  W1ADR21         .off                //
    RegBit  W1ADR22         .off                //
    RegBit  W1ADR23         .off                //
  endRegDef
  //
  // ----- G1W2ADRL Bits -----------------------------------------------------
  regDef G1W2ADRL
    RegBit  W2ADR0          .off                //
    RegBit  W2ADR1          .off                //
    RegBit  W2ADR2          .off                //
    RegBit  W2ADR3          .off                //
    RegBit  W2ADR4          .off                //
    RegBit  W2ADR5          .off                //
    RegBit  W2ADR6          .off                //
    RegBit  W2ADR7          .off                //
    RegBit  W2ADR8          .off                //
    RegBit  W2ADR9          .off                //
    RegBit  W2ADR10         .off                //
    RegBit  W2ADR11         .off                //
    RegBit  W2ADR12         .off                //
    RegBit  W2ADR13         .off                //
    RegBit  W2ADR14         .off                //
    RegBit  W2ADR15         .off                //
  endRegDef
  //
  // ----- G1W2ADRH Bits -----------------------------------------------------
  regDef G1W2ADRH
    RegBit  W2ADR16         .off                //
    RegBit  W2ADR17         .off                //
    RegBit  W2ADR18         .off                //
    RegBit  W2ADR19         .off                //
    RegBit  W2ADR20         .off                //
    RegBit  W2ADR21         .off                //
    RegBit  W2ADR22         .off                //
    RegBit  W2ADR23         .off                //
  endRegDef
  //
  // ----- G1PUW Bits -----------------------------------------------------
  regDef G1PUW
    RegBit  PUW0            .off                //
    RegBit  PUW1            .off                //
    RegBit  PUW2            .off                //
    RegBit  PUW3            .off                //
    RegBit  PUW4            .off                //
    RegBit  PUW5            .off                //
    RegBit  PUW6            .off                //
    RegBit  PUW7            .off                //
    RegBit  PUW8            .off                //
    RegBit  PUW9            .off                //
    RegBit  PUW10           .off                //
  endRegDef
  //
  // ----- G1PUH Bits -----------------------------------------------------
  regDef G1PUH
    RegBit  PUH0            .off                //
    RegBit  PUH1            .off                //
    RegBit  PUH2            .off                //
    RegBit  PUH3            .off                //
    RegBit  PUH4            .off                //
    RegBit  PUH5            .off                //
    RegBit  PUH6            .off                //
    RegBit  PUH7            .off                //
    RegBit  PUH8            .off                //
    RegBit  PUH9            .off                //
    RegBit  PUH10           .off                //
  endRegDef
  //
  // ----- G1DPADRL Bits -----------------------------------------------------
  regDef G1DPADRL
    RegBit  DPADR0          .off                //
    RegBit  DPADR1          .off                //
    RegBit  DPADR2          .off                //
    RegBit  DPADR3          .off                //
    RegBit  DPADR4          .off                //
    RegBit  DPADR5          .off                //
    RegBit  DPADR6          .off                //
    RegBit  DPADR7          .off                //
    RegBit  DPADR8          .off                //
    RegBit  DPADR9          .off                //
    RegBit  DPADR10         .off                //
    RegBit  DPADR11         .off                //
    RegBit  DPADR12         .off                //
    RegBit  DPADR13         .off                //
    RegBit  DPADR14         .off                //
    RegBit  DPADR15         .off                //
  endRegDef
  //
  // ----- G1DPADRH Bits -----------------------------------------------------
  regDef G1DPADRH
    RegBit  DPADR16         .off                //
    RegBit  DPADR17         .off                //
    RegBit  DPADR18         .off                //
    RegBit  DPADR19         .off                //
    RegBit  DPADR20         .off                //
    RegBit  DPADR21         .off                //
    RegBit  DPADR22         .off                //
    RegBit  DPADR23         .off                //
  endRegDef
  //
  // ----- G1DPW Bits -----------------------------------------------------
  regDef G1DPW
    RegBit  DPW0            .off                //
    RegBit  DPW1            .off                //
    RegBit  DPW2            .off                //
    RegBit  DPW3            .off                //
    RegBit  DPW4            .off                //
    RegBit  DPW5            .off                //
    RegBit  DPW6            .off                //
    RegBit  DPW7            .off                //
    RegBit  DPW8            .off                //
    RegBit  DPW9            .off                //
    RegBit  DPW10           .off                //
  endRegDef
  //
  // ----- G1DPH Bits -----------------------------------------------------
  regDef G1DPH
    RegBit  DPH0            .off                //
    RegBit  DPH1            .off                //
    RegBit  DPH2            .off                //
    RegBit  DPH3            .off                //
    RegBit  DPH4            .off                //
    RegBit  DPH5            .off                //
    RegBit  DPH6            .off                //
    RegBit  DPH7            .off                //
    RegBit  DPH8            .off                //
    RegBit  DPH9            .off                //
    RegBit  DPH10           .off                //
  endRegDef
  //
  // ----- G1DPWT Bits -----------------------------------------------------
  regDef G1DPWT
    RegBit  DPWT0           .off                //
    RegBit  DPWT1           .off                //
    RegBit  DPWT2           .off                //
    RegBit  DPWT3           .off                //
    RegBit  DPWT4           .off                //
    RegBit  DPWT5           .off                //
    RegBit  DPWT6           .off                //
    RegBit  DPWT7           .off                //
    RegBit  DPWT8           .off                //
    RegBit  DPWT9           .off                //
    RegBit  DPWT10          .off                //
  endRegDef
  //
  // ----- G1DPHT Bits -----------------------------------------------------
  regDef G1DPHT
    RegBit  DPHT0           .off                //
    RegBit  DPHT1           .off                //
    RegBit  DPHT2           .off                //
    RegBit  DPHT3           .off                //
    RegBit  DPHT4           .off                //
    RegBit  DPHT5           .off                //
    RegBit  DPHT6           .off                //
    RegBit  DPHT7           .off                //
    RegBit  DPHT8           .off                //
    RegBit  DPHT9           .off                //
    RegBit  DPHT10          .off                //
  endRegDef
  //
  // ----- G1CON2 Bits -----------------------------------------------------
  regDef G1CON2
    RegBit  DPMODE0         .off                //
    RegBit  DPMODE1         .off                //
    RegBit  DPMODE2         .off                //
    RegBit  DPBPP0          .off                //
    RegBit  DPBPP1          .off                //
    RegBit  DPBPP2          .off                //
    RegBit  DPTEST0         .off                //
    RegBit  DPTEST1         .off                //
    RegBit  DPSTGER0        .off                //
    RegBit  DPSTGER1        .off                //
    RegBit  DPGWDTH0        .off                //
    RegBit  DPGWDTH1        .off                //
  endRegDef
  //
  // ----- G1CON3 Bits -----------------------------------------------------
  regDef G1CON3
    RegBit  DPHSOE          .off                //
    RegBit  DPVSOE          .off                //
    RegBit  DPENOE          .off                //
    RegBit  DPPWROE         .off                //
    RegBit  DPHSPOL         .off                //
    RegBit  DPVSPOL         .off                //
    RegBit  DPENPOL         .off                //
    RegBit  DPCLKPOL        .off                //
    RegBit  DPPOWER         .off                //
    RegBit  DPPINOE         .off                //
  endRegDef
  //
  // ----- G1ACTDA Bits -----------------------------------------------------
  regDef G1ACTDA
    RegBit  ACTPIX0         .off                //
    RegBit  ACTPIX1         .off                //
    RegBit  ACTPIX2         .off                //
    RegBit  ACTPIX3         .off                //
    RegBit  ACTPIX4         .off                //
    RegBit  ACTPIX5         .off                //
    RegBit  ACTPIX6         .off                //
    RegBit  ACTPIX7         .off                //
    RegBit  ACTLINE0        .off                //
    RegBit  ACTLINE1        .off                //
    RegBit  ACTLINE2        .off                //
    RegBit  ACTLINE3        .off                //
    RegBit  ACTLINE4        .off                //
    RegBit  ACTLINE5        .off                //
    RegBit  ACTLINE6        .off                //
    RegBit  ACTLINE7        .off                //
  endRegDef
  //
  // ----- G1HSYNC Bits -----------------------------------------------------
  regDef G1HSYNC
    RegBit  HSST0           .off                //
    RegBit  HSST1           .off                //
    RegBit  HSST2           .off                //
    RegBit  HSST3           .off                //
    RegBit  HSST4           .off                //
    RegBit  HSST5           .off                //
    RegBit  HSST6           .off                //
    RegBit  HSST7           .off                //
    RegBit  HSLEN0          .off                //
    RegBit  HSLEN1          .off                //
    RegBit  HSLEN2          .off                //
    RegBit  HSLEN3          .off                //
    RegBit  HSLEN4          .off                //
    RegBit  HSLEN5          .off                //
    RegBit  HSLEN6          .off                //
    RegBit  HSLEN7          .off                //
  endRegDef
  //
  // ----- G1VSYNC Bits -----------------------------------------------------
  regDef G1VSYNC
    RegBit  VSST0           .off                //
    RegBit  VSST1           .off                //
    RegBit  VSST2           .off                //
    RegBit  VSST3           .off                //
    RegBit  VSST4           .off                //
    RegBit  VSST5           .off                //
    RegBit  VSST6           .off                //
    RegBit  VSST7           .off                //
    RegBit  VSLEN0          .off                //
    RegBit  VSLEN1          .off                //
    RegBit  VSLEN2          .off                //
    RegBit  VSLEN3          .off                //
    RegBit  VSLEN4          .off                //
    RegBit  VSLEN5          .off                //
    RegBit  VSLEN6          .off                //
    RegBit  VSLEN7          .off                //
  endRegDef
  //
  // ----- G1DBLCON Bits -----------------------------------------------------
  regDef G1DBLCON
    RegBit  HENST0          .off                //
    RegBit  HENST1          .off                //
    RegBit  HENST2          .off                //
    RegBit  HENST3          .off                //
    RegBit  HENST4          .off                //
    RegBit  HENST5          .off                //
    RegBit  HENST6          .off                //
    RegBit  HENST7          .off                //
    RegBit  VENST0          .off                //
    RegBit  VENST1          .off                //
    RegBit  VENST2          .off                //
    RegBit  VENST3          .off                //
    RegBit  VENST4          .off                //
    RegBit  VENST5          .off                //
    RegBit  VENST6          .off                //
    RegBit  VENST7          .off                //
  endRegDef
  //
  // ----- G1CLUT Bits -----------------------------------------------------
  regDef G1CLUT
    RegBit  CLUTRWEN        .off                //
    RegBit  CLUTTRD         .off                //
    RegBit  CLUTBUSY        .off                //
    RegBit  CLUTEN          .off                //

    RegBit  CLUTADR0        .off                //
    RegBit  CLUTADR1        .off                //
    RegBit  CLUTADR2        .off                //
    RegBit  CLUTADR3        .off                //
    RegBit  CLUTADR4        .off                //
    RegBit  CLUTADR5        .off                //
    RegBit  CLUTADR6        .off                //
    RegBit  CLUTADR7        .off                //
  endRegDef
  //
  // ----- G1CLUTWR Bits -----------------------------------------------------
  regDef G1CLUTWR
    RegBit  CLUTWR0         .off                //
    RegBit  CLUTWR1         .off                //
    RegBit  CLUTWR2         .off                //
    RegBit  CLUTWR3         .off                //
    RegBit  CLUTWR4         .off                //
    RegBit  CLUTWR5         .off                //
    RegBit  CLUTWR6         .off                //
    RegBit  CLUTWR7         .off                //
    RegBit  CLUTWR8         .off                //
    RegBit  CLUTWR9         .off                //
    RegBit  CLUTWR10        .off                //
    RegBit  CLUTWR11        .off                //
    RegBit  CLUTWR12        .off                //
    RegBit  CLUTWR13        .off                //
    RegBit  CLUTWR14        .off                //
    RegBit  CLUTWR15        .off                //
  endRegDef
  //
  // ----- G1CLUTRD Bits -----------------------------------------------------
  regDef G1CLUTRD
    RegBit  CLUTRD0         .off                //
    RegBit  CLUTRD1         .off                //
    RegBit  CLUTRD2         .off                //
    RegBit  CLUTRD3         .off                //
    RegBit  CLUTRD4         .off                //
    RegBit  CLUTRD5         .off                //
    RegBit  CLUTRD6         .off                //
    RegBit  CLUTRD7         .off                //
    RegBit  CLUTRD8         .off                //
    RegBit  CLUTRD9         .off                //
    RegBit  CLUTRD10        .off                //
    RegBit  CLUTRD11        .off                //
    RegBit  CLUTRD12        .off                //
    RegBit  CLUTRD13        .off                //
    RegBit  CLUTRD14        .off                //
    RegBit  CLUTRD15        .off                //
  endRegDef
  //
  // ----- G1MRGN Bits -----------------------------------------------------
  regDef G1MRGN
    RegBit  HBAMGN0         .off                //
    RegBit  HBAMGN1         .off                //
    RegBit  HBAMGN2         .off                //
    RegBit  HBAMGN3         .off                //
    RegBit  HBAMGN4         .off                //
    RegBit  HBAMGN5         .off                //
    RegBit  HBAMGN6         .off                //
    RegBit  HBAMGN7         .off                //
    RegBit  VBAMGN0         .off                //
    RegBit  VBAMGN1         .off                //
    RegBit  VBAMGN2         .off                //
    RegBit  VBAMGN3         .off                //
    RegBit  VBAMGN4         .off                //
    RegBit  VBAMGN5         .off                //
    RegBit  VBAMGN6         .off                //
    RegBit  VBAMGN7         .off                //
  endRegDef
  //
  // ----- G1CHRX Bits -----------------------------------------------------
  regDef G1CHRX
    RegBit  CURPOSX0        .off                //
    RegBit  CURPOSX1        .off                //
    RegBit  CURPOSX2        .off                //
    RegBit  CURPOSX3        .off                //
    RegBit  CURPOSX4        .off                //
    RegBit  CURPOSX5        .off                //
    RegBit  CURPOSX6        .off                //
    RegBit  CURPOSX7        .off                //
    RegBit  CURPOSX8        .off                //
    RegBit  CURPOSX9        .off                //
    RegBit  CURPOSX10       .off                //
  endRegDef
  //
  // ----- G1CHRY Bits -----------------------------------------------------
  regDef G1CHRY
    RegBit  CURPOSY0        .off                //
    RegBit  CURPOSY1        .off                //
    RegBit  CURPOSY2        .off                //
    RegBit  CURPOSY3        .off                //
    RegBit  CURPOSY4        .off                //
    RegBit  CURPOSY5        .off                //
    RegBit  CURPOSY6        .off                //
    RegBit  CURPOSY7        .off                //
    RegBit  CURPOSY8        .off                //
    RegBit  CURPOSY9        .off                //
    RegBit  CURPOSY10       .off                //
  endRegDef
  //
  // ----- G1IPU Bits -----------------------------------------------------
  regDef G1IPU
    RegBit  BFINAL          .off                //
    RegBit  IPUDONE         .off                //
    RegBit  WRAPERR         .off                //
    RegBit  LENERR          .off                //
    RegBit  BLCKERR         .off                //
    RegBit  HUFFERR         .off                //
  endRegDef
  //
  // ----- G1DBEN Bits -----------------------------------------------------
  regDef G1DBEN
    RegBit  GDBEN0          .off                //
    RegBit  GDBEN1          .off                //
    RegBit  GDBEN2          .off                //
    RegBit  GDBEN3          .off                //
    RegBit  GDBEN4          .off                //
    RegBit  GDBEN5          .off                //
    RegBit  GDBEN6          .off                //
    RegBit  GDBEN7          .off                //
    RegBit  GDBEN8          .off                //
    RegBit  GDBEN9          .off                //
    RegBit  GDBEN10         .off                //
    RegBit  GDBEN11         .off                //
    RegBit  GDBEN12         .off                //
    RegBit  GDBEN13         .off                //
    RegBit  GDBEN14         .off                //
    RegBit  GDBEN15         .off                //
  endRegDef


  // =======================================================================
  // RCON - Reset Control register
  //        see "08-reset.pdf" page 8-4

  // POR : power up reset occurred when set

  // BOR : brown out reset has occurred when set

  // IDLE : Device woke up from idle mode when set

  // SLEEP : Device woke up from sleep when set

  // WDTO : WDT timout occurred when set

  // SWDTEN
  // 0 : WDT is off (only when FWDTEN in config is off)
  // 1 : WDT is on

  // SWR : A software reset was executed when set

  // EXTR : Master Clear (pin) reset occurred when set

  // VREGS :
  // 0 : voltage regulator goes standby in sleep
  // 1 : voltage regulator stays on in sleep

  // IOPWR : Illegal opcode or uninitialized wreg caused a reset when set

  // TRAPR : A trap conflict reset occurred when set

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef RCON
    RegBit  POR             .off                // Clear power on reset flag
    RegBit  BOR             .off                // Clear brown out flag
    RegBit  IDLE            .off                // clear idle recovery flag
    RegBit  SLEEP           .off                // clear sleep recovery flag
    RegBit  WDTO            .off                // clear watchdog error flag
    RegBit  SWDTEN                .on           // Enable watchdog even when set soft in device config
    RegBit  SWR             .off                // clear software reset flag
    RegBit  EXTR            .off                // clear external master reset flag
    RegBit  VREGS           .off                // Voltage regulator into standby on sleep
    RegBit  CM              .off                //
    RegBit  IOPUWR          .off                // Clear illegal opcode flag
    RegBit  TRAPR           .off                // clear trap conflict flag
  endRegDef


  // =======================================================================
  // OSCCON - Oscillator Control register
  //          See : "07-oscillator.pdf" page 7-9

  // OSWEN - OScillator Switch Enable
  // 0 - Oscillator Switch is complete
  // 1 - Request  switch to NOSC value

  // LPOSCEN - Secondary (SOSC) OScillator enable
  // 0 - Disable Secondary Oscillator
  // 1 - Enable  Secondary Oscillator

  // CF - Fail detect
  // 0 - FSCM has not detected clock failure
  // 1 - FSCM has     detected clock failure

  // LOCK - PLL lock status
  // 0 - PLL is out of lock, start-up timer in progress or PLL is off
  // 1 - PLL is in lock or start-up time satisfied

  // CLKLOCK - Clock Lock enable
  // 0 - Clock switching is enabled, system clock source may be modified
  // 1 - Clock switching is disabled, system clock source is locked

  // NOSC - New OScillator Selectio Bits
  // 210
  // 000 - FRC                Fast RC
  // 001 - FRCPLL             Fast RC + PLL
  // 010 - XT_HS_EC           Primary oscillator
  // 011 - XTPLL_HSPLL_ECPLL  Primary oscillator + PLL
  // 100 - SOSC               Scondary OScillator
  // 101 - LPRC               Low Power RC Oscillator (LPRC)
  // 110 - FRCFDIV16          Fast RC Oscillator div 16
  // 111 - FRCDIVN            Fast RC Oscillator div N (default)

    %111 8 bits NOSC_MASK
    %000 8 bits NOSC_FRC
    %001 8 bits NOSC_FRCPLL
    %010 8 bits NOSC_XT_HS_EC
    %011 8 bits NOSC_XTPLL_HSPLL_EC_PLL
    %100 8 bits NOSC_SOSC
    %101 8 bits NOSC_LPRC
    %110 8 bits NOSC_FRCFDIV16
    %111 8 bits NOSC_FRCDIVN

  // COSC - Curent OScillator Selection bits - Read Only
  // 210
  // 000 - FRC                Fast RC
  // 001 - FRCPLL             Fast RC + PLL
  // 010 - XT_HS_EC           Primary oscillator
  // 011 - XTPLL_HSPLL_ECPLL  Primary oscillator + PLL
  // 100 - SOSC               Scondary OScillator
  // 101 - LPRC               Low Power RC Oscillator (LPRC)
  // 110 - FRCFDIV16          Fast RC Oscillator div 16
  // 111 - FRCDIVN            Fast RC Oscillator div N (default)

    %111 12 bits COSC_MASK
    %000 12 bits COSC_FRC
    %001 12 bits COSC_FRCPLL
    %010 12 bits COSC_XT_HS_EC
    %011 12 bits COSC_XTPLL_HSPLL_EC_PLL
    %100 12 bits COSC_SOSC
    %101 12 bits COSC_LPRC
    %110 12 bits COSC_FRCFDIV16
    %111 12 bits COSC_FRCDIVN

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef OSCCON
    RegBit  OSWEN           .off                // Do not switch
    RegBit  LPOSCEN         .off                // Disable secondary oscillator
    RegBit  POSCEN          .off                //
    RegBit  (CF)            .off                // Read only - Clock Fail
    RegBit  LOCK            .off                // Read only - PLL lock status
    RegBit  IOLOCK          .off                //
    RegBit  CLKLOCK               .on           // Lock the clock
            COSC_FRCDIVN          .use          // Read only - default value
            NOSC_FRCDIVN          .use          // New config same as current
  endRegDef


  // =======================================================================
  // CLKDIV - Clock Divisor register
  //          See : "07-oscillator.pdf" page 7-11
  //
  // Fosc = Fin * M / (N1 * N2)

  // DOZEN - Doze Enable
  // 0 - Processor clock/peripheral clock ratio set 1:1
  // 1 - DOZE<2:0> field specifies ratio between peripheral and processor clocks.

  // ROI - Recover On Interrupt
  // 0 - Interrupts do not un-doze
  // 1 - Interrupts clear the DOZEN bit

  // FRCDIV - Internal Fast RC oscillator post scaler
  // a98
  // 000 - FRC /   1 (default)
  // 001 - FRC /   2
  // 010 - FRC /   4
  // 011 - FRC /   8
  // 100 - FRC /  16
  // 101 - FRC /  32
  // 110 - FRC /  64
  // 111 - FRC / 256

    %000 8 bits FRCDIV_1
    %001 8 bits FRCDIV_2
    %010 8 bits FRCDIV_4
    %011 8 bits FRCDIV_8
    %100 8 bits FRCDIV_16
    %101 8 bits FRCDIV_32
    %110 8 bits FRCDIV_64
    %111 8 bits FRCDIV_256

  // DOZE - Processor clock reduction delect
  // edc
  // 000 - Fcy /   1
  // 001 - Fcy /   2
  // 010 - Fcy /   4
  // 011 - Fcy /   8 (default)
  // 100 - Fcy /  16
  // 101 - Fcy /  32
  // 110 - Fcy /  64
  // 111 - Fcy / 128

    %000 12 bits DOZE_1
    %001 12 bits DOZE_2
    %010 12 bits DOZE_4
    %011 12 bits DOZE_8
    %100 12 bits DOZE_16
    %101 12 bits DOZE_32
    %110 12 bits DOZE_64
    %111 12 bits DOZE_128

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef CLKDIV
    RegBit  G1CLKSEL        .off                //
    RegBit  PLLEN           .off                //
    RegBit  DOZEN           .off                // Doze enable (off is default)
    RegBit  ROI                   .on           // Recover on interrupt (off is default)
    RegBit  CPDIV0          .off                //
    RegBit  CPDIV1          .off                //
            FRCDIV_1              .use          // RC post scaler (1, 2, 4, 8, 16, 32, 64, 256; 1 is default)
            DOZE_8                .use          // Doze ratio (1, 2, 4, 8, 16, 32, 64, 128; 8 is default)
  endRegDef


  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef CLKDIV2
    RegBit  GCLKDIV0        .off                //
    RegBit  GCLKDIV1        .off                //
    RegBit  GCLKDIV2        .off                //
    RegBit  GCLKDIV3        .off                //
    RegBit  GCLKDIV4        .off                //
    RegBit  GCLKDIV5        .off                //
    RegBit  GCLKDIV6        .off                //
  endRegDef


  // =======================================================================
  //

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef NVMCON
    RegBit  ERASE           .off                //
    RegBit  WRERR           .off                //
    RegBit  WREN            .off                //
    RegBit  WR              .off                //
    RegBit  NVMOP0          .off                //
    RegBit  NVMOP1          .off                //
    RegBit  NVMOP2          .off                //
    RegBit  NVMOP3          .off                //
  endRegDef


  // =======================================================================
  //

  0 constant NVMKEY_IniVal


  // =======================================================================
  //

  // . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  regDef PMD1
    RegBit  ADC1MD          .off                //
    RegBit  SPI1MD          .off                //
    RegBit  SPI2MD          .off                //
    RegBit  U1MD            .off                //
    RegBit  U2MD            .off                //
    RegBit  I2C1MD          .off                //
    RegBit  T1MD            .off                //
    RegBit  T2MD            .off                //
    RegBit  T3MD            .off                //
    RegBit  T4MD            .off                //
    RegBit  T5MD            .off                //
  endRegDef

  regDef PMD2
    RegBit  OC1MD           .off                //
    RegBit  OC2MD           .off                //
    RegBit  OC3MD           .off                //
    RegBit  OC4MD           .off                //
    RegBit  OC5MD           .off                //
    RegBit  OC6MD           .off                //
    RegBit  OC7MD           .off                //
    RegBit  OC8MD           .off                //
    RegBit  IC1MD           .off                //
    RegBit  IC2MD           .off                //
    RegBit  IC3MD           .off                //
    RegBit  IC4MD           .off                //
    RegBit  IC5MD           .off                //
    RegBit  IC6MD           .off                //
    RegBit  IC7MD           .off                //
    RegBit  IC8MD           .off                //
  endRegDef

  regDef PMD3
    RegBit  I2C2MD          .off                //
    RegBit  I2C3MD          .off                //
    RegBit  U3MD            .off                //
    RegBit  CRCMD           .off                //
    RegBit  RTCCMD          .off                //
    RegBit  CMPMD           .off                //
  endRegDef

  regDef PMD4
    RegBit  USB1MD          .off                //
    RegBit  LVDMD           .off                //
    RegBit  CTMUMD          .off                //
    RegBit  REFOMD          .off                //
    RegBit  U4MD            .off                //
    RegBit  UPWMMD          .off                //
  endRegDef

  regDef PMD5
    RegBit  OC9MD           .off                //
    RegBit  IC9MD           .off                //
  endRegDef

  regDef PMD6
    RegBit  SPI3MD          .off                //
    RegBit  GFX1MD          .off                //
  endRegDef


  // =======================================================================

endModule

